/**********************************************************************************************************************************/
/* Project      : TAPER                                                                                                           */
/* Module       : TPCUST.C                                                                                                        */
/* Description  : Custom Info - Customised handlers                                                                               */
/* Version type : Program module                                                                                                  */
/* Last changed : 02-01-1999  15:40                                                                                               */
/* Update count : 10                                                                                                              */
/* OS type      : PC (DOS)                                                                                                        */
/* Compiler     : Watcom C32 Optimizing Compiler Version 10.x                                                                     */
/* Linker       : Watcom C Linker                                                                                                 */
/*                                                                                                                                */
/*                Copyleft (C) 1997-2001 ThunderWare Research Center, written by Martijn van der Heide.                           */
/*                                                                                                                                */
/*                This program is free software; you can redistribute it and/or                                                   */
/*                modify it under the terms of the GNU General Public License                                                     */
/*                as published by the Free Software Foundation; either version 2                                                  */
/*                of the License, or (at your option) any later version.                                                          */
/*                                                                                                                                */
/*                This program is distributed in the hope that it will be useful,                                                 */
/*                but WITHOUT ANY WARRANTY; without even the implied warranty of                                                  */
/*                MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                                                   */
/*                GNU General Public License for more details.                                                                    */
/*                                                                                                                                */
/*                You should have received a copy of the GNU General Public License                                               */
/*                along with this program; if not, write to the Free Software                                                     */
/*                Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                                     */
/*                                                                                                                                */
/**********************************************************************************************************************************/

#define __TPCUST_MAIN__

#include "wbench32/wbench32.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <io.h>
#include <fcntl.h>
#include <dos.h>
#include <direct.h>
#include "tpmain.h"
#include "tpconfig.h"
#include "tpscreen.h"
#include "tptape.h"
#include "tpedit.h"
#include "tpcust.h"

#pragma pack                ()

/**********************************************************************************************************************************/
/* Define the static variables                                                                                                    */
/**********************************************************************************************************************************/

static char        _GeneralDescription[257];

/**********************************************************************************************************************************/
/* >>>>> Used by the POKEs block                                                                                                  */
/**********************************************************************************************************************************/

#define MAXPOKNAME    30
#define FLAG_IGNBANK  0x08
#define FLAG_ASK      0x10
#define FLAG_NOORG    0x20
#define FLAG_ALL      0xF8

static struct PokeEntry_s                                                                                       /* One poke entry */
{
  struct PokeEntry_s *Next;
  struct PokeEntry_s *Previous;
  byte                Bank;
  word                Address;
  byte                Set;
  byte                Org;
  byte                Flags;
};

static struct GamePokes_s                                                                                 /* One poke description */
{
  struct GamePokes_s *Next;
  struct GamePokes_s *Previous;
  char                Name[MAXPOKNAME + 1];
  struct PokeEntry_s *Pokes;
} *GamePokes = NULL;

static comboent       SelectedPoke       = 0;
static comboent       SelectedEntry      = 0;

/**********************************************************************************************************************************/
/* >>>>> Used by the Instructions block                                                                                           */
/**********************************************************************************************************************************/

static char          _TempFileName[MAXDIR + 1];
static char          _Path[MAXDIR + 1]  = "\0";
static char          _File[13]          = "\0";
static char          _Mask[13]          = "*.SCR";

/**********************************************************************************************************************************/
/* >>>>> Used by the Spectrum Screen block                                                                                        */
/**********************************************************************************************************************************/

static word          _BlockOffset;
static short         _AddHeight;

static struct ComboInfo PokesWindow;
static struct ComboInfo EntriesWindow;

static struct WindowInfo _TBWindow = {147, 65, 347, 358, 347, 358, 347, 358, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                      WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL,
                                      "Edit Custom Info - Spectrum Screen"};

/**********************************************************************************************************************************/
/* Define the static functions                                                                                                    */
/**********************************************************************************************************************************/

static byte _GetGeneralDescription (char *Header, char *DefaultText);
static bool _GetNewFileName        (char *Operation, char *OldName, char *ExtraOptions, char *ExtReason);
static void _DestroyPokes          (void);
static void _DestroyEntries        (struct PokeEntry_s *PokeEntries);
static bool _SetupPokesWindow      (void);
static bool _SetupEntriesWindow    (comboent SelectedLine);
static bool _ReadGamePokes         (dword OldSize);
static bool _WriteGamePokes        (dword *NewSize);

static int  _HandleAddPoke         (struct WindowInfo *Window, struct ButtonInfo *Button);
static int  _HandleDelPoke         (struct WindowInfo *Window, struct ButtonInfo *Button);
static bool _HandleUpdatePoke      (void);
static int  _HandleAddEntry        (struct WindowInfo *Window, struct ButtonInfo *Button);
static int  _HandleDelEntry        (struct WindowInfo *Window, struct ButtonInfo *Button);
static bool _HandleUpdateEntry     (void);

static void _TBTimedFlash          (struct WindowInfo *Window);

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> START OF THE INTERNAL LIBRARY ROUTINES <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/**********************************************************************************************************************************/

static byte _GetGeneralDescription (char *Header, char *DefaultText)

/**********************************************************************************************************************************/
/* Pre   : `Header' holds the Custom Info ID to be written in the window header, `DefaultText' points to the default value.       */
/* Post  : The General Description field has been edited into the global _GeneralDescription and its length is returned.          */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct WindowInfo      CWindow = {195, 160, 251, 160, 251, 160, 251, 160, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                    WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL,
                                    "Edit General Description"};
  struct InputBoxInfo    BGenDes = {32, 30, 30, 7, DBLACK, DWHITE, NULL, 0, _GeneralDescription, BF_NOSURROUND | BF_KEYSELECT,
                                    255, 10};
  struct ButtonInfo      BOk     = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 98, 135, 8, DBLACK, DWHITE, NULL,
                                    1, "@Ok", NULL};
  struct ButtonInfo     *ButtonP;
  bool                   Done    = FALSE;
  char                   ShadString[80];

  strcpy (_GeneralDescription, DefaultText);
  AddButton (&CWindow, &BOk, TRUE);
  AddInputBox (&CWindow, &BGenDes, FALSE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  sprintf (ShadString, "%s - General Description", Header);
  WDrawShadowBox (&CWindow, 1, 0, 11, 40, ShadString);
  while (!Done)
  {
    switch (HandleWindow (&CWindow, &ButtonP))
    {
      case WINH_RNOHAND : Done = TRUE; break;                                                                             /* (Ok) */
      case WINH_CLOSE   : Done = TRUE; break;
    }
    if (GlobalKey == K_ESC)                                                                                    /* <ESC> stops too */
      Done = TRUE;
    GlobalKey = 0x0000;
    MouseStatus ();
  }
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
  if (!stricmp (_GeneralDescription, DefaultText))
    _GeneralDescription[0] = '\0';
  return (strlen (_GeneralDescription));
}

static bool _GetNewFileName (char *Operation, char *OldName, char *ExtraOptions, char *ExtReason)

/**********************************************************************************************************************************/
/* Pre   : `Operation' is a string for the file requester header, `OldName' is the current path name.                             */
/*         `ExtraOptions' is non-NULL if extra command line options may be required to `ExtReason'.                               */
/*         Both `OldName' and `ExtraOption' must be path fields in the TPConfig structure!                                        */
/* Post  : A new filename has been asked with a FileRequester, which is returned in `OldName'.                                    */
/*         If a new filename has been selected, TRUE is returned.                                                                 */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  char           Path[MAXDIR + 1];
  char           File[13];
  char           Mask[13] = "*.*";
  char           ExtBuffer[65];
  short register Cnt;
  bool           Selected;

  if (OldName[0] != '\0')                                                                              /* A path has been given ? */
  {
    Cnt = strlen (OldName);
    while (-- Cnt >= 0 && OldName[Cnt] != '\\' && OldName[Cnt] != ':')
      ;
    if (Cnt >= 0)                                                                                    /* Is there a leading path ? */
    {
      strncpy (Path, OldName, Cnt);
      Path[Cnt] = '\0';
    }
    else
      Path[0] = '\0';
    strcpy (File, OldName + Cnt + 1);
  }
  else
    Path[0] = File[0] = '\0';
  if ((Selected = FileRequester (180, 50, 48, 29, 20, Path, File, Mask, Operation, TRUE, FALSE)))
  {
    MakeFullPath (OldName, Path, File);
    if (ExtraOptions != NULL)                                              /* The program could need extra command line options ? */
    {
      if (InputRequester (-1, -1, NULL, NULL, ExtBuffer, 64, TRUE, FALSE, "Add Command-Line Options",
                          "If this program requires extra command line options to %s, please provide these now.", ExtReason))
        strcpy (ExtraOptions, ExtBuffer);
    }
  }
  if (Selected || MaxFloppyIdChanged)
    WriteConfig ();
  return (Selected);
}

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> START OF THE INTERNAL POKE LIBRARY ROUTINES <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/**********************************************************************************************************************************/

static void _DestroyEntries (struct PokeEntry_s *PokeEntries)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : All memory taken by the given PokeEntry array has been released.                                                       */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct PokeEntry_s *OEntry;

  OEntry = PokeEntries;
  while (OEntry->Next != NULL)
  {
    OEntry = OEntry->Next;
    free (OEntry->Previous);
  }
  free (OEntry);
}

static void _DestroyPokes (void)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : All memory taken by the `GamePokes' database has been released.                                                        */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct GamePokes_s *OPokes = NULL;

  if (GamePokes == NULL)
    return;
  OPokes = GamePokes;
  while (OPokes->Next != NULL)
  {
    _DestroyEntries (OPokes->Pokes);
    OPokes = OPokes->Next;
    free (OPokes->Previous);
  }
  _DestroyEntries (OPokes->Pokes);
  free (OPokes);
  GamePokes = NULL;
}

static bool _SetupPokesWindow (void)

/**********************************************************************************************************************************/
/* Pre   : The PokesWindow must have been created.                                                                                */
/* Post  : All poke descriptions are inserted into a reinitialized PokesWindow and the entries have been redrawn.                 */
/*         The return value is FALSE if an out-of-memory error occured (which is not reported here).                              */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct GamePokes_s *OPokes;
  struct ComboEntry   NEntry;
  bool                AllOk = TRUE;

  OPokes = GamePokes;
  NEntry.Selected = 0x00;
  ReInitComboWindow (&PokesWindow, 16);
  while (OPokes != NULL)
  {
    strcpy (NEntry.Text, OPokes->Name);
    NEntry.Identifier = (dword)OPokes;
    if (!InsertComboEntry (&PokesWindow, &NEntry, NULL, -1))
      AllOk = FALSE;
    OPokes = OPokes->Next;
  }
  PrintComboEntries (&PokesWindow);
  return (AllOk);
}

static bool _SetupEntriesWindow (comboent SelectedLine)

/**********************************************************************************************************************************/
/* Pre   : The EntriesWindow must have been created. `SelectedLine' holds the poke description number.                            */
/* Post  : All poke values belonging to the given description are inserted into a reinitialized EntriesWindow and the entries     */
/*         have been redrawn. The return value is FALSE if an out-of-memory error occured (which is not reported here).           */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct PokeEntry_s *OEntry;
  struct ComboEntry   NEntry;
  bool                AllOk = TRUE;
  char                ABank[2];
  char                ASet[4];

  if (PokesWindow.NumEntries == -1)
    OEntry = NULL;
  else
  {
    GetComboEntry (&PokesWindow, &NEntry, SelectedLine);
    OEntry = ((struct GamePokes_s *)NEntry.Identifier)->Pokes;
  }
  NEntry.Selected = 0x00;
  ReInitComboWindow (&EntriesWindow, 16);
  while (OEntry != NULL)
  {
    if (!(OEntry->Flags & FLAG_IGNBANK))
      sprintf (ABank, "%u", OEntry->Bank);
    else
      strcpy (ABank, "-");
    if (!(OEntry->Flags & FLAG_ASK))
      sprintf (ASet, "%3u", OEntry->Set);
    else
      strcpy (ASet, "ASK");
    if (!(OEntry->Flags & FLAG_NOORG))
      sprintf (NEntry.Text, "%s  %5u,%s (%3u)", ABank, OEntry->Address, ASet, OEntry->Org);
    else
      sprintf (NEntry.Text, "%s  %5u,%s", ABank, OEntry->Address, ASet);
    NEntry.Identifier = (dword)OEntry;
    if (!InsertComboEntry (&EntriesWindow, &NEntry, NULL, -1))
      AllOk = FALSE;
    OEntry = OEntry->Next;
  }
  PrintComboEntries (&EntriesWindow);
  return (AllOk);
}

static bool _ReadGamePokes (dword OldSize)

/**********************************************************************************************************************************/
/* Pre   : `OldSize' holds the current size of the block.                                                                         */
/* Post  : The table, which is in `WorkBuffer' has been setup into the pokes lists.                                               */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct GamePokes_s *NPokes;
  struct GamePokes_s *OPokes = NULL;
  struct PokeEntry_s *NEntry;
  struct PokeEntry_s *OEntry = NULL;
  byte      register  CntTrainers;
  byte      register  CntPokes;
  byte               *BIndex;

  _GeneralDescription[0] = '\0';
  if (OldSize == 0)
    return (TRUE);
  strncpy (_GeneralDescription, WorkBuffer + 1, *WorkBuffer);                                   /* Read General Description field */
  _GeneralDescription[*WorkBuffer] = '\0';
  OPokes = GamePokes;
  BIndex = WorkBuffer + (*WorkBuffer) + 1;                                                  /* Point to the start of the trainers */
  CntTrainers = *(BIndex ++);                                                              /* Get number of trainers in the block */
  for ( ; CntTrainers > 0 ; CntTrainers --)                                                                  /* Read all trainers */
  {
    if ((NPokes = (struct GamePokes_s *)malloc (sizeof (struct GamePokes_s))) == NULL)     return (FALSE);  
    NPokes->Next = NULL;
    NPokes->Pokes = NULL;
    if (OPokes == NULL)
    {
      GamePokes = NPokes;
      NPokes->Previous = NULL;
    }
    else
    {
      OPokes->Next = NPokes;
      NPokes->Previous = OPokes;
    }
    if (*BIndex > MAXPOKNAME)                                                            /* Description more than 30 characters ? */
    {
      strncpy (NPokes->Name, BIndex + 1, MAXPOKNAME);                                                     /* Strip the rest of it */
      NPokes->Name[MAXPOKNAME] = '\0';
    }
    else
    {
      strncpy (NPokes->Name, BIndex + 1, *BIndex);
      NPokes->Name[*BIndex] = '\0';
    }
    BIndex += (*BIndex) + 1;                                                                       /* Point past the trainer name */
    CntPokes = *(BIndex ++);                                                               /* Get number of pokes in this trainer */
    OEntry = NULL;
    for ( ; CntPokes > 0 ; CntPokes --)                                                          /* Read all pokes in the trainer */
    {
      if ((NEntry = (struct PokeEntry_s *)malloc (sizeof (struct PokeEntry_s))) == NULL)    return (FALSE);
      NEntry->Next = NULL;
      if (OEntry == NULL)
      {
        NPokes->Pokes = NEntry;
        NEntry->Previous = NULL;
      }
      else
      {
        OEntry->Next = NEntry;
        NEntry->Previous = OEntry;
      }
      NEntry->Flags = (*BIndex) & FLAG_ALL;                                                 /* Split the flags off the bank field */
      NEntry->Bank = (*BIndex) & ~FLAG_ALL;
      NEntry->Address = *((word *)(BIndex + 1));
      NEntry->Set = *(BIndex + 3);
      NEntry->Org = *(BIndex + 4);
      BIndex += 5;                                                                 /* Prepare to do the next poke in this trainer */
      OEntry = NEntry;
    }
    OPokes = NPokes;
  }
  return (TRUE);
}

static bool _WriteGamePokes (dword *NewSize)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : The pokes table has been reconstructed into WorkBuffer, from the lists in memory.                                      */
/*         The new size of the block is returned in `NewSize'.                                                                    */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct GamePokes_s *NPokes;
  struct PokeEntry_s *NEntry;
  byte               *BIndex;
  byte               *BNumTrainers;
  byte               *BNumPokes;

  NPokes = GamePokes;
  *WorkBuffer = strlen (_GeneralDescription);
  strcpy (WorkBuffer + 1, _GeneralDescription);
  BIndex = WorkBuffer + (*WorkBuffer) + 1;
  BNumTrainers = BIndex ++;                                                 /* Keep location where the number of trainers is kept */
  *BNumTrainers = 0;
  while (NPokes != NULL)
  {
    while (NPokes != NULL && NPokes->Pokes == NULL)                                           /* Only a name, no pokes attached ? */
      NPokes = NPokes->Next;                                                                                           /* Skip it */
    if (NPokes != NULL)                                                                                 /* Still anything to do ? */
    {
      (*BNumTrainers) ++;
      *BIndex = strlen (NPokes->Name);
      strcpy (BIndex + 1, NPokes->Name);
      BIndex += (*BIndex) + 1;
      BNumPokes = BIndex ++;                                       /* Keep location where number of pokes in this trainer is kept */
      *BNumPokes = 0;
      NEntry = NPokes->Pokes;
      while (NEntry != NULL)
      {
        (*BNumPokes) ++;
        *BIndex = (NEntry->Flags | NEntry->Bank);
        *((word *)(BIndex + 1)) = NEntry->Address;
        *(BIndex + 3) = NEntry->Set;
        *(BIndex + 4) = NEntry->Org;
        BIndex += 5;
        NEntry = NEntry->Next;
      }
      NPokes = NPokes->Next;
    }
  }
  *NewSize = ((dword)BIndex - (dword)WorkBuffer);
  return (TRUE);
}

static int _HandleAddPoke (struct WindowInfo *Window, struct ButtonInfo *Button)

/**********************************************************************************************************************************/
/* Pre   : The `Add' button has been pressed (or released).                                                                       */
/* Post  : A new poke entry has been created.                                                                                     */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct GamePokes_s *OPokes;
  struct GamePokes_s *NPokes;
  struct ComboEntry   NEntry;
  struct ComboEntry   CEntry;

  if (!(Button->Flags & BF_PRESSED))
  {
    NEntry.Text[0] = '\0';
    if (InputRequester (-1, -1, NULL, NULL, NEntry.Text, MAXPOKNAME, TRUE, FALSE, "Add Trainer Description",
                        "{CGive the description of the new trainer}C"))
    {
      if ((NPokes = (struct GamePokes_s *)malloc (sizeof (struct GamePokes_s))) == NULL)
        FatalError (1, _ErrOutOfMemory, sizeof (struct GamePokes_s));
      strcpy (NPokes->Name, NEntry.Text);
      NPokes->Next = NULL;
      NPokes->Pokes = NULL;
      if (GamePokes == NULL)
      {
        GamePokes = NPokes;
        NPokes->Previous = NULL;
      }
      else
      {
        GetComboEntry (&PokesWindow, &CEntry, PokesWindow.NumEntries);
        OPokes = (struct GamePokes_s *)CEntry.Identifier;                                                      /* (Last in chain) */
        OPokes->Next = NPokes;
        NPokes->Previous = OPokes;
      }
      NEntry.Identifier = (dword)NPokes;
      NEntry.Selected = 0x00;
      if (!InsertComboEntry (&PokesWindow, &NEntry, NULL, -1))
        FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
      SelectedPoke = PokesWindow.NumEntries;
      SelectedEntry = 0;
      _SetupEntriesWindow (SelectedPoke);
      SelectEntry (&PokesWindow, SelectedPoke);
      SelectEntry (&EntriesWindow, SelectedEntry);
      _HandleAddEntry (Window, Button);                                                                /* Auto-insert first entry */
      return (WINH_EDITEND);                                                             /* Signal: (un)ghost appropriate buttons */
    }
  }
  return (WINH_NONE);
}

static int _HandleDelPoke (struct WindowInfo *Window, struct ButtonInfo *Button)

/**********************************************************************************************************************************/
/* Pre   : The `Delete' button has been pressed (or released).                                                                    */
/* Post  : The current poke has been removed.                                                                                     */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct GamePokes_s *OPokes;
  struct PokeEntry_s *OEntry;
  struct ComboEntry   NEntry;

  Window = Window; /* keep compiler happy */
  if (!(Button->Flags & BF_PRESSED) && PokesWindow.NumEntries >= 0)
  {
    GetComboEntry (&PokesWindow, &NEntry, SelectedPoke);
    if (TPConfig.ConfirmDelete)                                                               /* Confirmation needed: handle it ? */
      if (QuestionRequester (-1, -1, FALSE, FALSE, NULL, "@Cancel", NULL, NULL, NULL, "Confirm Delete",
                             "Are you sure you want to delete the trainer\n\"%s\"?", NEntry.Text))
        return (WINH_NONE);
    OPokes = (struct GamePokes_s *)NEntry.Identifier;
    OEntry = OPokes->Pokes;
    if (OEntry != NULL)                                                                                    /* Remove poke entries */
    {
      while (OEntry->Next != NULL)
      {
        OEntry = OEntry->Next;
        free (OEntry->Previous);
      }
      free (OEntry);
    }
    if (SelectedPoke == 0)                                                                                /* Removing first one ? */
    {
      GamePokes = OPokes->Next;
      OPokes->Next->Previous = NULL;
    }
    else
    {
      OPokes->Previous->Next = OPokes->Next;                                                                            /* Unlink */
      OPokes->Next->Previous = OPokes->Previous;
    }
    free (OPokes);
    DeleteComboEntry (&PokesWindow, SelectedPoke);
    if (SelectedPoke > PokesWindow.NumEntries)
      SelectedPoke --;
    SelectedEntry = 0;
    _SetupEntriesWindow (SelectedPoke);
    SelectEntry (&PokesWindow, SelectedPoke);
    SelectEntry (&EntriesWindow, SelectedEntry);
    return (WINH_EDITEND);                                                               /* Signal: (un)ghost appropriate buttons */
  }
  return (WINH_NONE);
}

static bool _HandleUpdatePoke (void)

/**********************************************************************************************************************************/
/* Pre   : A poke description has been double-clicked.                                                                            */
/* Post  : The current poke description has been updated.                                                                         */
/*         The return value is TRUE if the value has been changed.                                                                */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct GamePokes_s *OPokes;
  char               *CEntry;
  struct ComboEntry   NEntry;

  GetComboEntry (&PokesWindow, &NEntry, SelectedPoke);
  OPokes = (struct GamePokes_s *)NEntry.Identifier;
  CEntry = NEntry.Text;
  if (InputRequester (-1, -1, NULL, NULL, CEntry, MAXPOKNAME, TRUE, FALSE, "Update Trainer Description",
                      "{CGive the new description of the trainer}C"))
    if (strcmp (CEntry, OPokes->Name))                                                                      /* Actually changed ? */
    {
      PutComboEntry (&PokesWindow, &NEntry, SelectedPoke, TRUE);
      strcpy (OPokes->Name, CEntry);
      SelectEntry (&PokesWindow, SelectedPoke);
      return (TRUE);
    }
  return (FALSE);
}

static int _HandleAddEntry (struct WindowInfo *Window, struct ButtonInfo *Button)

/**********************************************************************************************************************************/
/* Pre   : The `Insert' button has been pressed (or released).                                                                    */
/* Post  : A new poke entry has been inserted.                                                                                    */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct WindowInfo       CWindow  = {190, 220, 160, 160, 160, 160, 160, 160, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                      WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "Add Poke"};
  struct ButtonInfo       BBank    = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 72, 40, 1, DBLACK, DWHITE, NULL,
                                      1, NULL, NULL};
  struct ButtonInfo       BNoBank  = {BT_PUSH, BC_GADGET, BF_FWIDTH | BF_KEYSELECT, 85, 40, 6, DBLACK, DWHITE, NULL,
                                      9, "\007", NULL};
  struct ButtonInfo       BAddress = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 72, 60, 5, DBLACK, DWHITE, NULL,
                                      2, NULL, NULL};
  struct ButtonInfo       BSet     = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 72, 80, 3, DBLACK, DWHITE, NULL,
                                      3, NULL, NULL};
  struct ButtonInfo       BAskSet  = {BT_PUSH, BC_GADGET, BF_FWIDTH | BF_KEYSELECT, 97, 80, 6, DBLACK, DWHITE, NULL,
                                      8, "\007", NULL};
  struct ButtonInfo       BOrg     = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 72, 100, 3, DBLACK, DWHITE, NULL,
                                      4, NULL, NULL};
  struct ButtonInfo       BNoOrg   = {BT_PUSH, BC_GADGET, BF_FWIDTH | BF_KEYSELECT, 97, 100, 6, DBLACK, DWHITE, NULL,
                                      5, "\007", NULL};
  struct ButtonInfo       BCancel  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 10, 135, 8, DBLACK, DWHITE, NULL,
                                      6, "@Cancel", NULL};
  struct ButtonInfo       BOk      = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT | BF_GHOSTED, 96, 135, 8, DBLACK, DWHITE, NULL,
                                      7, "@Ok", NULL};
  struct ButtonInfo      *ButtonP;
  struct GamePokes_s     *OPokes;
  struct PokeEntry_s     *OEntry;
  struct PokeEntry_s     *NEntry;
  struct ComboEntry       NCEntry;
  char                    ABank[2];
  char                    AAddress[6];
  char                    ASet[4];
  char                    AOrg[4];
  bool                    Finished;
  bool                    Cancelled = FALSE;
  bool                    Accepted = FALSE;
  bool                    ChangeOKButton;
  byte                    Flags = FLAG_IGNBANK | FLAG_NOORG;

  Window = Window; /* keep compiler happy */
  BBank.Contents = ABank;
  BAddress.Contents = AAddress;
  BSet.Contents = ASet;
  BOrg.Contents = AOrg;
  if (!(Button->Flags & BF_PRESSED))
  {
    AAddress[0] = ASet[0] = '\0';
    if (Flags & FLAG_NOORG)
    {
      strcpy (AOrg, "???");
      BOrg.Flags |= BF_GHOSTED;
    }
    else
      AOrg[0] = '\0';
    if (Flags & FLAG_IGNBANK)
    {
      strcpy (ABank, "-");
      BBank.Flags |= BF_GHOSTED;
    }
    else
      ABank[0] = '\0';
    AddButton (&CWindow, &BAddress, FALSE);
    AddButton (&CWindow, &BSet, FALSE);
    AddButton (&CWindow, &BAskSet, FALSE);
    AddButton (&CWindow, &BOrg, FALSE);
    AddButton (&CWindow, &BNoOrg, FALSE);
    AddButton (&CWindow, &BCancel, FALSE);
    AddButton (&CWindow, &BOk, TRUE);
    AddButton (&CWindow, &BBank, FALSE);
    AddButton (&CWindow, &BNoBank, FALSE);
    DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
    Wprintf (&CWindow, 1, TEXT_CENTER, "Setup new poke parameters");
    Wprintf (&CWindow,  4, 2, "{+1Bank}N");
    Wprintf (&CWindow,  6, 2, "{+1Address}N");
    Wprintf (&CWindow,  8, 2, "{+1Value}N");
    Wprintf (&CWindow, 10, 2, "{+1Original}N");
    PushKey (&CWindow, K_ENTER, 0x0000);                                            /* Put <RETURN> in the keyboard queue, so the */
                                                                                    /* address box will be started up immediately */
    while (!Accepted)
    {
      Finished = FALSE;
      Cancelled = TRUE;
      while (!Finished)
      {
        ChangeOKButton = FALSE;
        switch (HandleWindow (&CWindow, &ButtonP))
        {
          case WINH_CLOSE   : Finished = TRUE; break;
          case WINH_RNOHAND : switch (ButtonP->ReactNum)
                              {
                                case 6 : Finished = TRUE; Cancelled = TRUE; break;                                    /* (Cancel) */
                                case 7 : Finished = TRUE; Cancelled = FALSE; break;                                       /* (Ok) */
                                case 5 : if (Flags & FLAG_NOORG)                                                     /* ORG cycle */
                                         {
                                           Flags &= ~FLAG_NOORG;
                                           strcpy (AOrg, "0");
                                           BOrg.Flags &= ~BF_GHOSTED;
                                           UpdateButtonText (&CWindow, &BOrg, BF_FITTXT);
                                         }
                                         else
                                         {
                                           Flags |= FLAG_NOORG;
                                           strcpy (AOrg, "???");
                                           BOrg.Flags |= BF_GHOSTED;
                                           UpdateButtonText (&CWindow, &BOrg, BF_FITTXT);
                                         }
                                         break;
                                case 8 : if (Flags & FLAG_ASK)                                                       /* SET cycle */
                                         {
                                           Flags &= ~FLAG_ASK;
                                           strcpy (ASet, "0");
                                           BSet.Flags &= ~BF_GHOSTED;
                                           UpdateButtonText (&CWindow, &BSet, BF_FITTXT);
                                         }
                                         else
                                         {
                                           Flags |= FLAG_ASK;
                                           strcpy (ASet, "ASK");
                                           BSet.Flags |= BF_GHOSTED;
                                           UpdateButtonText (&CWindow, &BSet, BF_FITTXT);
                                         }
                                         ChangeOKButton = TRUE;
                                         break;
                                case 9 : if (Flags & FLAG_IGNBANK)                                                  /* BANK cycle */
                                         {
                                           Flags &= ~FLAG_IGNBANK;
                                           strcpy (ABank, "0");
                                           BBank.Flags &= ~BF_GHOSTED;
                                           UpdateButtonText (&CWindow, &BBank, BF_FITTXT);
                                         }
                                         else
                                         {
                                           Flags |= FLAG_IGNBANK;
                                           strcpy (ABank, "-");
                                           BBank.Flags |= BF_GHOSTED;
                                           UpdateButtonText (&CWindow, &BBank, BF_FITTXT);
                                         }
                                         break;
                              }
                              break;
          case WINH_EDITEND : ChangeOKButton = TRUE;
                              break;
        }
        if (ChangeOKButton)
        {
          if (AAddress[0] != '\0' && ASet[0] != '\0')
            BOk.Flags &= ~BF_GHOSTED;
          else
            BOk.Flags |= BF_GHOSTED;
          UpdateButtonText (&CWindow, &BOk, 0);
        }
        MouseStatus ();
        if (GlobalKey == K_ESC)
          Finished = TRUE;
      }
      if (Cancelled)
        Accepted = TRUE;
      else
        if (((Flags & FLAG_IGNBANK) || atoi (ABank) < 8) &&
            atol (AAddress) >= 16384 && atol (AAddress) < 65536 &&
            ((Flags & FLAG_ASK) || atol (ASet) < 256) &&
            ((Flags & FLAG_NOORG) || atol (AOrg) < 256))
          Accepted = TRUE;
      if (!Accepted)
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - One or more values are out of range");
    }
    DestroyWindow (&CWindow);
    GlobalKey = 0x0000;
    if (!Cancelled)
    {
      if ((NEntry = (struct PokeEntry_s *)malloc (sizeof (struct PokeEntry_s))) == NULL)
        FatalError (1, _ErrOutOfMemory, sizeof (struct PokeEntry_s));
      NEntry->Next = NULL;
      if (Flags & FLAG_IGNBANK)
        NEntry->Bank = 0;
      else
        NEntry->Bank = atoi (ABank);
      NEntry->Address = atoi (AAddress);
      if (Flags & FLAG_ASK)
        NEntry->Set = 0;
      else
        NEntry->Set = atoi (ASet);
      if (Flags & FLAG_NOORG)
        NEntry->Org = 0;
      else
        NEntry->Org = atoi (AOrg);
      NEntry->Flags = Flags;
      GetComboEntry (&PokesWindow, &NCEntry, SelectedPoke);
      OPokes = (struct GamePokes_s *)NCEntry.Identifier;
      if (OPokes->Pokes == NULL)                                                                                  /* First poke ? */
      {
        OPokes->Pokes = NEntry;
        NEntry->Previous = NULL;
      }
      else
      {
        GetComboEntry (&EntriesWindow, &NCEntry, EntriesWindow.NumEntries);
        OEntry = (struct PokeEntry_s *)NCEntry.Identifier;
        OEntry->Next = NEntry;
        NEntry->Previous = OEntry;
      }
      NCEntry.Identifier = (dword)NEntry;
      NCEntry.Selected = 0x00;
      if (!(NEntry->Flags & FLAG_IGNBANK))
        sprintf (ABank, "%u", NEntry->Bank);
      else
        strcpy (ABank, "-");
      if (!(NEntry->Flags & FLAG_ASK))
        sprintf (ASet, "%3u", NEntry->Set);
      else
        strcpy (ASet, "ASK");
      if (!(NEntry->Flags & FLAG_NOORG))
        sprintf (NCEntry.Text, "%s  %5u,%s (%3u)", ABank, NEntry->Address, ASet, NEntry->Org);
      else
        sprintf (NCEntry.Text, "%s  %5u,%s", ABank, NEntry->Address, ASet);
      if (!InsertComboEntry (&EntriesWindow, &NCEntry, NULL, -1))
        FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
      SelectedEntry = EntriesWindow.NumEntries;
      SelectEntry (&EntriesWindow, SelectedEntry);
      return (WINH_EDITEND);                                                             /* Signal: (un)ghost appropriate buttons */
    }
  }
  return (WINH_NONE);
}

static int _HandleDelEntry (struct WindowInfo *Window, struct ButtonInfo *Button)

/**********************************************************************************************************************************/
/* Pre   : The `Remove' button has been pressed (or released).                                                                    */
/* Post  : The current poke entry has been removed.                                                                               */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct GamePokes_s *OPokes;
  struct PokeEntry_s *OEntry;
  struct ComboEntry   NEntry;

  Window = Window; /* keep compiler happy */
  if (!(Button->Flags & BF_PRESSED) && EntriesWindow.NumEntries >= 0)
  {
    GetComboEntry (&EntriesWindow, &NEntry, SelectedEntry);
    if (TPConfig.ConfirmDelete)                                                               /* Confirmation needed: handle it ? */
      if (QuestionRequester (-1, -1, FALSE, FALSE, NULL, "@Cancel", NULL, NULL, NULL, "Confirm Delete",
                             "Are you sure you want to delete the poke\n\"%s\"?", NEntry.Text))
        return (WINH_NONE);
    OEntry = (struct PokeEntry_s *)NEntry.Identifier;
    GetComboEntry (&PokesWindow, &NEntry, SelectedPoke);
    OPokes = (struct GamePokes_s *)NEntry.Identifier;
    if (SelectedEntry == 0)                                                                               /* Removing first one ? */
    {
      OPokes->Pokes = OEntry->Next;
      OEntry->Next->Previous = NULL;
    }
    else
    {
      OEntry->Previous->Next = OEntry->Next;                                                                            /* Unlink */
      OEntry->Next->Previous = OEntry->Previous;
    }
    free (OEntry);
    DeleteComboEntry (&EntriesWindow, SelectedEntry);
    if (SelectedEntry > EntriesWindow.NumEntries)
      SelectedEntry --;
    SelectEntry (&EntriesWindow, SelectedEntry);
    return (WINH_EDITEND);                                                               /* Signal: (un)ghost appropriate buttons */
  }
  return (WINH_NONE);
}

static bool _HandleUpdateEntry (void)

/**********************************************************************************************************************************/
/* Pre   : A poke entry has been double-clicked.                                                                                  */
/* Post  : The current poke entry has been updated.                                                                               */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct WindowInfo       CWindow  = {190, 220, 160, 160, 160, 160, 160, 160, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                      WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "Update Poke"};
  struct ButtonInfo       BBank    = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 72, 40, 1, DBLACK, DWHITE, NULL,
                                      1, NULL, NULL};
  struct ButtonInfo       BNoBank  = {BT_PUSH, BC_GADGET, BF_FWIDTH | BF_KEYSELECT, 85, 40, 6, DBLACK, DWHITE, NULL,
                                      9, "\007", NULL};
  struct ButtonInfo       BAddress = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 72, 60, 5, DBLACK, DWHITE, NULL,
                                      2, NULL, NULL};
  struct ButtonInfo       BSet     = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 72, 80, 3, DBLACK, DWHITE, NULL,
                                      3, NULL, NULL};
  struct ButtonInfo       BAskSet  = {BT_PUSH, BC_GADGET, BF_FWIDTH | BF_KEYSELECT, 97, 80, 6, DBLACK, DWHITE, NULL,
                                      8, "\007", NULL};
  struct ButtonInfo       BOrg     = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 72, 100, 3, DBLACK, DWHITE, NULL,
                                      4, NULL, NULL};
  struct ButtonInfo       BNoOrg   = {BT_PUSH, BC_GADGET, BF_FWIDTH | BF_KEYSELECT, 97, 100, 6, DBLACK, DWHITE, NULL,
                                      5, "\007", NULL};
  struct ButtonInfo       BCancel  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 10, 135, 8, DBLACK, DWHITE, NULL,
                                      6, "@Cancel", NULL};
  struct ButtonInfo       BOk      = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 96, 135, 8, DBLACK, DWHITE, NULL,
                                      7, "@Ok", NULL};
  struct ButtonInfo      *ButtonP;
  struct PokeEntry_s     *OEntry;
  char                   *CEntry;
  struct ComboEntry       NEntry;
  char                    OldCEntry[30];
  char                    ABank[2];
  char                    AAddress[6];
  char                    ASet[4];
  char                    AOrg[4];
  bool                    Finished;
  bool                    Cancelled;
  bool                    Accepted = FALSE;
  bool                    ChangeOKButton;

  BBank.Contents = ABank;
  BAddress.Contents = AAddress;
  BSet.Contents = ASet;
  BOrg.Contents = AOrg;
  GetComboEntry (&EntriesWindow, &NEntry, SelectedEntry);
  OEntry = (struct PokeEntry_s *)NEntry.Identifier;
  CEntry = NEntry.Text;
  strcpy (OldCEntry, CEntry);
  AAddress[0] = ASet[0] = '\0';
  if (OEntry->Flags & FLAG_NOORG)
  {
    strcpy (AOrg, "???");
    BOrg.Flags |= BF_GHOSTED;
  }
  else
    sprintf (AOrg, "%d", OEntry->Org);
  if (OEntry->Flags & FLAG_IGNBANK)
  {
    strcpy (ABank, "-");
    BBank.Flags |= BF_GHOSTED;
  }
  else
    sprintf (ABank, "%d", OEntry->Bank);
  if (OEntry->Flags & FLAG_ASK)
  {
    strcpy (ASet, "ASK");
    BSet.Flags |= BF_GHOSTED;
  }
  else
    sprintf (ASet, "%d", OEntry->Set);
  sprintf (AAddress, "%u", OEntry->Address);
  AddButton (&CWindow, &BOk, TRUE);
  AddButton (&CWindow, &BBank, FALSE);
  AddButton (&CWindow, &BNoBank, FALSE);
  AddButton (&CWindow, &BAddress, FALSE);
  AddButton (&CWindow, &BSet, FALSE);
  AddButton (&CWindow, &BAskSet, FALSE);
  AddButton (&CWindow, &BOrg, FALSE);
  AddButton (&CWindow, &BNoOrg, FALSE);
  AddButton (&CWindow, &BCancel, FALSE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  Wprintf (&CWindow, 1, TEXT_CENTER, "Update poke parameters");
  Wprintf (&CWindow,  4, 2, "{+1Bank}N");
  Wprintf (&CWindow,  6, 2, "{+1Address}N");
  Wprintf (&CWindow,  8, 2, "{+1Value}N");
  Wprintf (&CWindow, 10, 2, "{+1Original}N");
  while (!Accepted)
  {
    Finished = FALSE;
    Cancelled = TRUE;
    while (!Finished)
    {
      ChangeOKButton = FALSE;
      switch (HandleWindow (&CWindow, &ButtonP))
      {
        case WINH_CLOSE   : Finished = TRUE; break;
        case WINH_RNOHAND : switch (ButtonP->ReactNum)
                            {
                              case 6 : Finished = TRUE; break;                                                        /* (Cancel) */
                              case 7 : Finished = TRUE; Cancelled = FALSE; break;                                         /* (Ok) */
                              case 5 : if (OEntry->Flags & FLAG_NOORG)                                               /* ORG cycle */
                                       {
                                         OEntry->Flags &= ~FLAG_NOORG;
                                         strcpy (AOrg, "0");
                                         BOrg.Flags &= ~BF_GHOSTED;
                                         UpdateButtonText (&CWindow, &BOrg, BF_FITTXT);
                                       }
                                       else
                                       {
                                         OEntry->Flags |= FLAG_NOORG;
                                         strcpy (AOrg, "???");
                                         BOrg.Flags |= BF_GHOSTED;
                                         UpdateButtonText (&CWindow, &BOrg, BF_FITTXT);
                                       }
                                       break;
                              case 8 : if (OEntry->Flags & FLAG_ASK)                                                 /* SET cycle */
                                       {
                                         OEntry->Flags &= ~FLAG_ASK;
                                         strcpy (ASet, "0");
                                         BSet.Flags &= ~BF_GHOSTED;
                                         UpdateButtonText (&CWindow, &BSet, BF_FITTXT);
                                       }
                                       else
                                       {
                                         OEntry->Flags |= FLAG_ASK;
                                         strcpy (ASet, "ASK");
                                         BSet.Flags |= BF_GHOSTED;
                                         UpdateButtonText (&CWindow, &BSet, BF_FITTXT);
                                       }
                                       ChangeOKButton = TRUE;
                                       break;
                              case 9 : if (OEntry->Flags & FLAG_IGNBANK)                                            /* BANK cycle */
                                       {
                                         OEntry->Flags &= ~FLAG_IGNBANK;
                                         strcpy (ABank, "0");
                                         BBank.Flags &= ~BF_GHOSTED;
                                         UpdateButtonText (&CWindow, &BBank, BF_FITTXT);
                                       }
                                       else
                                       {
                                         OEntry->Flags |= FLAG_IGNBANK;
                                         strcpy (ABank, "-");
                                         BBank.Flags |= BF_GHOSTED;
                                         UpdateButtonText (&CWindow, &BBank, BF_FITTXT);
                                       }
                                       break;
                            }
                            break;
        case WINH_EDITEND : ChangeOKButton = TRUE;
                            break;
      }
      if (ChangeOKButton)
      {
        if (AAddress[0] != '\0' && ASet[0] != '\0')
          BOk.Flags &= ~BF_GHOSTED;
        else
          BOk.Flags |= BF_GHOSTED;
        UpdateButtonText (&CWindow, &BOk, 0);
      }
      MouseStatus ();
      if (GlobalKey == K_ESC)
        Finished = TRUE;
    }
    if (Cancelled)
      Accepted = TRUE;
    else
      if (((OEntry->Flags & FLAG_IGNBANK) || atoi (ABank) < 8) &&
          atol (AAddress) >= 16384 && atol (AAddress) < 65536 &&
          ((OEntry->Flags & FLAG_ASK) || atol (ASet) < 256) &&
          ((OEntry->Flags & FLAG_NOORG) || atol (AOrg) < 256))
        Accepted = TRUE;
    if (!Accepted)
      ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - One or more values are out of range");
  }
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
  if (!Cancelled)
  {
    if (OEntry->Flags & FLAG_IGNBANK)
      OEntry->Bank = 0;
    else
      OEntry->Bank = atoi (ABank);
    OEntry->Address = atoi (AAddress);
    if (OEntry->Flags & FLAG_ASK)
      OEntry->Set = 0;
    else
      OEntry->Set = atoi (ASet);
    if (OEntry->Flags & FLAG_NOORG)
      OEntry->Org = 0;
    else
      OEntry->Org = atoi (AOrg);
    if (!(OEntry->Flags & FLAG_IGNBANK))
      sprintf (ABank, "%u", (OEntry->Bank & ~FLAG_ALL));
    else
      strcpy (ABank, "-");
    if (!(OEntry->Flags & FLAG_ASK))
      sprintf (ASet, "%3u", OEntry->Set);
    else
      strcpy (ASet, "ASK");
    if (!(OEntry->Flags & FLAG_NOORG))
      sprintf (NEntry.Text, "%s  %5u,%s (%3u)", ABank, OEntry->Address, ASet, OEntry->Org);
    else
      sprintf (NEntry.Text, "%s  %5u,%s", ABank, OEntry->Address, ASet);
    PutComboEntry (&EntriesWindow, &NEntry, SelectedEntry, TRUE);
    SelectEntry (&EntriesWindow, SelectedEntry);
    return ((bool)(strcmp (OldCEntry, CEntry) == 0) ? FALSE : TRUE);
  }
  return (FALSE);
}

static void _TBTimedFlash (struct WindowInfo *Window)

/**********************************************************************************************************************************/
/* Pre   : A spectrum screen has to be set up and drawn already.                                                                  */
/* Post  : This is the timed flash handler. Each run toggles the flashing parts.                                                  */
/* Import: PutSpectrumScreen.                                                                                                     */
/**********************************************************************************************************************************/

{
  Window = Window; /* keep compiler happy */
  FlashState = !FlashState;
  PutSpectrumScreen (WorkBuffer + _BlockOffset, _TBWindow.StartX + 6, _TBWindow.StartY + 127 + _AddHeight);
}

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> START OF THE EXPORTED POKE LIBRARY ROUTINES <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/**********************************************************************************************************************************/

bool SpawnApplication (char *Path, char *Options, char *File)

/**********************************************************************************************************************************/
/* Pre   : `Path' holds the path to the application, `Options' hold any extra command line options, `File' holds the file to be   */
/*         processed by the spawned application.                                                                                  */
/* Post  : The application has been run from its own directory.                                                                   */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  char           Params[300] = {"\0"};
  char           PathBuffer[MAXDIR];
  char           StartPath[MAXDIR];
  char           StartName[13];
  unsigned       Dummy;
  bool           SpawnReturn;
  short register Cnt;

  Cnt = strlen (Path);                                                                                 /* Split path and filename */
  while (-- Cnt >= 0 && Path[Cnt] != '\\' && Path[Cnt] != ':')
    ;
  if (Cnt >= 0)                                                                                   /* Is there a preceeding path ? */
  {
    if (Cnt > 0)
      strncpy (StartPath, Path, Cnt);
    StartPath[Cnt] = '\0';
    if (StartPath[1] == ':')                                                                            /* Preceeded by a drive ? */
    {
      _dos_setdrive (ToNumber (StartPath[0]), &Dummy);                                                      /* Change drive first */
      getcwd (PathBuffer, MAXDIR);
      if (StartPath[2] != '\0')                                                                                 /* Only a drive ? */
        chdir (StartPath + 2);                                                                     /* Change directory too if not */
      else if (Cnt == 2 && Path[2] == '\\')
        chdir ("\\");
    }
    else
    {
      getcwd (PathBuffer, MAXDIR);
      if (Cnt == 0 && Path[0] == '\\')
        chdir ("\\");
      else
        chdir (StartPath);                                                                          /* No drive, just change path */
    }
  }
  strcpy (StartName, Path + Cnt + 1);
  if (Options[0] != '\0')                                                              /* Do we have extra command line options ? */
  {
    strcat (Params, Options);
    strcat (Params, " ");
  }
  strcat (Params, File);                                                                                /* Add file to be handled */
  SpawnReturn = SwapOut (TRUE, StartName, Params, NULL, NULL, FALSE, FALSE, FALSE);                      /* Start the application */
  chdir (PathBuffer);
  _dos_setdrive (CurDrv, &Dummy);
  chdir (CurPath);                                                                                 /* Restore drive and directory */
  return (SpawnReturn);
}

bool EditGamePokes (dword OldSize, dword *NewSize)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : The pokes file for the currently selected games has been edited and written back.                                      */
/* Import: _ReadGamePokes, _WriteGamePokes, _SetupPokesWindow, _SetupEntriesWindow, _DestroyPokes.                                */
/**********************************************************************************************************************************/

{
  struct WindowInfo      CWindow = {120, 85, 382, 318, 382, 318, 382, 318, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                    WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL,
                                    "Edit Custom Info - POKEs block"};
  struct InputBoxInfo    BGenDes = {100, 30, 30, 7, DBLACK, DWHITE, NULL, 7, _GeneralDescription, BF_NOSURROUND | BF_KEYSELECT,
                                    255, 10};
  struct ButtonInfo      BAddP   = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 18, 251, 8, DBLACK, DWHITE, NULL,
                                    1, "@Add", _HandleAddPoke};
  struct ButtonInfo      BDelP   = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 78, 251, 8, DBLACK, DWHITE, NULL,
                                    2, "@Delete", _HandleDelPoke};
  struct ButtonInfo      BAddE   = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 246, 251, 8, DBLACK, DWHITE, NULL,
                                    3, "@Insert", _HandleAddEntry};
  struct ButtonInfo      BDelE   = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 306, 251, 8, DBLACK, DWHITE, NULL,
                                    4, "Re@move", _HandleDelEntry};
  struct ButtonInfo      BCancel = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 70, 290, 8, DBLACK, DWHITE, NULL,
                                    5, "@Cancel", NULL};
  struct ButtonInfo      BOk     = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 260, 290, 8, DBLACK, DWHITE, NULL,
                                    6, "@Save", NULL};
  struct ButtonInfo     *ButtonP;
  comboent               SelectedLine;
  bool                   Finished     = FALSE;
  bool                   Cancelled    = TRUE;
  bool                   PokesChanged = FALSE;
  bool                   PokesExisted = TRUE;

  if (!_ReadGamePokes (OldSize))
  { ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Out of memory while setting up"); return (FALSE); }
  AddButton (&CWindow, &BOk, TRUE);
  AddInputBox (&CWindow, &BGenDes, FALSE);
  AddButton (&CWindow, &BAddP, FALSE);
  AddButton (&CWindow, &BDelP, FALSE);
  AddButton (&CWindow, &BAddE, FALSE);
  AddButton (&CWindow, &BDelE, FALSE);
  AddButton (&CWindow, &BCancel, FALSE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  WDrawShadowBox (&CWindow, 1, 0, 11, 62, "General Description");
  WDrawShadowBox (&CWindow, 12, 0, 27, 36, "Trainers");
  WDrawShadowBox (&CWindow, 12, 38, 27, 62, "Pokes");
  MakeComboWindow (&PokesWindow, &CWindow, COMT_SELECT, COMF_NOBUFFER | COMF_VSLIDER | COMF_KEYSELECT | COMF_SYNCSLIDER,
                   CWindow.StartX + 16, CWindow.StartY + 140, 15, 64, MAXPOKNAME + 5, MAXPOKNAME, 10, NULL, FALSE,
                   0, NULL, DBLACK, DWHITE, 0, 0);
  MakeComboWindow (&EntriesWindow, &CWindow, COMT_SELECT, COMF_NOBUFFER | COMF_VSLIDER | COMF_SYNCSLIDER,
                   CWindow.StartX + 244, CWindow.StartY + 140, 15, 64, 19, 18, 10, NULL, FALSE,
                   0, NULL, DBLACK, DWHITE, 0, 0);
  SelectedPoke = 0;
  SelectedEntry = 0;
  _SetupPokesWindow ();
  _SetupEntriesWindow (SelectedPoke);
  SelectEntry (&PokesWindow, SelectedPoke);
  SelectEntry (&EntriesWindow, SelectedEntry);
  if (PokesWindow.NumEntries == -1)                                                                          /* No pokes at all ? */
  {
    BDelP.Flags |= BF_GHOSTED;
    BAddE.Flags |= BF_GHOSTED;
    BOk.Flags |= BF_GHOSTED;
    UpdateButtonText (&CWindow, &BDelP, 0);
    UpdateButtonText (&CWindow, &BAddE, 0);
    UpdateButtonText (&CWindow, &BOk, 0);
  }
  if (EntriesWindow.NumEntries == -1)
  {
    BDelE.Flags |= BF_GHOSTED;
    UpdateButtonText (&CWindow, &BDelE, 0);
  }
  while (!Finished)
  {
    switch (HandleWindow (&CWindow, &ButtonP))
    {
      case WINH_CLOSE     : Finished = TRUE; break;
      case WINH_RNOHAND   : switch (ButtonP->ReactNum)
                            {
                              case 5 : Finished = TRUE; break;                                                        /* (Cancel) */
                              case 6 : Finished = TRUE; Cancelled = FALSE; break;                                         /* (Ok) */
                            }
                            break;
      case WINH_EDITEND   : PokesChanged = TRUE;                                                            /* (Add/Del handlers) */
                            if (PokesWindow.NumEntries == -1)                                                /* No pokes at all ? */
                            {
                              BDelP.Flags |= BF_GHOSTED;
                              BAddE.Flags |= BF_GHOSTED;
                              if (!PokesExisted)                                            /* Allow full delete if not inserting */
                                BOk.Flags |= BF_GHOSTED;
                            }
                            else
                            {
                              BDelP.Flags &= ~BF_GHOSTED;
                              BAddE.Flags &= ~BF_GHOSTED;
                              BOk.Flags &= ~BF_GHOSTED;
                            }
                            if (EntriesWindow.NumEntries == -1)
                              BDelE.Flags |= BF_GHOSTED;
                            else
                              BDelE.Flags &= ~BF_GHOSTED;
                            UpdateButtonText (&CWindow, &BDelP, 0);
                            UpdateButtonText (&CWindow, &BAddE, 0);
                            UpdateButtonText (&CWindow, &BOk, 0);
                            UpdateButtonText (&CWindow, &BDelE, 0);
                            break;
    }
    switch (HandleComboWindow (&PokesWindow, &SelectedLine))
    {
      case WINH_LINEFIRST : SelectedPoke = SelectedLine;
                            SelectedEntry = 0;
                            _SetupEntriesWindow (SelectedPoke);
                            SelectEntry (&EntriesWindow, SelectedEntry);
                            break;
      case WINH_LINESEL   : if (_HandleUpdatePoke ())
                              PokesChanged = TRUE;
                            break;
    }
    if (HandleComboWindow (&EntriesWindow, &SelectedEntry) == WINH_LINESEL)
      if (_HandleUpdateEntry ())
        PokesChanged = TRUE;
    MouseStatus ();
    if (GlobalKey == K_ESC)
      Finished = TRUE;
  }
  if (!Cancelled && PokesChanged)
    _WriteGamePokes (NewSize);
  DestroyComboWindow (&PokesWindow, 0);
  DestroyComboWindow (&EntriesWindow, 0);
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
  _DestroyPokes ();
  return (!Cancelled && PokesChanged);
}

bool EditSpectrumScreen (short WhichWindow, struct TapeEntry_s *Entry, dword OldSize, dword *NewSize)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points the the tape entry, `OldSize' is the size of this block,     */
/*         not counting the Custom Info header.                                                                                   */
/* Post  : The tape entry has been displayed. The user had the possibility to change any part of it.                              */
/*         The return value is TRUE if any change has been made and `NewSize' then holds the new size for this block.             */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  static byte IBorderColors[8] = { 0, 1, 4, 5, 2, 3, 6, 7 };
  struct InputBoxInfo BGenDes = {85, 30, 30, 7, DBLACK, DWHITE, NULL, 1, _GeneralDescription, BF_NOSURROUND | BF_KEYSELECT,
                                 255, 10};
  struct ButtonInfo   BFile   = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 274, 128, 10, DBLACK, DWHITE, NULL, 2, "@File",
                                 NULL};
  struct ButtonInfo   BMkGIF  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 274, 168, 10, DBLACK, DWHITE, NULL, 3, "Make @GIF",
                                 NULL};
  struct ButtonInfo   BMkSCR  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 274, 188, 10, DBLACK, DWHITE, NULL, 4, "Make @SCR",
                                 NULL};
  struct ButtonInfo   BBCol   = {BT_PUSH, BC_GADGET, BF_FWIDTH | BF_KEYSELECT, 325, 230, 1, DBLACK, DWHITE, NULL, 7, "\0", NULL};
  struct ButtonInfo   BCancel = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 75, 335, 8, DBLACK, DWHITE, NULL,
                                 5, "@Cancel", NULL};
  struct ButtonInfo   BOk     = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 215, 335, 8, DBLACK, DWHITE, NULL,
                                 6, "@Ok", NULL};
  struct ComboInfo    PWindow;
  struct ComboEntry   PEntry;
  struct ButtonInfo  *ButtonP;
  char               *FullName;
  short     register  Cnt;
  comboent            SelectedLine;
  bool                Done        = FALSE;
  bool                Cancelled   = FALSE;
  bool                Changed     = TRUE;
  byte                BorderColor = 7;                                                                        /* Default to white */
  char                FullPath[MAXDIR + 1];
  int                 Fh;
  word                NewOffset;

  _AddHeight = (OldSize == 0 ? 0 : 30);
  if (OldSize == 0)                                                                                        /* Inserting a block ? */
  {
    Done = FALSE;
    strcpy (_Mask, "*.SCR");
    while (!Done)
    {
      if (!FileRequester (180, 50, 48, 29, 20, _Path, _File, _Mask, "Select Screen File", TRUE, FALSE))
        return (FALSE);
      MakeFullPath (FullPath, _Path, _File);
      if ((Fh = open (FullPath, O_BINARY | O_RDONLY)) == -1)
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Cannot open %s\n      - %s", _File, StrError (errno));
      else if (filelength (Fh) != 6912)
      {
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - %s is not a Spectrum screen file!", _File);
        close (Fh);
      }
      else
      {
        strcpy (_GeneralDescription, "Loading Screen");
        _BlockOffset = 0;
        if (read (Fh, WorkBuffer, 6912) != 6912)
        {
          ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Cannot read %s\n      - %s", _File, StrError (errno));
          memset (WorkBuffer, 0, 6144);                                                            /* Clear 'display memory' area */
          memset (WorkBuffer + 6144, 0x38, 768);                                                       /* Clear 'attributes' area */
        }
        else
          Changed = TRUE;
        close (Fh);
        Done = TRUE;                                                                                  /* Accept the provided file */
      }
    }
    Done = FALSE;
  }
  else                                                                                                               /* (Editing) */
  {
    if (*WorkBuffer == 0)                                                                               /* No description given ? */
      strcpy (_GeneralDescription, "Loading Screen");
    else
    {
      strncpy (_GeneralDescription, WorkBuffer + 1, *WorkBuffer);                               /* Read General Description field */
      _GeneralDescription[*WorkBuffer] = '\0';
    }
    BorderColor = *(WorkBuffer + (*WorkBuffer) + 1);
    _BlockOffset = (word)(*WorkBuffer) + 2;
  }
  BBCol.BPen = IBorderColors[BorderColor];
  BGenDes.StartY += _AddHeight;
  BFile.StartY += _AddHeight;
  BMkGIF.StartY += _AddHeight;
  BMkSCR.StartY += _AddHeight;
  BBCol.StartY += _AddHeight;
  BCancel.StartY += _AddHeight;
  BOk.StartY += _AddHeight;
  _TBWindow.Height = _TBWindow.MinHeight = _TBWindow.MaxHeight = 358 + _AddHeight;
  _TBWindow.StartY = 65 - (_AddHeight / 2);
  AddButton (&_TBWindow, &BOk, TRUE);
  AddInputBox (&_TBWindow, &BGenDes, FALSE);
  AddButton (&_TBWindow, &BFile, FALSE);
  AddButton (&_TBWindow, &BMkGIF, FALSE);
  AddButton (&_TBWindow, &BMkSCR, FALSE);
  AddButton (&_TBWindow, &BBCol, FALSE);
  AddButton (&_TBWindow, &BCancel, FALSE);
  DrawWindow (&_TBWindow, NULL, WING_CLOSE, TRUE);
  if (OldSize != 0)
  {
    GetBlockTypeInfo (Entry->BlockTypeID, NULL, NULL, &FullName, NULL, NULL);
    Wprintf (&_TBWindow, 1, 2, " {+1Block type       }N %s (%02X)", FullName, Entry->BlockTypeID);
    Wprintf (&_TBWindow, 2, 2, " {+1Block length     }N %lu", Entry->BlockLength);
  }
  WDrawShadowBox (&_TBWindow, (_AddHeight / 10) + 1, 0, (_AddHeight / 10) + 11, 56, "Description");
  Wprintf (&_TBWindow, (_AddHeight / 10) + 23, 46, "{+1Border}N");
  SurroundScreen (&_TBWindow, 4, 125 + _AddHeight);
  PutSpectrumScreen (WorkBuffer + _BlockOffset, _TBWindow.StartX + 6, _TBWindow.StartY + 127 + _AddHeight);
  ReleaseTimedHandler (_TBTimedFlash, &_TBWindow);
  if (ScreenContainsFlash && TPConfig.EmulateFlash)
    RegisterTimedHandler (TPConfig.SlowFlash ? 960 : 320, _TBTimedFlash, &_TBWindow);
  while (!Done)
  {
    switch (HandleWindow (&_TBWindow, &ButtonP))
    {
      case WINH_EDITEND : if (ButtonP->Flags & BF_FITINPUT)                                        /* Contents actually changed ? */
                            Changed = TRUE;                                                         /* Edited General Description */
                          break;
      case WINH_RNOHAND : switch (ButtonP->ReactNum)
                          {
                            case 2 : Done = FALSE;                                                                 /* File button */
                                     while (!Done)
                                     {
                                       if (!FileRequester (180, 50, 48, 29, 20, _Path, _File, _Mask, "Select Screen File", TRUE,
                                           FALSE))
                                         Done = TRUE;
                                       else
                                       {
                                         MakeFullPath (FullPath, _Path, _File);
                                         if ((Fh = open (FullPath, O_BINARY | O_RDONLY)) == -1)
                                           ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Cannot open %s\n      - %s",
                                                           _File, StrError (errno));
                                         else if (filelength (Fh) != 6912)
                                         {
                                           ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - %s is not a Spectrum screen file!",
                                                           _File);
                                           close (Fh);
                                         }
                                         else
                                         {
                                           ReleaseTimedHandler (_TBTimedFlash, &_TBWindow);
                                           if (read (Fh, WorkBuffer + _BlockOffset, 6912) != 6912)
                                           {
                                             ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Cannot read %s\n      - %s",
                                                             _File, StrError (errno));
                                             memset (WorkBuffer + _BlockOffset, 0, 6144);          /* Clear 'display memory' area */
                                             memset (WorkBuffer + _BlockOffset + 6144, 0x38, 768);     /* Clear 'attributes' area */
                                           }
                                           PutSpectrumScreen (WorkBuffer + _BlockOffset, _TBWindow.StartX + 6,
                                                              _TBWindow.StartY + 127 + _AddHeight);
                                           if (ScreenContainsFlash && TPConfig.EmulateFlash)
                                             RegisterTimedHandler (TPConfig.SlowFlash ? 960 : 320, _TBTimedFlash, &_TBWindow);
                                           close (Fh);
                                           Done = TRUE;                                               /* Accept the provided file */
                                           Changed = TRUE;
                                         }
                                       }
                                     }
                                     Done = FALSE;
                                     break;
                            case 3 : WriteGIFFile (WhichWindow, _TBWindow.StartX + 6, _TBWindow.StartY + 127 + _AddHeight);
                                     break;                                                                         /* GIF button */
                            case 4 : WriteSCRFile (WhichWindow, WorkBuffer + _BlockOffset);                         /* SCR button */
                                     break;
                            case 5 : Cancelled = Done = TRUE; break;                                                    /* Cancel */
                            case 6 : Done = TRUE; break;                                                                    /* Ok */
                            case 7 : MakeComboWindow (&PWindow, &_TBWindow, COMT_SELECT, COMF_KEYSELECT | COMF_EXCLUSIVE,
                                                      _TBWindow.StartX + BBCol.StartX, _TBWindow.StartY + BBCol.StartY + 16,
                                                      0, 1024, 8, 1, 8, NULL, FALSE, 0, NULL, DBLACK, DWHITE, 0, 0);
                                     for (Cnt = 0 ; Cnt < 8 ; Cnt ++)
                                     {
                                       PEntry.Selected = 0;
                                       PEntry.Identifier = Cnt;
                                       sprintf (PEntry.Text, "{+%d\xDB}N", IBorderColors[Cnt]);
                                       if (!InsertComboEntry (&PWindow, &PEntry, NULL, -1))
                                         FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                                     }
                                     PrintComboEntries (&PWindow);
                                     Done = FALSE;
                                     while (!Done)
                                     {
                                       MouseStatus ();
                                       if (HandleComboWindow (&PWindow, &SelectedLine) == WINH_LINEFIRST)
                                       {
                                         Done = TRUE;
                                         BorderColor = PWindow.LastSelected;
                                         Changed = TRUE;
                                       }
                                       else
                                       {
                                         HandleWindow (&(PWindow.ComboWindow), &ButtonP);             /* (Just read the keyboard) */
                                         if (Mouse.Xx < PWindow.OverlayWindow.A_StartX || Mouse.Xx > PWindow.OverlayWindow.A_EndX ||
                                             Mouse.Yy < PWindow.OverlayWindow.A_StartY || Mouse.Yy > PWindow.OverlayWindow.A_EndY)
                                           if (Mouse.Left)                                 /* Left click outside the select box ? */
                                           {
                                             Done = TRUE;
                                             DelayLeftMouseButton ();
                                           }
                                       }
                                       if (GlobalKey == K_ESC)
                                         Done = TRUE;
                                     }
                                     GlobalKey = 0x0000;
                                     DestroyComboWindow (&PWindow, 0);
                                     BBCol.BPen = IBorderColors[BorderColor];
                                     DrawButton (&_TBWindow, &BBCol);
                                     Done = FALSE;
                                     break;
                          }
                          break;
      case WINH_CLOSE   : Cancelled = Done = TRUE;
    }
    if (GlobalKey == K_ESC)                                                                                    /* <ESC> stops too */
      Cancelled = Done = TRUE;
    GlobalKey = 0x0000;
    MouseStatus ();
  }
  DestroyWindow (&_TBWindow);
  GlobalKey = 0x0000;
  if (Cancelled || !Changed)
    return (FALSE);
  if (!strcmp (_GeneralDescription, "Loading Screen"))                                                          /* Default text ? */
    _GeneralDescription[0] = '\0';
  NewOffset = strlen (_GeneralDescription) + 1;                                     /* Point to the location for the Border Color */
  if (OldSize == 0)                                                                                            /* Inserting now ? */
    memmove (WorkBuffer + NewOffset + 1, WorkBuffer, 6912);                                         /* Move the screen into place */
  else
    memmove (WorkBuffer + NewOffset + 1, WorkBuffer + (*WorkBuffer) + 2, 6912);
  *WorkBuffer = strlen (_GeneralDescription);                                              /* Write length of General Description */
  if (*WorkBuffer != 0)                                                                                              /* Nothing ? */
    strncpy (WorkBuffer + 1, _GeneralDescription, *WorkBuffer);                             /* Add the General Description itself */
  *(WorkBuffer + (*WorkBuffer) + 1) = BorderColor;                                                /* Write the border color value */
  *NewSize = 6914 + *WorkBuffer;
  return (TRUE);                                                                 /* Signal: write everything (back) into the tape */
}

bool ViewInstructionsBlock (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points the the tape entry.                                          */
/* Post  : This is an alternate handler for the 'View Tape Block Information' routine, as the width of the window has to be       */
/*         much larger (80 - the maximum width of a text file). The return value is TRUE if the user edited the block.            */
/* Import: EditInstructionsBlock.                                                                                                 */
/**********************************************************************************************************************************/

{
#define MAXVWLINES   44
  struct ButtonInfo   BOk   = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 295, 250, 8, DBLACK, DWHITE, NULL,
                               1, "@Ok", NULL};
  struct ButtonInfo   BEdit = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 163, 250, 8, DBLACK, DWHITE, NULL,
                               2, "@Edit", NULL};
  struct MouseShape          *OldShape;
  char                       *FullName;
  struct ButtonInfo          *ButtonP;
  struct WindowInfo           ViewWindow;
  struct ComboInfo            TextWindow;
  struct ComboEntry           NEntry;
  comboent                    Dummy;
  short             register  Index        = 0;
  char                        Bt;
  char                        PrevBt       = 0;
  bool                        More         = FALSE;
  bool                        Done         = FALSE;
  bool                        Cancelled    = FALSE;
  bool                        Changed      = FALSE;
  bool                        Resizing     = FALSE;
  bool                        Reload       = TRUE;
  bool                        FirstRun     = TRUE;
  char                        Line[1024];
  byte                        CFlags = COMF_NOBUFFER | COMF_VSLIDER | COMF_SYNCSLIDER | COMF_KEYSELECT;
  short                       CurrentWidth;
  short                       CurrentHeight;
  dword                       XMSIndex;
  dword                       LengthToGo;
  dword                       ReadLength;
  dword                       DReadLength;
  word                        BufferIndex;

  SelectFont (NULL, FONT_NORMAL);
  memset (&ViewWindow, 0, sizeof (struct WindowInfo));                                           /* Initialize the interface part */
  CurrentWidth           = 80;
  CurrentHeight          = 20;
  ViewWindow.StartX      = 65;
  ViewWindow.StartY      = 40;
  ViewWindow.Width       = ViewWindow.MinWidth = ViewWindow.MaxWidth = CurrentFont->Width * CurrentWidth + 30;
  ViewWindow.Height      = ViewWindow.MinHeight = (CurrentFont->Height + 2) * CurrentHeight + 80;
  ViewWindow.MaxHeight   = (CurrentFont->Height + 2) * MAXVWLINES;
  ViewWindow.BorderFlags = WINB_OUT | WINB_SURROUND;
  ViewWindow.BorderSize  = 1;
  ViewWindow.APen        = SystemColors.ReqAPen;
  ViewWindow.BPen        = SystemColors.ReqOPen;
  ViewWindow.Header      = "Tape Block Information";
  ViewWindow.Flags       = WINF_ACTIVATE | WINF_HUSE | WINF_HCENTER | WINF_XMSBUFFER;
  BEdit.StartY = BOk.StartY = (CurrentFont->Height + 2) * CurrentHeight + 55;
  AddButton (&ViewWindow, &BOk, TRUE);
  AddButton (&ViewWindow, &BEdit, FALSE);
  DrawWindow (&ViewWindow, NULL, WING_CLOSE | WING_RESIZEMAX, TRUE);
  GetBlockTypeInfo (Entry->BlockTypeID, NULL, NULL, &FullName, NULL, NULL);
  Wprintf (&ViewWindow, 1, 2, " {+1Block type       }N %s (%02X)", FullName, Entry->BlockTypeID);
  Wprintf (&ViewWindow, 2, 2, " {+1Block length     }N %lu", Entry->BlockLength);
  More = FALSE;
  OldShape = CurrentMouseShape;
  while (Reload)
  {
    Reload = FALSE;
    if (FirstRun)
      MakeComboWindow (&TextWindow, &ViewWindow, COMT_VIEW, CFlags, ViewWindow.StartX + 8, ViewWindow.StartY + 40, 0, 2048, 100,
                       CurrentWidth, CurrentHeight, CurrentFont, FALSE, 0, NULL, SystemColors.ReqAPen, SystemColors.ReqBPen, 0, 0);
    else
      ReInitComboWindow (&TextWindow, 0);
    FirstRun = FALSE;
    SetMouseShape (DefaultWaitingMouseShape);
    XMSIndex = GetTapeBlockXMSIndex (WhichWindow, Entry);
    LengthToGo = Entry->BlockLength;
    ReadLength = (LengthToGo > 49152 ? 49152 : LengthToGo);
    DReadLength = (ReadLength & 0xFFFFFFFC) + (ReadLength % 4 ? 4 : 0);
    CopyFromXMS (DReadLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
    LengthToGo -= ReadLength;
    XMSIndex += ReadLength;
    BufferIndex = 16;                                                                          /* Skip the Custom Info Identifier */
    Index = 0;
    More = TRUE;
    Done = FALSE;
    while (More && !Done)                                                             /* Read the entire file into the entrytable */
    {
      if (BufferIndex < ReadLength - 1)
        Bt = *(WorkBuffer + (BufferIndex ++));
      else if (LengthToGo == 0)
        Done = TRUE;
      else
      {
        ReadLength = (LengthToGo > 49152 ? 49152 : LengthToGo);
        DReadLength = (ReadLength & 0xFFFFFFFC) + (ReadLength % 4 ? 4 : 0);
        CopyFromXMS (DReadLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
        LengthToGo -= ReadLength;
        XMSIndex += ReadLength;
        Bt = *WorkBuffer;
        BufferIndex = 0;
      }
      if (More && !Done)
      {
        if (Bt != 0x0A && Bt != 0x0D && Index < 1023)
        {
          if (Bt == '{' || Bt == '}')
            Line[Index ++] = Bt;
          Line[Index ++] = Bt;
        }
        else if ((Bt == 0x0A && PrevBt != 0x0D) ||                                                                  /* CR or LF ? */
                 (Bt == 0x0D && PrevBt != 0x0A) ||                                      /* Eliminate the combination of CR and LF */
                 (PrevBt != 0x0A && PrevBt != 0x0D))
        {
          Line[Index] = '\0';
          More = FitComboString (&TextWindow, NULL, Line, 80, 0, 0, FALSE);
          Index = 0;
        }
        if (Bt == 0x0A && PrevBt == 0x0D)
          PrevBt = 0;
        else
          PrevBt = Bt;
      }
    }
    if (More && Index != 0)                                                                       /* Resulting line without eol ? */
    {
      Line[Index] = '\0';
      More = FitComboString (&TextWindow, NULL, Line, 80, 0, 0, FALSE);
    }
    if (!More)                                                                                     /* File truncated prematurally */
    {
      if (TextWindow.NumEntries == -1)                                                                       /* No entry at all ? */
        FatalError (1, _ErrOutOfMemory, 0);                                                                          /* Then quit */
      GetComboEntry (&TextWindow, &NEntry, TextWindow.NumEntries);
      strcpy (NEntry.Text, "{+1>>> TRUNCATED <<<}N");
      PutComboEntry (&TextWindow, &NEntry, TextWindow.NumEntries, FALSE);
      More = TRUE;
    }
    SetMouseShape (OldShape);
    PrintComboEntries (&TextWindow);                                                                        /* Print the contents */
    while (More)                                                                                  /* Handle until close requested */
    {
      if (Resizing)
      {
        CurrentWidth = (short)((ViewWindow.Width - 30) / CurrentFont->Width);
        CurrentHeight = (short)((ViewWindow.Height - 80) / (CurrentFont->Height + 2));
        ViewWindow.Width  = CurrentFont->Width * CurrentWidth + 30;
        ViewWindow.Height = (CurrentFont->Height + 2) * CurrentHeight + 80;
        BEdit.StartY = BOk.StartY = (CurrentFont->Height + 2) * CurrentHeight + 55;
        AddButton (&ViewWindow, &BOk, TRUE);
        AddButton (&ViewWindow, &BEdit, FALSE);
        DrawWindow (&ViewWindow, NULL, WING_CLOSE | WING_RESIZEMAX, TRUE);
        GetBlockTypeInfo (Entry->BlockTypeID, NULL, NULL, &FullName, NULL, NULL);
        Wprintf (&ViewWindow, 1, 2, " {+1Block type       }N %s (%02X)", FullName, Entry->BlockTypeID);
        Wprintf (&ViewWindow, 2, 2, " {+1Block length     }N %lu", Entry->BlockLength);
        MakeComboWindow (&TextWindow, &ViewWindow, COMT_VIEW, CFlags, ViewWindow.StartX + 8, ViewWindow.StartY + 40, 0, 2048, 100,
                        CurrentWidth, CurrentHeight, CurrentFont, FALSE, 0, NULL, SystemColors.ReqAPen, SystemColors.ReqBPen, 0, 0);
        PrintComboEntries (&TextWindow);
        Resizing = FALSE;
      }
      HandleComboWindow (&TextWindow, &Dummy);
      if (GlobalKey == K_ESC)
      { More = FALSE; Cancelled = TRUE; }
      else
      {
        switch (HandleWindow (&ViewWindow, &ButtonP))
        {
          case WINH_CLOSE     : More = FALSE; Cancelled = TRUE; break;
          case WINH_WINRESIZE : Resizing = TRUE;
                                DestroyComboWindow (&TextWindow, COMF_RESIZING);
                                DestroyWindow (&ViewWindow);
                                CFlags |= COMF_RESIZING;
                                break;
          case WINH_RNOHAND   : switch (ButtonP->ReactNum)
                                {
                                  case 1 : More = FALSE; break;                                                           /* 'Ok' */
                                  case 2 : Changed = EditInstructionsBlock (WhichWindow, Entry,                         /* 'Edit' */
                                                                            Entry->BlockLength - 16, &LengthToGo);
                                           if (Changed)
                                           { More = FALSE; Reload = TRUE; }                                  /* Reload if changed */
                                           break;
                                }
        }
        if (More && GlobalKey == K_ESC)
        { More = FALSE; Cancelled = TRUE; }
      }
      MouseStatus ();
    }
  }
  DestroyComboWindow (&TextWindow, 0);
  DestroyWindow (&ViewWindow);
  GlobalKey = 0x0000;
  return (!Cancelled && Changed);
}

bool EditInstructionsBlock (short WhichWindow, struct TapeEntry_s *Entry, dword OldSize, dword *NewSize)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points the the tape entry, `OldSize' is 0 if the entry has just     */
/*         been inserted, otherwise holds the length of the text part.                                                            */
/*         This handler is used for both plain ASCII 'Instructions' and 'ZX-Edit Document' blocks.                                */
/* Post  : The user has edited this Custom Info block. If the block was now being inserted, he has been presented a file          */
/*         requester to import a .TXT file. If that was escaped, an empty editor has been started so the user can supply his own  */
/*         text. If it was an existing piece of text, this text has been written out in a temporary text file, the editor has     */
/*         started and if the file had been changed, it is re-inserted into the XMS block.                                        */
/*         If anything has been changed (either now inserted or edited), the tape is adjusted already when leaving and TRUE is    */
/*         returned. Any error is reported here, in which case FALSE is returned. If the text block has not been changed, FALSE   */
/*         is returned as well. If it should be created now and no text has actually been inserted, the tape block has been       */
/*         removed already.                                                                                                       */
/* Import: SpawnApplication, _GetGeneralDescription.                                                                              */
/**********************************************************************************************************************************/

{
  struct stat    StatBefore;
  struct stat    StatAfter;
  short register Index     = 0;
  char           Line[1026];
  bool           More      = TRUE;
  bool           Done      = FALSE;
  bool           WriteIt;
  bool           FileIsTmp = FALSE;
  bool           Continue;
  dword          XMSIndex;
  dword          LengthToGo;
  dword          ReadLength;
  dword          DReadLength;
  dword          Skipped   = 0;
  dword          StartOffset;
  word           BufferIndex;
  dword          DBufferIndex;
  byte           Bt;
  byte           FirstBt;
  bool           FoundEOL;
  FILE          *Fp;
  unsigned       FileAttributes;
  char          *UsedEditor;
  char           NoOptions = '\0';

  if (Entry->FlagByte == INSRTCI_ZXEDIT)                                                                    /* ZX-Edit Document ? */
    UsedEditor = TPConfig.ZXEditPath;
  else                                                                                     /* No, use normal ASCII editor instead */
    UsedEditor = TPConfig.EditPath;
  if (OldSize == 0)                                                                                        /* Inserting a block ? */
  {
    Done = FALSE;
    if (Entry->FlagByte == INSRTCI_ZXEDIT)
      strcpy (_Mask, "*.ZED");
    else
      strcpy (_Mask, "*.TXT");
    while (!Done)
    {
      if (!FileRequester (180, 50, 48, 29, 20, _Path, _File, _Mask, "Select Text File", TRUE, FALSE))           /* Ask for a file */
      {                                                           /* User doesn't want to insert a file, but rather something new */
        Continue = TRUE;
        if (UsedEditor[0] == '\0')                                                                         /* No editor defined ? */
        {
          Continue = FALSE;
          if (Entry->FlagByte == INSRTCI_ZXEDIT)
          {
            if (!QuestionRequester (-1, -1, FALSE, FALSE, NULL, "@Cancel", NULL, NULL, NULL, "ZX-Edit Wizard",
                                    "The ZX-Edit program path has not been set in your configuration.\n"
                                    "Do you want to select it now?"))
              Continue = _GetNewFileName ("Select ZX-Edit Path", TPConfig.ZXEditPath, NULL, NULL);
          }
          else
            ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - No editor has been defined");
          if (!Continue)
          {
            DeleteTapeBlock (WhichWindow, Entry, FALSE, FALSE);
            return (FALSE);
          }
        }
        if (Continue)
        {
          MakeFullPath (_TempFileName, TmpDirectory, mktemp (NULL));
          strcpy (_TempFileName + strlen (_TempFileName) - 3, Entry->FlagByte == INSRTCI_ZXEDIT ? "ZED" : "TXT");
          FileIsTmp = TRUE;
          SpawnApplication (UsedEditor, &NoOptions, _TempFileName);                                           /* Start the editor */
          if (SwapReturn != SWAP_OK)                                                                    /* Something went wrong ? */
          { unlink (_TempFileName);                                                                          /* Cleanup and leave */
            DeleteTapeBlock (WhichWindow, Entry, FALSE, FALSE); return (FALSE); }
          if ((Fp = fopen (_TempFileName, "rb")) == NULL)                                             /* Open the typed text file */
          {
            if (errno != ENOENT)                                              /* User simply aborted his editor without writing ? */
              ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Cannot open %s\n      - %s", _TempFileName, StrError (errno));
            DeleteTapeBlock (WhichWindow, Entry, FALSE, FALSE);
            return (FALSE);
          }
          Done = TRUE;
        }
      }
      else
      {
        MakeFullPath (_TempFileName, _Path, _File);
        FileIsTmp = FALSE;
        if ((Fp = fopen (_TempFileName, "rb")) == NULL)
        {
          ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Cannot open %s\n      - %s", _File, StrError (errno));
          return (FALSE);
        }
        Done = TRUE;
      }
    }
  }
  else                                                                                        /* (Editing an existing text block) */
  {
    Continue = TRUE;
    if (UsedEditor[0] == '\0')                                                                             /* No editor defined ? */
    {
      Continue = FALSE;
      if (Entry->FlagByte == INSRTCI_ZXEDIT)
      {
        if (!QuestionRequester (-1, -1, FALSE, FALSE, NULL, "@Cancel", NULL, NULL, NULL, "ZX-Edit Wizard",
                                "The ZX-Edit program path has not been set in your configuration.\n"
                                "Do you want to select it now?"))
          Continue = _GetNewFileName ("Select ZX-Edit Path", TPConfig.ZXEditPath, NULL, NULL);
      }
      else
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - No editor has been defined");
      if (!Continue)
        return (FALSE);
    }
    if (Continue)
    {
      MakeFullPath (_TempFileName, TmpDirectory, mktemp (NULL));                       /* Write the block out to a temporary file */
      strcpy (_TempFileName + strlen (_TempFileName) - 3, Entry->FlagByte == INSRTCI_ZXEDIT ? "ZED" : "TXT");
      FileIsTmp = TRUE;
      if ((Fp = fopen (_TempFileName, "wb")) == NULL)                                                /* Create the temp text file */
      {
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Cannot create %s\n      - %s", _TempFileName, StrError (errno));
        return (FALSE);
      }
      /* Write the text data from XMS out into the temporary file */
      XMSIndex = GetTapeBlockXMSIndex (WhichWindow, Entry);
      LengthToGo = Entry->BlockLength;
      ReadLength = (LengthToGo > 49152 ? 49152 : LengthToGo);
      DReadLength = (ReadLength & 0xFFFFFFFC) + (ReadLength % 4 ? 4 : 0);
      CopyFromXMS (DReadLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
      LengthToGo -= ReadLength;
      XMSIndex += ReadLength;
      if (Entry->FlagByte == INSRTCI_ZXEDIT)
        BufferIndex = 17 + *(WorkBuffer + 16);                                                 /* Skip the Custom Info Identifier */
      else
        BufferIndex = 16;
      Index = 0;
      More = TRUE;                                                                                 /* Signal: no write errors yet */
      Done = FALSE;
      while (More && !Done)                                                           /* Read the entire file into the entrytable */
      {
        if (Entry->FlagByte == INSRTCI_ZXEDIT)                                                              /* ZX-Edit Document ? */
        {
          if (fwrite (WorkBuffer, 1, (ReadLength - BufferIndex), Fp) != (ReadLength - BufferIndex))             /* Write the line */
            More = FALSE;
          if (LengthToGo == 0)
            Done = TRUE;
          else
          {
            ReadLength = (LengthToGo > 49152 ? 49152 : LengthToGo);
            DReadLength = (ReadLength & 0xFFFFFFFC) + (ReadLength % 4 ? 4 : 0);
            CopyFromXMS (DReadLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
            LengthToGo -= ReadLength;
            XMSIndex += ReadLength;
            BufferIndex = 0;
          }
        }
        else
        {                                                                                     /* Nope, need to check the contents */
          if (BufferIndex < ReadLength - 1)
            Bt = *(WorkBuffer + (BufferIndex ++));
          else if (LengthToGo == 0)
            Done = TRUE;
          else
          {
            ReadLength = (LengthToGo > 49152 ? 49152 : LengthToGo);
            DReadLength = (ReadLength & 0xFFFFFFFC) + (ReadLength % 4 ? 4 : 0);
            CopyFromXMS (DReadLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
            LengthToGo -= ReadLength;
            XMSIndex += ReadLength;
            Bt = *WorkBuffer;
            BufferIndex = 0;
          }
          if (More && !Done)
          {
            if (Bt != 0x0A && Bt != 0x0D && Index < 1023)
              Line[Index ++] = Bt;
            else if (Bt != 0x0A)                                                                       /* Eliminate CR characters */
            {
              if (Bt == 0x0D)                                                         /* Expand CR characters to CR-LF (PC style) */
              {
                Line[Index ++] = 0x0D;
                Line[Index ++] = 0x0A;
                Line[Index] = '\0';
              }
              else
                Line[Index] = '\0';
              if (fwrite (Line, 1, Index, Fp) != Index)                                                         /* Write the line */
                More = FALSE;
              Index = 0;
            }
          }
        }
      }
      if (Entry->FlagByte != INSRTCI_ZXEDIT)
        if (More && Index != 0)                                                                   /* Resulting line without eol ? */
        {
          Line[Index ++] = 0x0D;                                                          /* DOS likes a CRLF after the last line */
          Line[Index ++] = 0x0A;
          Line[Index] = '\0';
          if (fwrite (Line, 1, Index, Fp) != Index)                                                             /* Write the line */
            More = FALSE;
        }
      fclose (Fp);
      if (!More)                                                                                            /* An error occured ? */
      {
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Error writing %s\n      - %s", _TempFileName, StrError (errno));
        unlink (_TempFileName);                                                                              /* Cleanup and leave */
        return (FALSE);
      }
      _dos_setfileattr (_TempFileName, _A_NORMAL);                                                       /* Reset the archive bit */
      if (stat (_TempFileName, &StatBefore))                                                           /* Get current file status */
      {
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Cannot get status of %s\n      - %s", _TempFileName, StrError (errno));
        unlink (_TempFileName);                                                                              /* Cleanup and leave */
        return (FALSE);
      }
      SpawnApplication (UsedEditor, &NoOptions, _TempFileName);                                               /* Start the editor */
      if (SwapReturn != SWAP_OK)                                                                        /* Something went wrong ? */
      { unlink (_TempFileName); return (FALSE); }                                                            /* Cleanup and leave */
      More = FALSE;
      if (!_dos_getfileattr (_TempFileName, &FileAttributes))                                            /* Get current file info */
        if (!stat (_TempFileName, &StatAfter))
          if ((FileAttributes & _A_ARCH) ||                                                         /* Archive bit has been set ? */
              StatBefore.st_ctime != StatAfter.st_ctime ||                           /* File modification date has been changed ? */
              StatBefore.st_size != StatAfter.st_size)                                                /* File size is different ? */
            More = TRUE;                                                                             /* Yes, file has been edited */
      if (!More)                                                                                       /* Step out if not changed */
      { unlink (_TempFileName); return (FALSE); }
      if ((Fp = fopen (_TempFileName, "rb")) == NULL)                                              /* Re-open the typed text file */
      {
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Cannot open %s\n      - %s", _TempFileName, StrError (errno));
        return (FALSE);
      }
    }
  }
  /* We have an open file now, that must be read (back) into the tape */
  fseek (Fp, 0L, SEEK_END);
  *NewSize = ftell (Fp);
  fseek (Fp, 0L, SEEK_SET);
  if (Entry->FlagByte == INSRTCI_ZXEDIT)                                                                    /* ZX-Edit Document ? */
    StartOffset = _GetGeneralDescription ("ZX-Edit Document", "Instructions") + 17;
  else
    StartOffset = 16;
  XMSIndex = ResizeTapeBlock (WhichWindow, Entry, *NewSize + StartOffset, 0);                  /* Resize the data block if needed */
  if (Entry->FlagByte == INSRTCI_ZXEDIT)                                                                    /* ZX-Edit Document ? */
  {
    strcpy (WorkBuffer, "ZX-Edit Document");                                                                  /* Build identifier */
    *(WorkBuffer + 16) = StartOffset - 17;                                                      /* Add General Description if any */
    strcpy (WorkBuffer + 17, _GeneralDescription);
    LengthToGo = *NewSize;
    DBufferIndex = 49152 - StartOffset;                                                              /* Set max for the first run */
    while (LengthToGo > 0)
    {
      ReadLength = (LengthToGo > DBufferIndex ? DBufferIndex : LengthToGo);
      if (fread (WorkBuffer + StartOffset, 1, ReadLength, Fp) != ReadLength)                           /* Read each chunk in turn */
      {
        fclose (Fp);
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Error reading %s\n      - %s", _TempFileName, StrError (errno));
        if (FileIsTmp)
          unlink (_TempFileName);
        return (TRUE);
      }
      DReadLength = ((ReadLength + StartOffset) & 0xFFFFFFFC) + ((ReadLength + StartOffset) % 4 ? 4 : 0);
      CopyToXMS (DReadLength, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, XMSIndex);
      LengthToGo -= ReadLength;
      DBufferIndex = 49152;
      StartOffset = 0;
    }
  }
  else
  {
    CopyToXMS (16, TapeWindow[WhichWindow].Tape.XMSHandle, "Instructions    ", XMSIndex);
    XMSIndex += StartOffset;
    LengthToGo = *NewSize;
    BufferIndex = 0;
    Done = FALSE;
    FoundEOL = FALSE;
    while (LengthToGo > 0)
    {
      if (fread (&Bt, 1, 1, Fp) != 1)                                                                   /* Read each byte in turn */
      {
        fclose (Fp);
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Error reading %s\n      - %s", _TempFileName, StrError (errno));
        if (FileIsTmp)
          unlink (_TempFileName);
        return (TRUE);
      }
      LengthToGo --;
      WriteIt = TRUE;
      if (Entry->FlagByte != INSRTCI_ZXEDIT)                                                                /* ZX-Edit Document ? */
      {                                                                                       /* Nope, need to check the contents */
        if (Bt == 0x0A || Bt == 0x0D)
          if (!FoundEOL || Bt == FirstBt)                                                         /* Multiple returns are allowed */
          {
            FoundEOL = TRUE;
            FirstBt = Bt;
            Bt = 0x0D;
          }
          else
          {
            WriteIt = FALSE;
            Skipped ++;
          }
        else
          FoundEOL = FALSE;
      }
      if (WriteIt)
      {
        if (BufferIndex < 49152)
          *(WorkBuffer + (BufferIndex ++)) = Bt;
        else
        {
          CopyToXMS (49152, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, XMSIndex);
          XMSIndex += 49152;
          BufferIndex = 1;
          *WorkBuffer = Bt;
        }
      }
    }
    if (BufferIndex > 0)                                                                     /* Partly buffer at the end (likely) */
    {
      DBufferIndex = (BufferIndex & 0xFFFFFFFC) + (BufferIndex % 4 ? 4 : 0);
      CopyToXMS (DBufferIndex, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, XMSIndex);
    }
    if (Skipped > 0)                                                                   /* Did we skip some CR-LF sequence bytes ? */
      Entry->BlockLength -= Skipped;                                                                  /* Then disaccount for them */
  }
  fclose (Fp);
  if (FileIsTmp)
    unlink (_TempFileName);
  return (TRUE);
}

bool EditPictureBlock (short WhichWindow, struct TapeEntry_s *Entry, dword OldSize, dword *NewSize)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points the the tape entry, `OldSize' is 0 if the entry has just     */
/*         been inserted, otherwise holds the length of the picture part.                                                         */
/* Post  : The user has viewed/inserted a Picture Block. If the block was now being inserted, he has been presented a file        */
/*         requester to import a .GIF or .JPG file. If that was escaped, the entire routine is aborted.                           */
/*         If anything has been inserted, the tape is adjusted already when leaving and TRUE is returned. Any error is reported   */
/*         here, in which case FALSE is returned. If the picture block has not been changed, FALSE is returned as well. If it     */
/*         should be created now and no picture has actually been inserted, the tape block has been  removed already.             */
/* Import: SpawnApplication, _GetGeneralDescription.                                                                              */
/**********************************************************************************************************************************/

{
  bool           More      = TRUE;
  bool           Done      = FALSE;
  bool           Continue;
  byte           PicType;
  dword          XMSIndex;
  dword          LengthToGo;
  dword          ReadLength;
  dword          DReadLength;
  dword          StartOffset;
  word           BufferIndex;
  dword          DBufferIndex;
  int            Fh;
  char          *UsedViewer;
  char          *UsedViewerOptions;

  if (OldSize == 0)                                                                                        /* Inserting a block ? */
  {
    Done = FALSE;
    strcpy (_Mask, "*.*");                                                                   /* (We need either a .GIF or a .JPG) */
    while (!Done)
    {
      if (!FileRequester (180, 50, 48, 29, 20, _Path, _File, _Mask, "Select Picture File", TRUE, FALSE))        /* Ask for a file */
      {                                                           /* User doesn't want to insert a file, but rather something new */
        DeleteTapeBlock (WhichWindow, Entry, FALSE, FALSE);
        return (FALSE);
      }
      else
      {
        MakeFullPath (_TempFileName, _Path, _File);
        PicType = 0xFF;
        if (!strcmp (_File + strlen (_File) - 4, ".GIF"))
          PicType = 0x00;
        else if (!strcmp (_File + strlen (_File) - 4, ".JPG"))
          PicType = 0x01;
        if (PicType == 0xFF)
          ErrorRequester (-1, -1, NULL, NULL, NULL, "Please select a .GIF or a .JPG file");
        else if ((Fh = open (_TempFileName, O_BINARY | O_RDONLY)) == -1)
          ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Cannot open %s\n      - %s", _File, StrError (errno));
        else
          Done = TRUE;
      }
    }
    /* We have an open file now, that must be read into the tape */
    *NewSize = filelength (Fh);
    StartOffset = _GetGeneralDescription ("Picture", "Inlay Card") + 18;
    XMSIndex = ResizeTapeBlock (WhichWindow, Entry, *NewSize + StartOffset, 0);                /* Resize the data block if needed */
    strcpy (WorkBuffer, "Picture         ");                                                                  /* Build identifier */
    *(WorkBuffer + 16) = PicType;                                                                             /* Set picture type */
    *(WorkBuffer + 17) = StartOffset - 18;                                                      /* Add General Description if any */
    strcpy (WorkBuffer + 18, _GeneralDescription);
    LengthToGo = *NewSize;
    DBufferIndex = 49152 - StartOffset;                                                              /* Set max for the first run */
    while (LengthToGo > 0)
    {
      ReadLength = (LengthToGo > DBufferIndex ? DBufferIndex : LengthToGo);
      if (read (Fh, WorkBuffer + StartOffset, ReadLength) < 0)                                         /* Read each chunk in turn */
      {
        close (Fh);
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Error reading %s\n      - %s", _TempFileName, StrError (errno));
        DeleteTapeBlock (WhichWindow, Entry, FALSE, FALSE);
        return (FALSE);
      }
      DReadLength = ((ReadLength + StartOffset) & 0xFFFFFFFC) + ((ReadLength + StartOffset) % 4 ? 4 : 0);
      CopyToXMS (DReadLength, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, XMSIndex);
      XMSIndex += DReadLength;
      LengthToGo -= ReadLength;
      DBufferIndex = 49152;
      StartOffset = 0;
    }
    close (Fh);
    return (TRUE);
  }
  else                                                                                     /* (Viewing an existing picture block) */
  {
    /* Write the picture data from XMS out into the temporary file */
    XMSIndex = GetTapeBlockXMSIndex (WhichWindow, Entry);
    LengthToGo = Entry->BlockLength;
    ReadLength = (LengthToGo > 49152 ? 49152 : LengthToGo);
    DReadLength = (ReadLength & 0xFFFFFFFC) + (ReadLength % 4 ? 4 : 0);
    CopyFromXMS (DReadLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
    LengthToGo -= ReadLength;
    XMSIndex += DReadLength;
    PicType = *(WorkBuffer + 16);
    if (PicType != 0x00 && PicType != 0x01)
    { ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Unknown picture type (0x%02X)", PicType); return (FALSE); }
    UsedViewer = (PicType == 0x00 ? TPConfig.GIFViewerPath : TPConfig.JPGViewerPath);
    UsedViewerOptions = (PicType == 0x00 ? TPConfig.GIFViewerOptions : TPConfig.JPGViewerOptions);
    Continue = TRUE;
    if (UsedViewer[0] == '\0')                                                                             /* No viewer defined ? */
    {
      Continue = FALSE;
      if (!QuestionRequester (-1, -1, FALSE, FALSE, NULL, "@Cancel", NULL, NULL, NULL, "Picture Viewer Wizard",
                              "No %s viewer program path has not been set in your configuration.\n"
                              "Do you want to select it now?", PicType == 0x00 ? "GIF" : "JPG"))
        Continue = _GetNewFileName (PicType == 0x00 ? "Select GIF Viewer Path" : "Select JPG Viewer Path", UsedViewer,
                                    UsedViewerOptions, "view a file");
      if (!Continue)
        return (FALSE);
    }
    MakeFullPath (_TempFileName, TmpDirectory, mktemp (NULL));
    strcpy (_TempFileName + strlen (_TempFileName) - 3, PicType == 0x00 ? "GIF" : "JPG");
    if ((Fh = open (_TempFileName, O_BINARY | O_RDWR | O_CREAT, 0777)) == -1)                        /* Create the temporary file */
    {
      ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Cannot create %s\n      - %s", _TempFileName, StrError (errno));
      return (FALSE);
    }
    BufferIndex = 18 + *(WorkBuffer + 17);                                                     /* Skip the Custom Info Identifier */
    More = TRUE;                                                                                   /* Signal: no write errors yet */
    Done = FALSE;
    while (More && !Done)
    {
      ReadLength -= BufferIndex;
      if (write (Fh, WorkBuffer + BufferIndex, ReadLength) != ReadLength)                                      /* Write the chunk */
        More = FALSE;
      if (LengthToGo == 0)
        Done = TRUE;
      else
      {
        ReadLength = (LengthToGo > 49152 ? 49152 : LengthToGo);
        DReadLength = (ReadLength & 0xFFFFFFFC) + (ReadLength % 4 ? 4 : 0);
        CopyFromXMS (DReadLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
        LengthToGo -= ReadLength;
        XMSIndex += DReadLength;
        BufferIndex = 0;
      }
    }
    close (Fh);
    if (!More)                                                                                              /* An error occured ? */
      ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Error writing %s\n      - %s", _TempFileName, StrError (errno));
    else
      SpawnApplication (UsedViewer, UsedViewerOptions, _TempFileName);                                        /* Start the editor */
    unlink (_TempFileName);                                                                                  /* Cleanup and leave */
    return (FALSE);
  }
}
