/**********************************************************************************************************************************/
/* Project      : TAPER                                                                                                           */
/* Module       : TPEDIT.C                                                                                                        */
/* Description  : Edit/insert of tape entries                                                                                     */
/* Version type : Program module                                                                                                  */
/* Last changed : 25-06-1999  7:00                                                                                                */
/* Update count : 39                                                                                                              */
/* OS type      : PC (DOS)                                                                                                        */
/* Compiler     : Watcom C32 Optimizing Compiler Version 10.x                                                                     */
/* Linker       : Watcom C Linker                                                                                                 */
/*                                                                                                                                */
/*                Copyleft (C) 1997-2001 ThunderWare Research Center, written by Martijn van der Heide.                           */
/*                                                                                                                                */
/*                This program is free software; you can redistribute it and/or                                                   */
/*                modify it under the terms of the GNU General Public License                                                     */
/*                as published by the Free Software Foundation; either version 2                                                  */
/*                of the License, or (at your option) any later version.                                                          */
/*                                                                                                                                */
/*                This program is distributed in the hope that it will be useful,                                                 */
/*                but WITHOUT ANY WARRANTY; without even the implied warranty of                                                  */
/*                MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                                                   */
/*                GNU General Public License for more details.                                                                    */
/*                                                                                                                                */
/*                You should have received a copy of the GNU General Public License                                               */
/*                along with this program; if not, write to the Free Software                                                     */
/*                Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                                     */
/*                                                                                                                                */
/**********************************************************************************************************************************/

#define __TPEDIT_MAIN__

#include "wbench32/wbench32.h"
#include <dos.h>
#include <fcntl.h>
#include <io.h>
#include <bios.h>
#include <sys/types.h>
#include "tpconfig.h"
#include "tpscreen.h"
#include "tpaudio.h"
#include "tpfile.h"
#include "tptape.h"
#include "tpmenu.h"
#include "tpmain.h"
#include "tpcust.h"
#include "tpbasic.h"
#include "tpedit.h"

#pragma pack                ()

/**********************************************************************************************************************************/
/* Define the static variables                                                                                                    */
/**********************************************************************************************************************************/

static char     _Path[MAXDIR + 1] = "\0";
static char     _File[13]         = "\0";
static char     _Mask[13]         = "*.*";

static comboent _EntryNumber      = -1;
static bool     _Cancelled        = FALSE;

static byte     _InsertCustomType = INSRTCI_NONE;

/**********************************************************************************************************************************/
/* Define the static routines                                                                                                     */
/**********************************************************************************************************************************/

static bool  _EditSelectValue      (char *SelectString, comboent *Block, comboent MaxBlock, comboent WallyBlock);
static bool  _InsertRawDataFile    (short WhichWindow, struct TapeEntry_s *Entry);
static bool  _InsertHeaderBlock    (short WhichWindow, struct TapeEntry_s *Entry);
static bool  _InsertBASICBlock     (short WhichWindow, struct TapeEntry_s *Entry);
static bool  _InsertCIPOKEs        (short WhichWindow, struct TapeEntry_s *Entry);
static bool  _InsertCIInstructions (short WhichWindow, struct TapeEntry_s *Entry);
static bool  _InsertCIScreen       (short WhichWindow, struct TapeEntry_s *Entry);
static bool  _InsertCIZXEdit       (short WhichWindow, struct TapeEntry_s *Entry);
static bool  _InsertCIPicture      (short WhichWindow, struct TapeEntry_s *Entry);

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> START OF THE INTERNAL LIBRARY ROUTINES <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/**********************************************************************************************************************************/

static bool _EditSelectValue (char *SelectString, comboent *Block, comboent MaxBlock, comboent WallyBlock)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : The entry has been adjusted by the user and saved in the tape directory.                                               */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct WindowInfo CWindow = {189, 220, 262, 90, 262, 90, 262, 90, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                               WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "Edit Menu Selection"};
  struct ButtonInfo PCancel = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 63, 60, 8, DBLACK, DWHITE, NULL, 1, "@Cancel", NULL};
  struct ButtonInfo POk     = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 141, 60, 8, DBLACK, DWHITE, NULL, 0, "@Ok", NULL};
  struct ButtonInfo BString = {BT_INPUT, BC_TEXT, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 70, 10, 30, DBLACK, DWHITE, NULL, 2,
                               NULL, NULL};
  struct ButtonInfo BBlock  = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 70, 30, 3, DBLACK, DWHITE, NULL, 3,
                               NULL, NULL};
  struct ButtonInfo *ButtonP;
  char              ValBlock[4];
  bool              Done      = FALSE;
  bool              Cancelled = FALSE;
  bool              Changed   = FALSE;
  bool              Accepted  = FALSE;

  BString.Contents = SelectString;
  BBlock.Contents = ValBlock;
  sprintf (ValBlock, "%d", *Block);
  AddButton (&CWindow, &POk, TRUE);
  AddButton (&CWindow, &BString, FALSE);
  AddButton (&CWindow, &BBlock, FALSE);
  AddButton (&CWindow, &PCancel, FALSE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  Wprintf (&CWindow, 1, 1, "{+1Selection}N");
  Wprintf (&CWindow, 3, 1, "{+1Jumps to}N");
  while (!Accepted)
  {
    while (!Done)
    {
      switch (HandleWindow (&CWindow, &ButtonP))
      {
        case WINH_CLOSE   : Done = TRUE; Cancelled = TRUE; break;
        case WINH_RNOHAND : switch (ButtonP->ReactNum)
                            {
                              case 0  : Done = TRUE; Cancelled = FALSE; break;                                            /* (Ok) */
                              case 1  : Done = TRUE; Cancelled = TRUE; break;                                         /* (Cancel) */
                            }
                            break;
        case WINH_EDITEND : if (ButtonP->Flags & BF_FITINPUT)                                          /* Contents were changed ? */
                              Changed = TRUE;
                            break;
      }
      if (GlobalKey == K_ESC)  { Done = TRUE; Cancelled = TRUE; }
      MouseStatus ();
    }
    if (Cancelled)
      Accepted = TRUE;
    else
    {
      if (ValBlock[0] != '\0' && atoi (ValBlock) == WallyBlock)                                                 /* Wrong choise ? */
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Don't be a wally!");
      else if (ValBlock[0] != '\0' && atoi (ValBlock) <= MaxBlock)
      {
        *Block = atoi (ValBlock);
        Accepted = TRUE;
      }
      else
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - The value is out of range\nUse values 0 through %d", MaxBlock);
      if (!Accepted)
        Done = FALSE;
    }
  }
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
  return (!Cancelled && Changed);
}

static bool _InsertRawDataFile (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : The selected raw file has been inserted. The routine ends with the timing editor.                                      */
/* Import: ResizeTapeBlock, EditTapeDataBlock.                                                                                    */
/**********************************************************************************************************************************/

{
  int           Fh;
  char          FullPath[MAXDIR + 1];
  long          FileLength;
  dword         XMSIndex;
  dword         XMSSubIndex;
  dword         LengthToGo;
  dword         DLengthToGo;
  word          ReadLength;
  word register Cnt;
  bool          Error = FALSE;

  Entry->FlagByte = 0xFF;                                                                             /* Assume normal data block */
  Entry->Flags.FlagByteUsed = TRUE;
  Entry->Flags.ParityByteUsed = TRUE;
  Entry->CalculatedParity = Entry->FlagByte;                                                                   /* Init the parity */
  Entry->Timing.LengthPilotPulse = NS_LENPILOT;                                                           /* Make it normal speed */
  Entry->Timing.NumPilotPulses = NS_DTPILOT;
  Entry->Timing.LengthSyncPulse[0] = NS_LENSYNC1;
  Entry->Timing.LengthSyncPulse[1] = NS_LENSYNC2;
  Entry->Timing.LengthZeroBitPulse = NS_LEN0BIT;
  Entry->Timing.LengthOneBitPulse = NS_LEN1BIT;
  Entry->Timing.NumBitsInFirstByte = 8;
  Entry->Timing.NumBitsInLastByte = 8;
  Entry->Flags.IsNormalSpeed = TRUE;
  MakeFullPath (FullPath, _Path, _File);
  if ((Fh = open (FullPath, O_BINARY | O_RDONLY)) == -1)
    return (TRUE);
  FileLength = filelength (Fh);
  if (FileLength <= 65535L)                                                                /* Could fit in a Normal Speed block ? */
    Entry->BlockTypeID = 0x10;                                                                                 /* Then convert it */
  XMSIndex = ResizeTapeBlock (WhichWindow, Entry, FileLength, 0);                                       /* Make room for the file */
  XMSSubIndex = 0;                                                                                      /* Start at the beginning */
  LengthToGo = Entry->BlockLength;
  while (LengthToGo > 0 && !Error)                                                                     /* Read the data in chunks */
  {
    ReadLength = (word)(LengthToGo > (dword)49152 ? 49152 : LengthToGo);
    if (read (Fh, WorkBuffer, ReadLength) < ReadLength)                                                        /* Premature EOF ? */
    {
      ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Unexpected end of file");
      Error = TRUE;
    }
    for (Cnt = 0 ; Cnt < ReadLength ; Cnt ++)                                                                 /* Calculate parity */
      Entry->CalculatedParity ^= *(WorkBuffer + Cnt);
    DLengthToGo = (dword)(ReadLength & 0xFFFC) + (ReadLength % 4 ? 4 : 0);
    CopyToXMS (DLengthToGo, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, XMSIndex + XMSSubIndex);
    XMSSubIndex += ReadLength;
    LengthToGo -= ReadLength;
  }
  close (Fh);
  Entry->ParityByte = Entry->CalculatedParity;                                                       /* Make the parity 'correct' */
  return (EditTapeDataBlock (WhichWindow, Entry));                                                           /* Adjust all timers */
}

static bool _InsertHeaderBlock (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : An empty header has been inserted. The routine ends with the header editor.                                            */
/* Import: ResizeTapeBlock, EditHeaderBlock.                                                                                      */
/**********************************************************************************************************************************/

{
  dword         XMSIndex;
  word register Cnt;

  Entry->BlockTypeID = 0x10;
  Entry->FlagByte = 0x00;                                                                           /* Assume normal header block */
  Entry->Flags.FlagByteUsed = TRUE;
  Entry->Flags.ParityByteUsed = TRUE;
  Entry->Timing.LengthPilotPulse = NS_LENPILOT;                                                           /* Make it normal speed */
  Entry->Timing.NumPilotPulses = NS_DTPILOT;
  Entry->Timing.LengthSyncPulse[0] = NS_LENSYNC1;
  Entry->Timing.LengthSyncPulse[1] = NS_LENSYNC2;
  Entry->Timing.LengthZeroBitPulse = NS_LEN0BIT;
  Entry->Timing.LengthOneBitPulse = NS_LEN1BIT;
  Entry->Timing.NumBitsInFirstByte = 8;
  Entry->Timing.NumBitsInLastByte = 8;
  Entry->Flags.IsNormalSpeed = TRUE;
  GetBlockTypeInfo (Entry->BlockTypeID, &(Entry->IsTapeData), &(Entry->IsIdentifiedData), NULL, NULL, NULL);
  *WorkBuffer = 0x03;                                                                               /* Default is a 'bytes' block */
  memset (WorkBuffer + 1, ' ', 10);                                               /* Filename is padded with spaces on a Spectrum */
  memset (WorkBuffer + 11, 0x00, 5);
  *(WorkBuffer + 16) = 0x80;
  Entry->CalculatedParity = Entry->FlagByte;                                                                   /* Init the parity */
  for (Cnt = 0 ; Cnt < 17 ; Cnt ++)
    Entry->CalculatedParity ^= *(WorkBuffer + Cnt);                                                       /* Calculate the parity */
  Entry->ParityByte = Entry->CalculatedParity;                                                       /* Make the parity 'correct' */
  XMSIndex = ResizeTapeBlock (WhichWindow, Entry, 17, 0);                                        /* Make room for the header data */
  CopyToXMS (20, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, XMSIndex);                         /* Insert the header data */
  return (EditHeaderBlock (WhichWindow, Entry));                                                              /* Adjust all items */
}

static bool _InsertBASICBlock (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : The user has inserted a BASIC block.                                                                                   */
/* Import: none.                                                                                                                  */
/**********************************************************************************************************************************/

{
  word DummyBlockLength;
  bool Cancelled;

  Cancelled = !EditBASICBlock (WhichWindow, Entry, NULL, &DummyBlockLength, 0, TRUE);                     /* Go create the block! */
  if (Cancelled)
  {
    if (_EntryNumber != -1)                                                                     /* Came from the insert handler ? */
      DeleteTapeBlock (WhichWindow, Entry, FALSE, FALSE);                         /* Then we don't have to keep the entry anymore */
    _Cancelled = TRUE;
    return (FALSE);
  }
  return (TRUE);
}

static bool _InsertCIPOKEs (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : A flag is set and the Custom Info editor is called.                                                                    */
/* Import: EditCustomInfoBlock.                                                                                                   */
/**********************************************************************************************************************************/

{
  Entry->BlockTypeID = 0x35;
  Entry->FlagByte = _InsertCustomType = INSRTCI_POKES;
  return (EditCustomInfoBlock (WhichWindow, Entry));                                                              /* Go insert it */
}

static bool _InsertCIInstructions (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : A flag is set and the Special editor is called.                                                                        */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  dword NewSize;

  Entry->BlockTypeID = 0x35;
  Entry->FlagByte = _InsertCustomType = INSRTCI_INSTRUCT;
  return (EditInstructionsBlock (WhichWindow, Entry, 0, &NewSize));                                               /* Go insert it */
}

static bool _InsertCIScreen (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : A flag is set and the Custom Info editor is called.                                                                    */
/* Import: EditCustomInfoBlock.                                                                                                   */
/**********************************************************************************************************************************/

{
  Entry->BlockTypeID = 0x35;
  Entry->FlagByte = _InsertCustomType = INSRTCI_SCREEN;
  return (EditCustomInfoBlock (WhichWindow, Entry));                                                              /* Go insert it */
}

static bool _InsertCIZXEdit (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : A flag is set and the Special editor is called.                                                                        */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  dword NewSize;

  Entry->BlockTypeID = 0x35;
  Entry->FlagByte = _InsertCustomType = INSRTCI_ZXEDIT;
  return (EditInstructionsBlock (WhichWindow, Entry, 0, &NewSize));                                               /* Go insert it */
}

static bool _InsertCIPicture (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : A flag is set and the Custom Info editor is called.                                                                    */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  dword NewSize;

  Entry->BlockTypeID = 0x35;
  Entry->FlagByte = _InsertCustomType = INSRTCI_PICTURE;
  return (EditPictureBlock (WhichWindow, Entry, 0, &NewSize));                                                    /* Go insert it */
}

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> START OF THE EXPORTED LIBRARY ROUTINES <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/**********************************************************************************************************************************/

dword ResizeTapeBlock (short WhichWindow, struct TapeEntry_s *Entry, dword NewSize, dword AddedSize)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be considered, `NewSize'      */
/*         `NewSize' holds the new BlockSize value for this Entry.                                                                */
/* Post  : The XMS data block belonging to this entry has been resized if necessary. The XMSIndex has been returned.              */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  dword NewLength;
  dword NextXMSIndex;
  dword SizeDiff;
  dword XMSIndex;

  Entry->BlockLength = NewSize;
  NewLength = ((NewSize + AddedSize) & 0xFFFFFFFC) + ((NewSize + AddedSize) % 4 ? 4 : 0);          /* Determine XMS needed length */
  if (TapeWindow[WhichWindow].Tape.XMSHandle == -1)                                                               /* Empty tape ? */
  {
    Entry->AllocLength = NewLength;
    XMSAlloc (&(TapeWindow[WhichWindow].Tape.XMSHandle), NewLength);                                      /* Create the tape body */
    TapeWindow[WhichWindow].Tape.Size = NewLength;
    TapeWindow[WhichWindow].EmptyTape = FALSE;
    return (0);
  }
  XMSIndex = GetTapeBlockXMSIndex (WhichWindow, Entry);
  NextXMSIndex = XMSIndex + Entry->AllocLength;
  if (NewLength == Entry->AllocLength)                                                                   /* No resize necessary ? */
    return (XMSIndex);
  if (NewLength < Entry->AllocLength)                                                       /* New string shorter than old string */
  {
    SizeDiff = Entry->AllocLength - NewLength;
    if (NextXMSIndex != TapeWindow[WhichWindow].Tape.Size)
      CopyInXMS (TapeWindow[WhichWindow].Tape.Size - NextXMSIndex, TapeWindow[WhichWindow].Tape.XMSHandle,
                 NextXMSIndex, TapeWindow[WhichWindow].Tape.XMSHandle, NextXMSIndex - SizeDiff);      /* Move the next entry back */
    XMSRealloc (TapeWindow[WhichWindow].Tape.XMSHandle, TapeWindow[WhichWindow].Tape.Size - SizeDiff, FALSE);
    TapeWindow[WhichWindow].Tape.Size -= SizeDiff;                                                        /* Adjust the tape size */
  }
  else                                                                                     /* (New string longer than old string) */
  {
    SizeDiff = NewLength - Entry->AllocLength;
    XMSRealloc (TapeWindow[WhichWindow].Tape.XMSHandle, TapeWindow[WhichWindow].Tape.Size + SizeDiff, FALSE);
    if (NextXMSIndex != TapeWindow[WhichWindow].Tape.Size)
      CopyInXMS (TapeWindow[WhichWindow].Tape.Size - NextXMSIndex, TapeWindow[WhichWindow].Tape.XMSHandle,
                 NextXMSIndex, TapeWindow[WhichWindow].Tape.XMSHandle, NextXMSIndex + SizeDiff);   /* Move the room to this entry */
    TapeWindow[WhichWindow].Tape.Size += SizeDiff;                                                        /* Adjust the tape size */
  }
  Entry->AllocLength = NewLength;
  return (XMSIndex);
}

void LoadHardwareTable (void)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : The file 'HARDWARE.INC' has been loaded into its own XMS space.                                                        */
/*         Mo return is made if any error occured.                                                                                */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  int   Fh;
  dword FileLength;
  char  FullName[MAXDIR + 1];

  _makepath (FullName, ProgramDrive, ProgramPath, "HARDWARE", "INC");
  if ((Fh = open (FullName, O_RDONLY)) < 0)
    FatalError (7, "Could not load HARDWARE.INC: %s", StrError (errno));
  FileLength = filelength (Fh);
  if (read (Fh, WorkBuffer, (word)FileLength) != FileLength)
    FatalError (6, "Error loading HARDWARE.INC");
  HWTableLength = (FileLength & 0xFFFFFFFC) + (FileLength % 4 ? 4 : 0);
  XMSAlloc (&HWTableXMSHandle, HWTableLength);
  CopyToXMS (HWTableLength, HWTableXMSHandle, WorkBuffer, 0);
  close (Fh);
}

void GetHardwareTable (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry of the `Hardware Type' block.    */
/* Post  : The data of the `Hardware Type' block has been loaded and prepared into the WorkBuffer.                                */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  dword        *DHWTypes;
  byte         *BHWTypes;
  word          NumTypes;
  word register Cnt;
  word register Cnt2;
  dword         XMSIndex;
  dword         Swap;
  byte          Bt1;
  byte          Bt2;
  byte          Bt3;
  bool          More = TRUE;

  BHWTypes = GlobalFileCopyBuffer;                                                          /* Use this as an intermediate buffer */
  DHWTypes = (dword *)WorkBuffer;
  XMSIndex = GetTapeBlockXMSIndex (WhichWindow, Entry);
  CopyFromXMS (Entry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, BHWTypes);
  NumTypes = (word)(Entry->BlockLength / 3);
  for (Cnt = 0 ; Cnt < NumTypes ; Cnt ++)                                                                  /* Copy hardware table */
  {
    Bt1 = *(BHWTypes ++);                                                                                      /* (Hardware Type) */
    Bt2 = *(BHWTypes ++);                                                                                        /* (Hardware ID) */
    Bt3 = *(BHWTypes ++);                                                                                              /* (Value) */
    *(DHWTypes + Cnt) = ((dword)Bt3 << 16) | ((dword)Bt1 << 8) | Bt2;
  }
  for (Cnt = NumTypes - 1 ; Cnt > 0 && More ; Cnt --)                                                              /* Bubble-sort */
  {
    More = FALSE;
    for (Cnt2 = 0 ; Cnt2 < Cnt ; Cnt ++)
      if (*(DHWTypes + Cnt2) > *(DHWTypes + Cnt2 + 1))
      {
        Swap = *(DHWTypes + Cnt2);
        *(DHWTypes + Cnt2) = *(DHWTypes + Cnt2 + 1);
        *(DHWTypes + Cnt2 + 1) = Swap;
        More = TRUE;
      }
  }
}

bool EditHeaderBlock (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : The header block has been adjusted by the user and saved in the tape directory.                                        */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct WindowInfo CWindow   = {193, 150, 256, 150, 256, 150, 256, 150, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                 WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "Edit Header Block"};
  struct ButtonInfo PCancel   = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 63, 125, 8, DBLACK, DWHITE, NULL, 1, "@Cancel", NULL};
  struct ButtonInfo POk       = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 137, 125, 8, DBLACK, DWHITE, NULL, 0, "@Ok", NULL};
  struct ButtonInfo BBlType   = {BT_TEXT, BC_TEXT, BF_FWIDTH | BF_FITTXT, 112, 20, 20, DBLACK, DWHITE, NULL, 0, NULL, NULL};
  struct ButtonInfo BBlTypSel = {BT_PUSH, BC_GADGET, BF_FWIDTH | BF_KEYSELECT, 239, 20, 1, DBLACK, DWHITE, NULL, 2, "\012", NULL};
  struct ButtonInfo BHName    = {BT_INPUT, BC_TEXT, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 112, 40, 10, DBLACK, DWHITE, NULL, 3,
                                 NULL, NULL};
  struct ButtonInfo BBlLength = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 112, 60, 5, DBLACK, DWHITE, NULL, 4,
                                 NULL, NULL};
  struct ButtonInfo BStart    = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 112, 80, 5, DBLACK, DWHITE, NULL, 5,
                                 NULL, NULL};
  struct ButtonInfo BExtra    = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 112, 100, 5, DBLACK, DWHITE, NULL, 6,
                                 NULL, NULL};
  struct ComboInfo  SWindow;
  struct ComboEntry SEntry;
  char              ValBlockTypeID[21];
  char              ValHName[11];
  char              ValBlockLength[6];
  char              ValStart[6];
  char              ValLetter[6];
  char              ValExtra[6];
  byte              WType;
  word              WBlockLength;
  word              WStart;
  word              WExtra;
  comboent          SelectedLine;
  struct ButtonInfo *ButtonP;
  bool              Done      = FALSE;
  bool              Finished  = FALSE;
  bool              Cancelled = FALSE;
  bool              Changed   = FALSE;
  comboent          OldValue;
  short    register Cnt;
  dword             XMSIndex;

  BBlType.Contents = ValBlockTypeID;
  BHName.Contents = ValHName;
  BHName.Font = SpectrumFont;
  BBlLength.Contents = ValBlockLength;
  BStart.Contents = ValStart;
  BExtra.Contents = ValExtra;
  XMSIndex = GetTapeBlockXMSIndex (WhichWindow, Entry);
  CopyFromXMS (20, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);                        /* Fetch the header data */
  WType = *WorkBuffer;
  switch (WType)
  {
    case 0x00 : strcpy (ValBlockTypeID, "Program");
                BExtra.Flags &= ~BF_GHOSTED;
                break;
    case 0x01 : strcpy (ValBlockTypeID, "Numerical Array");
                BExtra.Flags |= BF_GHOSTED;
                BStart.Contents = ValLetter;
                BStart.ContentsType = BC_ALPHA;
                break;
    case 0x02 : strcpy (ValBlockTypeID, "Alphanumerical Array");
                BExtra.Flags |= BF_GHOSTED;
                BStart.Contents = ValLetter;
                BStart.ContentsType = BC_ALPHA;
                break;
    case 0x03 : strcpy (ValBlockTypeID, "Bytes");
                BExtra.Flags |= BF_GHOSTED;
                break;
    default   : strcpy (ValBlockTypeID, "<UNKNOWN>");
                BExtra.Flags |= BF_GHOSTED;
                break;
  }
  strncpy (ValHName, WorkBuffer + 1, 10);
  Cnt = 10;
  while (Cnt > 0 && !Done)
    if (ValHName[Cnt - 1] == ' ')
      Cnt --;
    else
      Done = TRUE;
  ValHName[Cnt] = '\0';
  WBlockLength = *((word *)(WorkBuffer + 11));
  if (WType == 0x01 || WType == 0x02)
    WStart = 0;
  else
    WStart = *((word *)(WorkBuffer + 13));
  WExtra = *((word *)(WorkBuffer + 15));
  sprintf (ValBlockLength, "%u", WBlockLength);
  sprintf (ValStart, "%u", WStart);
  switch (WType)
  {
    case 0x01 : sprintf (ValLetter, "%c", *(WorkBuffer + 14) - 0x40); break;
    case 0x02 : sprintf (ValLetter, "%c", *(WorkBuffer + 14) - 0x80); break;
    default   : strcpy (ValLetter, "A"); break;
  }
  sprintf (ValExtra, "%u", WExtra);
  AddButton (&CWindow, &POk, TRUE);
  AddButton (&CWindow, &BBlType, FALSE);
  AddButton (&CWindow, &BBlTypSel, FALSE);
  AddButton (&CWindow, &BHName, FALSE);
  AddButton (&CWindow, &BBlLength, FALSE);
  AddButton (&CWindow, &BStart, FALSE);
  AddButton (&CWindow, &BExtra, FALSE);
  AddButton (&CWindow, &PCancel, FALSE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  Wprintf (&CWindow, 2, 1, "{+1Block type}N");
  Wprintf (&CWindow, 4, 1, "{+1Block name}N");
  Wprintf (&CWindow, 6, 1, "{+1Block length}N");
  if (WType == 0x01 || WType == 0x02)
  {
    Wprintf (&CWindow, 8, 1, "{+1Variable}N");
    BStart.MaxInputLength = 1;
  }
  else
    Wprintf (&CWindow, 8, 1, "{+1Block start}N");
  if (WType == 0x00)
    Wprintf (&CWindow, 10, 1, "{+1BASIC length}N");
  else
    Wprintf (&CWindow, 10, 1, "{+1Extra}N");
  Done = FALSE;
  while (!Done)
  {
    switch (HandleWindow (&CWindow, &ButtonP))
    {
      case WINH_CLOSE   : Done = TRUE; Cancelled = TRUE; break;
      case WINH_RNOHAND : switch (ButtonP->ReactNum)
                          {
                            case 0  : Done = TRUE; Cancelled = FALSE; break;                                              /* (Ok) */
                            case 1  : Done = TRUE; Cancelled = TRUE; break;                                           /* (Cancel) */
                            case 2  : MakeComboWindow (&SWindow, &CWindow, COMT_SELECT, COMF_VSLIDER | COMF_SYNCSLIDER |
                                                       COMF_KEYSELECT | COMF_EXCLUSIVE, CWindow.StartX + BBlType.StartX + 1,
                                                       CWindow.StartY + BBlType.StartY + 16, 0, 1024, BBlType.FWidth + 2,
                                                       BBlType.FWidth, 4, NULL, FALSE, 0, NULL, DBLACK, DWHITE, 0, 0);
                                      SEntry.Selected = 0;
                                      SEntry.Identifier = Cnt;
                                      strcpy (SEntry.Text, "Program");
                                      if (!InsertComboEntry (&SWindow, &SEntry, NULL, -1))
                                        FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                                      strcpy (SEntry.Text, "Numerical Array");
                                      if (!InsertComboEntry (&SWindow, &SEntry, NULL, -1))
                                        FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                                      strcpy (SEntry.Text, "Alphanumerical Array");
                                      if (!InsertComboEntry (&SWindow, &SEntry, NULL, -1))
                                        FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                                      strcpy (SEntry.Text, "Bytes");
                                      if (!InsertComboEntry (&SWindow, &SEntry, NULL, -1))
                                        FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                                      if (WType > 3)
                                        SWindow.LastSelected = -1;
                                      else
                                        SWindow.LastSelected = WType;                       /* Auto-select the current block type */
                                      OldValue = SWindow.LastSelected;
                                      PrintComboEntries (&SWindow);
                                      Finished = FALSE;
                                      while (!Finished)
                                      {
                                        MouseStatus ();
                                        if (HandleComboWindow (&SWindow, &SelectedLine) == WINH_LINEFIRST)
                                        {
                                          Finished = TRUE;
                                          WType = SWindow.LastSelected;
                                        }
                                        else
                                        {
                                          HandleWindow (&(SWindow.ComboWindow), &ButtonP);            /* (Just read the keyboard) */
                                         if (Mouse.Xx < SWindow.OverlayWindow.A_StartX || Mouse.Xx > SWindow.OverlayWindow.A_EndX ||
                                              Mouse.Yy < SWindow.OverlayWindow.A_StartY || Mouse.Yy > SWindow.OverlayWindow.A_EndY)
                                            if (Mouse.Left)                                /* Left click outside the select box ? */
                                            {
                                              Finished = TRUE;
                                              DelayLeftMouseButton ();
                                            }
                                        }
                                        if (GlobalKey == K_ESC)
                                          Finished = TRUE;
                                        else if (GlobalKey == K_ENTER)
                                        { WType = SWindow.LastSelected; Finished = TRUE; }
                                      }
                                      GlobalKey = 0x0000;
                                      Finished = FALSE;
                                      DestroyComboWindow (&SWindow, 0);
                                      if (WType > 0x03)
                                        WType = 0x03;
                                      if (WType != OldValue)                                     /* Was a different type chosen ? */
                                      {
                                        switch (WType)
                                        {
                                          case 0x00 : strcpy (ValBlockTypeID, "Program");
                                                      BExtra.Flags &= ~BF_GHOSTED;
                                                      BStart.Contents = ValStart;
                                                      BStart.ContentsType = BC_NUMBER;
                                                      BStart.MaxInputLength = 5;
                                                      Wprintf (&CWindow, 8, 1, "{+1Auto start }N");
                                                      Wprintf (&CWindow, 10, 1, "{+1BASIC length}N");
                                                      break;
                                          case 0x01 : strcpy (ValBlockTypeID, "Numerical Array");
                                                      BExtra.Flags |= BF_GHOSTED;
                                                      BStart.Contents = ValLetter;
                                                      BStart.ContentsType = BC_ALPHA;
                                                      BStart.MaxInputLength = 1;
                                                      Wprintf (&CWindow, 8, 1, "{+1Variable   }N");
                                                      Wprintf (&CWindow, 10, 1, "{+1Extra       }N");
                                                      break;
                                          case 0x02 : strcpy (ValBlockTypeID, "Alphanumerical Array");
                                                      BExtra.Flags |= BF_GHOSTED;
                                                      BStart.Contents = ValLetter;
                                                      BStart.ContentsType = BC_ALPHA;
                                                      BStart.MaxInputLength = 1;
                                                      Wprintf (&CWindow, 8, 1, "{+1Variable   }N");
                                                      Wprintf (&CWindow, 10, 1, "{+1Extra       }N");
                                                      break;
                                          case 0x03 : strcpy (ValBlockTypeID, "Bytes");
                                                      BExtra.Flags |= BF_GHOSTED;
                                                      BStart.Contents = ValStart;
                                                      BStart.ContentsType = BC_NUMBER;
                                                      BStart.MaxInputLength = 5;
                                                      Wprintf (&CWindow, 8, 1, "{+1Block start}N");
                                                      Wprintf (&CWindow, 10, 1, "{+1Extra       }N");
                                                      break;
                                        }
                                        UpdateButtonText (&CWindow, &BBlType, BF_FITTXT);
                                        UpdateButtonText (&CWindow, &BExtra, BF_FITTXT);
                                        UpdateButtonText (&CWindow, &BStart, BF_FITTXT);
                                      }
                                      break;
                          }
      case WINH_EDITEND : if (ButtonP->Flags & BF_FITINPUT)                                            /* Contents were changed ? */
                            switch (ButtonP->ReactNum)
                            {
                              case 4  : WBlockLength = atoi (ValBlockLength);
                                        sprintf (ValBlockLength, "%u", WBlockLength);
                                        UpdateButtonText (&CWindow, &BBlLength, BF_FITTXT);
                                        break;
                              case 5  : if (BStart.ContentsType == BC_NUMBER)
                                        {
                                          WStart = atoi (ValStart);
                                          sprintf (ValStart, "%u", WStart);
                                          UpdateButtonText (&CWindow, &BStart, BF_FITTXT);
                                        }
                                        break;
                              case 6  : WExtra = atoi (ValExtra);
                                        sprintf (ValExtra, "%u", WExtra);
                                        UpdateButtonText (&CWindow, &BExtra, BF_FITTXT);
                                        break;
                              default : break;
                            }
    }
    if (GlobalKey == K_ESC)  { Done = TRUE; Cancelled = TRUE; }
    MouseStatus ();
  }
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
  if (Cancelled)
  {
    if (_EntryNumber != -1)                                                                     /* Came from the insert handler ? */
      DeleteTapeBlock (WhichWindow, Entry, FALSE, FALSE);                         /* Then we don't have to keep the entry anymore */
    _Cancelled = TRUE;
    return (FALSE);
  }
  *(WorkBuffer + 32) = WType;
  memset (WorkBuffer + 33, ' ', 10);                                            /* Ensure the filename part is padded with spaces */
  strncpy (WorkBuffer + 33, ValHName, strlen (ValHName));
  *((word *)(WorkBuffer + 43)) = WBlockLength;
  switch (WType)
  {
    case 0x00 : if (ValStart[0] == '\0' || WStart > 9999)                                                        /* No auto-start */
                  *((word *)(WorkBuffer + 45)) = 0x8000;
                else
                  *((word *)(WorkBuffer + 45)) = WStart;
                *((word *)(WorkBuffer + 47)) = WExtra;
                break;
    case 0x01 :
    case 0x02 : *(WorkBuffer + 45) = 0x00;
                if (ValLetter[0] == '\0')
                  *(WorkBuffer + 46) = 'A';
                else
                  *(WorkBuffer + 46) = toupper (ValLetter[0]) + (WType == 0x01 ? 0x40 : 0x80);
                *((word *)(WorkBuffer + 47)) = 0x8000;
                break;
    default   : *((word *)(WorkBuffer + 45)) = WStart;
                *((word *)(WorkBuffer + 47)) = 0x8000;
  }
  if ((Changed = memcmp (WorkBuffer, WorkBuffer + 32, 17)))                                          /* Header has been changed ? */
  {
    Entry->CalculatedParity = Entry->FlagByte;                                                                 /* Init the parity */
    for (Cnt = 0 ; Cnt < 17 ; Cnt ++)
      Entry->CalculatedParity ^= *(WorkBuffer + 32 + Cnt);                                                /* Calculate the parity */
    Entry->ParityByte = Entry->CalculatedParity;                                                     /* Make the parity 'correct' */
    CopyToXMS (20, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer + 32, XMSIndex);                   /* Store the header data */
  }
  return (Changed);
}

bool EditAllTiming (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : The entry has been adjusted by the user and saved in the tape directory.                                               */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  static struct _Fixed_s
  {
    bool            BlockType;
    bool            PilotPulse;
    bool            PilotLength;
    bool            SyncPulse1;
    bool            SyncPulse2;
    bool            ZeroBitPulse;
    bool            OneBitPulse;
    bool            LengthLastByte;
    bool            PauseLength;
  };
  struct _Fixed_s   Fixed     = { TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE };
  struct _Fixed_s   OldFixed;
  struct WindowInfo CWindow   = {193, 0, 256, 0, 256, 0, 256, 0, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                 WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL,
                                 "Edit Timing Of All Selected Blocks"};
  struct ButtonInfo PCancel   = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 63, 0, 8, DBLACK, DWHITE, NULL, 1, "@Cancel", NULL};
  struct ButtonInfo POk       = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 137, 0, 8, DBLACK, DWHITE, NULL, 0, "@Ok", NULL};
  struct ButtonInfo BBlType   = {BT_TEXT, BC_TEXT, BF_FWIDTH | BF_FITTXT, 112, 0, 20, DBLACK, DWHITE, NULL, 0, NULL, NULL};
  struct ButtonInfo BBlTypSel = {BT_PUSH, BC_GADGET, BF_FWIDTH | BF_KEYSELECT, 239, 0, 1, DBLACK, DWHITE, NULL, 2, "\007", NULL};
  struct ButtonInfo BPilotPul = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 112, 0, 5, DBLACK, DWHITE, NULL, 4,
                                 NULL, NULL};
  struct ButtonInfo BPilotLen = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 112, 0, 5, DBLACK, DWHITE, NULL, 5,
                                 NULL, NULL};
  struct ButtonInfo BSyncPul1 = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 112, 0, 5, DBLACK, DWHITE, NULL, 6,
                                 NULL, NULL};
  struct ButtonInfo BSyncPul2 = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 112, 0, 5, DBLACK, DWHITE, NULL, 7,
                                 NULL, NULL};
  struct ButtonInfo BZeroBit  = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 112, 0, 5, DBLACK, DWHITE, NULL, 8,
                                 NULL, NULL};
  struct ButtonInfo BOneBit   = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 112, 0, 5, DBLACK, DWHITE, NULL, 9,
                                 NULL, NULL};
  struct ButtonInfo BLenLast  = {BT_TEXT, BC_NUMBER, BF_FWIDTH | BF_FITTXT, 112, 0, 1, DBLACK, DWHITE, NULL, 10,
                                 NULL, NULL};
  struct ButtonInfo BDropEnd  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT | BF_NOSURROUND, 126, 0, 8, DBLACK, DWHITE, NULL, 12,
                                 "@Strip", NULL};
  struct ButtonInfo BPauseLen = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 112, 0, 5, DBLACK, DWHITE, NULL, 11,
                                 NULL, NULL};

  struct ButtonInfo      *ButtonP;
  struct TapeEntry_s      EditEntry;
  struct TapeEntry_s      CompareEntry;
  struct TapeEntry_s     *DirP;
  char                    ValBlockTypeID[30] = "\0";
  bool                    LockBlockType = FALSE;
  char                    ValPilotPulse[6] = "\0";
  char                    ValPilotLength[6] = "\0";
  char                    ValSyncPulse1[6] = "\0";
  char                    ValSyncPulse2[6] = "\0";
  char                    ValZeroBitPulse[6] = "\0";
  char                    ValOneBitPulse[6] = "\0";
  char                    ValLengthLastByte[2] = "\0";
  char                    ValPauseLength[6] = "\0";
  char                   *TmpString;
  bool                    Done      = FALSE;
  bool                    Cancelled = FALSE;
  bool                    Changed   = FALSE;
  short                   NumLines;
  short                   CurrentLine;

  BBlType.Contents = ValBlockTypeID;
  BPilotPul.Contents = ValPilotPulse;
  BPilotLen.Contents = ValPilotLength;
  BSyncPul1.Contents = ValSyncPulse1;
  BSyncPul2.Contents = ValSyncPulse2;
  BZeroBit.Contents = ValZeroBitPulse;
  BOneBit.Contents = ValOneBitPulse;
  BLenLast.Contents = ValLengthLastByte;
  BPauseLen.Contents = ValPauseLength;
  DirP = TapeWindow[WhichWindow].Directory;
  while (DirP)
  {
    if ((DirP->ComboSelected & SELECTBIT) && (DirP->BlockTypeID == 0x10 || DirP->BlockTypeID == 0x11))
    {
      if (!Done)                                                                                                   /* First one ? */
      {
        memcpy (&EditEntry, DirP, sizeof (struct TapeEntry_s));                                   /* Use a copy of the tape entry */
        if (DirP->BlockTypeID == 0x10)
        {
          EditEntry.Timing.LengthPilotPulse = NS_LENPILOT;
          EditEntry.Timing.NumPilotPulses = DirP->Flags.IsHeader ? NS_HDPILOT : NS_DTPILOT;
          EditEntry.Timing.LengthSyncPulse[0] = NS_LENSYNC1;
          EditEntry.Timing.LengthSyncPulse[1] = NS_LENSYNC2;
          EditEntry.Timing.LengthZeroBitPulse = NS_LEN0BIT;
          EditEntry.Timing.LengthOneBitPulse = NS_LEN1BIT;
          EditEntry.Timing.NumBitsInLastByte = 8;
        }
        if (DirP->BlockLength > 65534L)
          LockBlockType = TRUE;
        Done = TRUE;
      }
      else
      {
        memcpy (&CompareEntry, DirP, sizeof (struct TapeEntry_s));                                /* Use a copy of the tape entry */
        if (DirP->BlockTypeID == 0x10)
        {
          CompareEntry.Timing.LengthPilotPulse = NS_LENPILOT;
          CompareEntry.Timing.NumPilotPulses = DirP->Flags.IsHeader ? NS_HDPILOT : NS_DTPILOT;
          CompareEntry.Timing.LengthSyncPulse[0] = NS_LENSYNC1;
          CompareEntry.Timing.LengthSyncPulse[1] = NS_LENSYNC2;
          CompareEntry.Timing.LengthZeroBitPulse = NS_LEN0BIT;
          CompareEntry.Timing.LengthOneBitPulse = NS_LEN1BIT;
          CompareEntry.Timing.NumBitsInLastByte = 8;
        }
        if (CompareEntry.BlockTypeID != EditEntry.BlockTypeID)                                 Fixed.BlockType = FALSE;
        if (CompareEntry.Timing.LengthPilotPulse != EditEntry.Timing.LengthPilotPulse)         Fixed.PilotPulse = FALSE;
        if (CompareEntry.Timing.NumPilotPulses != EditEntry.Timing.NumPilotPulses)             Fixed.PilotLength = FALSE;
        if (CompareEntry.Timing.LengthSyncPulse[0] != EditEntry.Timing.LengthSyncPulse[0])     Fixed.SyncPulse1 = FALSE;
        if (CompareEntry.Timing.LengthSyncPulse[1] != EditEntry.Timing.LengthSyncPulse[1])     Fixed.SyncPulse2 = FALSE;
        if (CompareEntry.Timing.LengthZeroBitPulse != EditEntry.Timing.LengthZeroBitPulse)     Fixed.ZeroBitPulse = FALSE;
        if (CompareEntry.Timing.LengthOneBitPulse != EditEntry.Timing.LengthOneBitPulse)       Fixed.OneBitPulse = FALSE;
        if (CompareEntry.Timing.NumBitsInLastByte != EditEntry.Timing.NumBitsInLastByte)       Fixed.LengthLastByte = FALSE;
        if (CompareEntry.PauseLength != EditEntry.PauseLength)                                 Fixed.PauseLength = FALSE;
        if (CompareEntry.BlockLength > 65534L)
          LockBlockType = TRUE;
      }
    }
    DirP = DirP->Next;
  }
  if (!Done)                                                                                 /* No selected blocks 0x10 or 0x11 ? */
    return (FALSE);                                                                                /* Return: nothing has changed */
  Done = FALSE;
  NumLines = 9;
  SelectFont (DefaultFont, FONT_NORMAL);
  CWindow.StartY = (Screen.Height - (2 * NumLines * (CurrentFont->Height + 2))) / 2 - 30;
  CWindow.Height = CWindow.MinHeight = CWindow.MaxHeight = 2 * NumLines * (CurrentFont->Height + 2) + 60;
  PCancel.StartY = POk.StartY = 2 * NumLines * (CurrentFont->Height + 2) + 30;
  AddButton (&CWindow, &POk, TRUE);
  CurrentLine = 1;
  BBlType.StartY = BBlTypSel.StartY = CurrentLine * (CurrentFont->Height + 2);
  if (LockBlockType)
    EditEntry.BlockTypeID = 0x11;
  else if (!Fixed.BlockType)
    EditEntry.BlockTypeID = 0x10;                                                    /* (So next cycle button click goes to 0x11) */
  if (!Fixed.BlockType && !LockBlockType)
    BBlType.BPen = LBLACK;
  else
  {
    GetBlockTypeInfo (EditEntry.BlockTypeID, NULL, NULL, &TmpString, NULL, NULL);
    strcpy (ValBlockTypeID, TmpString);
  }
  if (!Fixed.PilotPulse)
  { BPilotPul.BPen = LBLACK; BPilotPul.ContentsType |= BC_STARTREDRAW; }
  else sprintf (ValPilotPulse, "%u", EditEntry.Timing.LengthPilotPulse);
  if (!Fixed.PilotLength)
  { BPilotLen.BPen = LBLACK; BPilotLen.ContentsType |= BC_STARTREDRAW; }
  else sprintf (ValPilotLength, "%u", EditEntry.Timing.NumPilotPulses);
  if (!Fixed.SyncPulse1)
  { BSyncPul1.BPen = LBLACK; BSyncPul1.ContentsType |= BC_STARTREDRAW; }
  else sprintf (ValSyncPulse1, "%u", EditEntry.Timing.LengthSyncPulse[0]);
  if (!Fixed.SyncPulse2)
  { BSyncPul2.BPen = LBLACK; BSyncPul2.ContentsType |= BC_STARTREDRAW; }
  else sprintf (ValSyncPulse2, "%u", EditEntry.Timing.LengthSyncPulse[1]);
  if (!Fixed.ZeroBitPulse)
  { BZeroBit.BPen = LBLACK; BZeroBit.ContentsType |= BC_STARTREDRAW; }
  else sprintf (ValZeroBitPulse, "%u", EditEntry.Timing.LengthZeroBitPulse);
  if (!Fixed.OneBitPulse)
  { BOneBit.BPen = LBLACK; BOneBit.ContentsType |= BC_STARTREDRAW; }
  else sprintf (ValOneBitPulse, "%u", EditEntry.Timing.LengthOneBitPulse);
  if (!Fixed.LengthLastByte)
  { BLenLast.BPen = LBLACK; BLenLast.ContentsType |= BC_STARTREDRAW; }
  else sprintf (ValLengthLastByte, "%u", EditEntry.Timing.NumBitsInLastByte);
  if (!Fixed.PauseLength)
  { BPauseLen.BPen = LBLACK; BPauseLen.ContentsType |= BC_STARTREDRAW; }
  else sprintf (ValPauseLength, "%u", EditEntry.PauseLength);
  AddButton (&CWindow, &BBlType, FALSE);
  AddButton (&CWindow, &BBlTypSel, FALSE);
  CurrentLine += 2;
  BPilotPul.StartY = CurrentLine * (CurrentFont->Height + 2);   AddButton (&CWindow, &BPilotPul, FALSE);  CurrentLine += 2;
  BPilotLen.StartY = CurrentLine * (CurrentFont->Height + 2);   AddButton (&CWindow, &BPilotLen, FALSE);  CurrentLine += 2;
  BSyncPul1.StartY = CurrentLine * (CurrentFont->Height + 2);   AddButton (&CWindow, &BSyncPul1, FALSE);  CurrentLine += 2;
  BSyncPul2.StartY = CurrentLine * (CurrentFont->Height + 2);   AddButton (&CWindow, &BSyncPul2, FALSE);  CurrentLine += 2;
  BZeroBit.StartY = CurrentLine * (CurrentFont->Height + 2);    AddButton (&CWindow, &BZeroBit, FALSE);   CurrentLine += 2;
  BOneBit.StartY = CurrentLine * (CurrentFont->Height + 2);     AddButton (&CWindow, &BOneBit, FALSE);    CurrentLine += 2;
  BLenLast.StartY = CurrentLine * (CurrentFont->Height + 2);    AddButton (&CWindow, &BLenLast, FALSE);
  BDropEnd.StartY = CurrentLine * (CurrentFont->Height + 2);    AddButton (&CWindow, &BDropEnd, FALSE);   CurrentLine += 2;
  BPauseLen.StartY = CurrentLine * (CurrentFont->Height + 2);   AddButton (&CWindow, &BPauseLen, FALSE);  CurrentLine += 2;
  if (EditEntry.BlockTypeID == 0x10)
  {
    BPilotPul.Flags |= BF_GHOSTED;
    BPilotLen.Flags |= BF_GHOSTED;
    BSyncPul1.Flags |= BF_GHOSTED;
    BSyncPul2.Flags |= BF_GHOSTED;
    BZeroBit.Flags |= BF_GHOSTED;
    BOneBit.Flags |= BF_GHOSTED;
  }
  if (Fixed.LengthLastByte && EditEntry.Timing.NumBitsInLastByte == 8)                                 /* No loose end to strip ? */
  {
    BLenLast.Flags |= BF_GHOSTED;
    BDropEnd.Flags |= BF_GHOSTED;
  }
  if (LockBlockType)                                                                        /* Cannot convert to Normal Speed, as */
  {                                                                                       /* such blocks have only 2 length bytes */
    BBlType.Flags |= BF_GHOSTED;
    BBlTypSel.Flags |= BF_GHOSTED;
  }
  AddButton (&CWindow, &PCancel, FALSE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  CurrentLine = 1;
  Wprintf (&CWindow, CurrentLine, 1, "{+1Block type}N");
  CurrentLine += 2;
  Wprintf (&CWindow, CurrentLine, 1, "{+1PILOT pulse}N");
  Wprintf (&CWindow, CurrentLine, 25, "{+1T}N  [%4u]", NS_LENPILOT);
  CurrentLine += 2;
  Wprintf (&CWindow, CurrentLine, 1, "{+1PILOT length}N");
  Wprintf (&CWindow, CurrentLine, 25, "   [%4u|%4u]", NS_HDPILOT, NS_DTPILOT);
  CurrentLine += 2;
  Wprintf (&CWindow, CurrentLine, 1, "{+1SYNC pulse 1}N");
  Wprintf (&CWindow, CurrentLine, 25, "{+1T}N  [%4u]", NS_LENSYNC1);
  CurrentLine += 2;
  Wprintf (&CWindow, CurrentLine, 1, "{+1SYNC pulse 2}N");
  Wprintf (&CWindow, CurrentLine, 25, "{+1T}N  [%4u]", NS_LENSYNC2);
  CurrentLine += 2;
  Wprintf (&CWindow, CurrentLine, 1, "{+1ZERO bit pulse}N");
  Wprintf (&CWindow, CurrentLine, 25, "{+1T}N  [%4u]", NS_LEN0BIT);
  CurrentLine += 2;
  Wprintf (&CWindow, CurrentLine, 1, "{+1ONE bit pulse}N");
  Wprintf (&CWindow, CurrentLine, 25, "{+1T}N  [%4u]", NS_LEN1BIT);
  CurrentLine += 2;
  Wprintf (&CWindow, CurrentLine, 1, "{+1Last byte length}N");
  CurrentLine += 2;
  Wprintf (&CWindow, CurrentLine, 1, "{+1Pause length}N");
  Wprintf (&CWindow, CurrentLine, 25, "{+1ms.}N");
  BBlType.BPen = BPilotPul.BPen = BPilotLen.BPen = BSyncPul1.BPen = BSyncPul2.BPen = BZeroBit.BPen = BOneBit.BPen =
  BLenLast.BPen = BPauseLen.BPen = DWHITE;
  memcpy (&OldFixed, &Fixed, sizeof (struct _Fixed_s));                                                  /* Keep old fixed status */
  memcpy (&CompareEntry, &EditEntry, sizeof (struct TapeEntry_s));                                        /* Keep old edit values */
  while (!Done)
  {
    switch (HandleWindow (&CWindow, &ButtonP))
    {
      case WINH_CLOSE   : Done = TRUE; Cancelled = TRUE; break;
      case WINH_RNOHAND : switch (ButtonP->ReactNum)
                          {
                            case 0  : Done = TRUE; Cancelled = FALSE; break;                                              /* (Ok) */
                            case 1  : Done = TRUE; Cancelled = TRUE; break;                                           /* (Cancel) */
                            case 2  : switch (EditEntry.BlockTypeID)                                     /* Toggle the block type */
                                      {
                                        case 0x10 : EditEntry.BlockTypeID = 0x11; break;
                                        case 0x11 : if (EditEntry.BlockLength > 65537L)            /* Type 0x10 cannot handle it! */
                                                      EditEntry.BlockTypeID = 0x11;            /* (65537 = 65535 + flag + parity) */
                                                    else
                                                      EditEntry.BlockTypeID = 0x10;
                                                    break;
                                      }
                                      if (EditEntry.BlockTypeID == 0x10)
                                      {
                                        BPilotPul.Flags |= BF_GHOSTED;
                                        BPilotLen.Flags |= BF_GHOSTED;
                                        BSyncPul1.Flags |= BF_GHOSTED;
                                        BSyncPul2.Flags |= BF_GHOSTED;
                                        BZeroBit.Flags |= BF_GHOSTED;
                                        BOneBit.Flags |= BF_GHOSTED;
                                      }
                                      else
                                      {
                                        BPilotPul.Flags &= ~BF_GHOSTED;
                                        BPilotLen.Flags &= ~BF_GHOSTED;
                                        BSyncPul1.Flags &= ~BF_GHOSTED;
                                        BSyncPul2.Flags &= ~BF_GHOSTED;
                                        BZeroBit.Flags &= ~BF_GHOSTED;
                                        BOneBit.Flags &= ~BF_GHOSTED;
                                      }
                                      GetBlockTypeInfo (EditEntry.BlockTypeID, NULL, NULL, &TmpString, NULL, NULL);
                                      strcpy (ValBlockTypeID, TmpString);
                                      Fixed.BlockType = TRUE;
                                      ClearButton (&CWindow, &BBlType);
                                      UpdateButtonText (&CWindow, &BBlType, BF_FITTXT);
                                      if (Fixed.PilotPulse)   UpdateButtonText (&CWindow, &BPilotPul, BF_FITTXT);
                                      if (Fixed.PilotLength)  UpdateButtonText (&CWindow, &BPilotLen, BF_FITTXT);
                                      if (Fixed.SyncPulse1)   UpdateButtonText (&CWindow, &BSyncPul1, BF_FITTXT);
                                      if (Fixed.SyncPulse2)   UpdateButtonText (&CWindow, &BSyncPul2, BF_FITTXT);
                                      if (Fixed.ZeroBitPulse) UpdateButtonText (&CWindow, &BZeroBit, BF_FITTXT);
                                      if (Fixed.OneBitPulse)  UpdateButtonText (&CWindow, &BOneBit, BF_FITTXT);
                                      break;
                            case 12 : EditEntry.Timing.NumBitsInLastByte = 8;                                          /* (Strip) */
                                      sprintf (ValLengthLastByte, "%u", EditEntry.Timing.NumBitsInLastByte);
                                      BDropEnd.Flags |= BF_GHOSTED;                                      /* Can be done only once */
                                      BLenLast.Flags |= BF_GHOSTED;
                                      UpdateButtonText (&CWindow, &BDropEnd, 0);
                                      UpdateButtonText (&CWindow, &BLenLast, BF_FITTXT);
                                      break;
                          }
                          break;
      case WINH_EDITESC :
      case WINH_EDITEND : if (ButtonP->Flags & BF_FITINPUT)                                            /* Contents were changed ? */
                            switch (ButtonP->ReactNum)
                            {
                              case 4  : EditEntry.Timing.LengthPilotPulse = atoi (ValPilotPulse);
                                        sprintf (ValPilotPulse, "%u", EditEntry.Timing.LengthPilotPulse);
                                        Fixed.PilotPulse = TRUE;
                                        BPilotPul.ContentsType &= ~BC_STARTREDRAW;
                                        UpdateButtonText (&CWindow, &BPilotPul, BF_FITTXT);
                                        break;
                              case 5  : EditEntry.Timing.NumPilotPulses = atoi (ValPilotLength);
                                        sprintf (ValPilotLength, "%u", EditEntry.Timing.NumPilotPulses);
                                        Fixed.PilotLength = TRUE;
                                        BPilotLen.ContentsType &= ~BC_STARTREDRAW;
                                        UpdateButtonText (&CWindow, &BPilotLen, BF_FITTXT);
                                        break;
                              case 6  : EditEntry.Timing.LengthSyncPulse[0] = atoi (ValSyncPulse1);
                                        sprintf (ValSyncPulse1, "%u", EditEntry.Timing.LengthSyncPulse[0]);
                                        Fixed.SyncPulse1 = TRUE;
                                        BSyncPul1.ContentsType &= ~BC_STARTREDRAW;
                                        UpdateButtonText (&CWindow, &BSyncPul1, BF_FITTXT);
                                        break;
                              case 7  : EditEntry.Timing.LengthSyncPulse[1] = atoi (ValSyncPulse2);
                                        sprintf (ValSyncPulse2, "%u", EditEntry.Timing.LengthSyncPulse[1]);
                                        Fixed.SyncPulse2 = TRUE;
                                        BSyncPul2.ContentsType &= ~BC_STARTREDRAW;
                                        UpdateButtonText (&CWindow, &BSyncPul2, BF_FITTXT);
                                        break;
                              case 8  : EditEntry.Timing.LengthZeroBitPulse = atoi (ValZeroBitPulse);
                                        sprintf (ValZeroBitPulse, "%u", EditEntry.Timing.LengthZeroBitPulse);
                                        Fixed.ZeroBitPulse = TRUE;
                                        BZeroBit.ContentsType &= ~BC_STARTREDRAW;
                                        UpdateButtonText (&CWindow, &BZeroBit, BF_FITTXT);
                                        break;
                              case 9  : EditEntry.Timing.LengthOneBitPulse = atoi (ValOneBitPulse);
                                        sprintf (ValOneBitPulse, "%u", EditEntry.Timing.LengthOneBitPulse);
                                        Fixed.OneBitPulse = TRUE;
                                        BOneBit.ContentsType &= ~BC_STARTREDRAW;
                                        UpdateButtonText (&CWindow, &BOneBit, BF_FITTXT);
                                        break;
                              case 10 : EditEntry.Timing.NumBitsInLastByte = atoi (ValLengthLastByte);
                                        sprintf (ValLengthLastByte, "%u", EditEntry.Timing.NumBitsInLastByte);
                                        Fixed.LengthLastByte = TRUE;
                                        BLenLast.ContentsType &= ~BC_STARTREDRAW;
                                        UpdateButtonText (&CWindow, &BLenLast, BF_FITTXT);
                                        break;
                              case 11 : EditEntry.PauseLength = atoi (ValPauseLength);
                                        sprintf (ValPauseLength, "%u", EditEntry.PauseLength);
                                        Fixed.PauseLength = TRUE;
                                        BPauseLen.ContentsType &= ~BC_STARTREDRAW;
                                        UpdateButtonText (&CWindow, &BPauseLen, BF_FITTXT);
                                        break;
                            }
                          else                                                                            /* No change or escaped */
                            switch (ButtonP->ReactNum)
                            {
                              case 4  : if (!Fixed.PilotPulse)
                                        { BPilotPul.BPen = LBLACK; ClearButton (&CWindow, &BPilotPul); BPilotPul.BPen = DWHITE; }
                                        break;
                              case 5  : if (!Fixed.PilotLength)
                                        { BPilotLen.BPen = LBLACK; ClearButton (&CWindow, &BPilotLen); BPilotLen.BPen = DWHITE;}
                                        break;
                              case 6  : if (!Fixed.SyncPulse1)
                                        { BSyncPul1.BPen = LBLACK; ClearButton (&CWindow, &BSyncPul1); BSyncPul1.BPen = DWHITE;}
                                        break;
                              case 7  : if (!Fixed.SyncPulse2)
                                        { BSyncPul2.BPen = LBLACK; ClearButton (&CWindow, &BSyncPul2); BSyncPul2.BPen = DWHITE;}
                                        break;
                              case 8  : if (!Fixed.ZeroBitPulse)
                                        { BZeroBit.BPen = LBLACK; ClearButton (&CWindow, &BZeroBit); BZeroBit.BPen = DWHITE;}
                                        break;
                              case 9  : if (!Fixed.OneBitPulse)
                                        { BOneBit.BPen = LBLACK; ClearButton (&CWindow, &BOneBit); BOneBit.BPen = DWHITE;}
                                        break;
                              case 10 : if (!Fixed.LengthLastByte)
                                        { BLenLast.BPen = LBLACK; ClearButton (&CWindow, &BLenLast); BLenLast.BPen = DWHITE;}
                                        break;
                              case 11 : if (!Fixed.PauseLength)
                                        { BPauseLen.BPen = LBLACK; ClearButton (&CWindow, &BPauseLen); BPauseLen.BPen = DWHITE;}
                                        break;
                            }
                          break;
    }
    if (GlobalKey == K_ESC)  { Done = TRUE; Cancelled = TRUE; }
    MouseStatus ();
  }
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
  if (Cancelled)
    return (FALSE);
  Changed = (memcmp (&CompareEntry, &EditEntry, sizeof (struct TapeEntry_s)) ||                           /* Was there a change ? */
             memcmp (&Fixed, &OldFixed, sizeof (struct _Fixed_s)));
  if (!Changed)
    return (FALSE);
  DirP = TapeWindow[WhichWindow].Directory;
  while (DirP)
  {                                                                                                       /* Do all entries again */
    if ((DirP->ComboSelected & SELECTBIT) && (DirP->BlockTypeID == 0x10 || DirP->BlockTypeID == 0x11))
    {                                                                                     /* Update the globally changed field(s) */
      if (DirP->BlockTypeID == 0x10 && Fixed.BlockType && EditEntry.BlockTypeID == 0x11)       /* Change Normal to Custom Speed ? */
      {                                                                    /* First set normal speed values in custom speed block */
        DirP->Timing.LengthPilotPulse = NS_LENPILOT;                           /* (To ensure non-changed values are correct too!) */
        DirP->Timing.NumPilotPulses = DirP->Flags.IsHeader ? NS_HDPILOT : NS_DTPILOT;
        DirP->Timing.LengthSyncPulse[0] = NS_LENSYNC1;
        DirP->Timing.LengthSyncPulse[1] = NS_LENSYNC2;
        DirP->Timing.LengthZeroBitPulse = NS_LEN0BIT;
        DirP->Timing.LengthOneBitPulse = NS_LEN1BIT;
      }
      if (Fixed.BlockType)       DirP->BlockTypeID = EditEntry.BlockTypeID;
      if (Fixed.PilotPulse)      DirP->Timing.LengthPilotPulse = EditEntry.Timing.LengthPilotPulse;
      if (Fixed.PilotLength)     DirP->Timing.NumPilotPulses = EditEntry.Timing.NumPilotPulses;
      if (Fixed.SyncPulse1)      DirP->Timing.LengthSyncPulse[0] = EditEntry.Timing.LengthSyncPulse[0];
      if (Fixed.SyncPulse2)      DirP->Timing.LengthSyncPulse[1] = EditEntry.Timing.LengthSyncPulse[1];
      if (Fixed.ZeroBitPulse)    DirP->Timing.LengthZeroBitPulse = EditEntry.Timing.LengthZeroBitPulse;
      if (Fixed.OneBitPulse)     DirP->Timing.LengthOneBitPulse = EditEntry.Timing.LengthOneBitPulse;
      if (Fixed.LengthLastByte)
      {
        if (DirP->BlockTypeID == 0x14 && DirP->Timing.NumBitsInLastByte != 8 && EditEntry.Timing.NumBitsInLastByte == 8)
          DirP->BlockLength --;                                                                     /* Strip from Pure Data block */
        DirP->Timing.NumBitsInLastByte = EditEntry.Timing.NumBitsInLastByte;
      }
      if (Fixed.PauseLength)     DirP->PauseLength = EditEntry.PauseLength;
      if (DirP->BlockTypeID == 0x10 ||                                                                 /* Is (now) Normal Speed ? */
          (DirP->Timing.LengthPilotPulse == NS_LENPILOT &&
           (DirP->Timing.NumPilotPulses == NS_HDPILOT || DirP->Timing.NumPilotPulses == NS_DTPILOT) &&
           DirP->Timing.LengthSyncPulse[0] == NS_LENSYNC1 &&
           DirP->Timing.LengthSyncPulse[1] == NS_LENSYNC2 &&
           DirP->Timing.LengthZeroBitPulse == NS_LEN0BIT &&
           DirP->Timing.LengthOneBitPulse == NS_LEN1BIT))
        DirP->Flags.IsNormalSpeed = TRUE;
      else
        DirP->Flags.IsNormalSpeed = FALSE;
      if (DirP->BlockLength == 17 && DirP->FlagByte == 0x00)                                               /* Is (now) a header ? */
        DirP->Flags.IsHeader = TRUE;
      else
        DirP->Flags.IsHeader = FALSE;
    }
    DirP = DirP->Next;
  }
  return (TRUE);
}

bool EditTapeDataBlock (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : The entry has been adjusted by the user and saved in the tape directory.                                               */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct WindowInfo CWindow   = {193, 0, 256, 0, 256, 0, 256, 0, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                 WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, NULL};
  struct ButtonInfo PCancel   = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 63, 0, 8, DBLACK, DWHITE, NULL, 1, "@Cancel", NULL};
  struct ButtonInfo POk       = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 137, 0, 8, DBLACK, DWHITE, NULL, 0, "@Ok", NULL};
  struct ButtonInfo BBlType   = {BT_TEXT, BC_TEXT, BF_FWIDTH | BF_FITTXT, 112, 0, 20, DBLACK, DWHITE, NULL, 0, NULL, NULL};
  struct ButtonInfo BBlTypSel = {BT_PUSH, BC_GADGET, BF_FWIDTH | BF_KEYSELECT, 239, 0, 1, DBLACK, DWHITE, NULL, 2, "\007", NULL};
  struct ButtonInfo BFlagByte = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 112, 0, 3, DBLACK, DWHITE, NULL, 13,
                                 NULL, NULL};
  struct ButtonInfo BParByte  = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 112, 0, 3, DBLACK, DWHITE, NULL, 3,
                                 NULL, NULL};
  struct ButtonInfo BPilotPul = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 112, 0, 5, DBLACK, DWHITE, NULL, 4,
                                 NULL, NULL};
  struct ButtonInfo BPilotLen = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 112, 0, 5, DBLACK, DWHITE, NULL, 5,
                                 NULL, NULL};
  struct ButtonInfo BSyncPul1 = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 112, 0, 5, DBLACK, DWHITE, NULL, 6,
                                 NULL, NULL};
  struct ButtonInfo BSyncPul2 = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 112, 0, 5, DBLACK, DWHITE, NULL, 7,
                                 NULL, NULL};
  struct ButtonInfo BZeroBit  = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 112, 0, 5, DBLACK, DWHITE, NULL, 8,
                                 NULL, NULL};
  struct ButtonInfo BOneBit   = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 112, 0, 5, DBLACK, DWHITE, NULL, 9,
                                 NULL, NULL};
  struct ButtonInfo BLenLast  = {BT_TEXT, BC_NUMBER, BF_FWIDTH | BF_FITTXT, 112, 0, 1, DBLACK, DWHITE, NULL, 10,
                                 NULL, NULL};
  struct ButtonInfo BDropEnd  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT | BF_NOSURROUND, 126, 0, 8, DBLACK, DWHITE, NULL, 12,
                                 "@Strip", NULL};
  struct ButtonInfo BDropLst  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT | BF_NOSURROUND, 192, 0, 8, DBLACK, DWHITE, NULL, 14,
                                 "Drop", NULL};
  struct ButtonInfo BPauseLen = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 112, 0, 5, DBLACK, DWHITE, NULL, 11,
                                 NULL, NULL};

  struct ButtonInfo      *ButtonP;
  struct TapeEntry_s      EditEntry;
  char                    WindowName[30];
  char                    ValBlockTypeID[30];
  char                    ValFlagByte[4];
  char                    ValParityByte[4];
  char                    ValPilotPulse[6];
  char                    ValPilotLength[6];
  char                    ValSyncPulse1[6];
  char                    ValSyncPulse2[6];
  char                    ValZeroBitPulse[6];
  char                    ValOneBitPulse[6];
  char                    ValLengthLastByte[2];
  char                    ValPauseLength[6];
  char                   *TmpString;
  bool                    Done      = FALSE;
  bool                    Cancelled = FALSE;
  bool                    Changed   = FALSE;
  short                   NumLines;
  short                   CurrentLine;
  short                   ParityLine;
  byte                    OldBlockTypeID;
  dword                   XMSIndex;
  dword                   DXMSSize;
  long          register  Cnt;

  CWindow.Header = WindowName;
  BBlType.Contents = ValBlockTypeID;
  BFlagByte.Contents = ValFlagByte;
  BParByte.Contents = ValParityByte;
  BPilotPul.Contents = ValPilotPulse;
  BPilotLen.Contents = ValPilotLength;
  BSyncPul1.Contents = ValSyncPulse1;
  BSyncPul2.Contents = ValSyncPulse2;
  BZeroBit.Contents = ValZeroBitPulse;
  BOneBit.Contents = ValOneBitPulse;
  BLenLast.Contents = ValLengthLastByte;
  BPauseLen.Contents = ValPauseLength;
  memcpy (&EditEntry, Entry, sizeof (struct TapeEntry_s));                                        /* Use a copy of the tape entry */
  OldBlockTypeID = Entry->BlockTypeID;
  switch (EditEntry.BlockTypeID)
  {
    case 0x10 : strcpy (WindowName, "Edit Normal Speed Data Block"); break;
    case 0x11 : strcpy (WindowName, "Edit Turbo Speed Data Block"); break;
    case 0x14 : strcpy (WindowName, "Edit Pure Data Block"); break;
  }
  NumLines = 11 - (EditEntry.BlockLength < 2 ? 1 : 0) - (EditEntry.BlockLength < 1 ? 1 : 0);
  GetBlockTypeInfo (EditEntry.BlockTypeID, NULL, NULL, &TmpString, NULL, NULL);
  strcpy (ValBlockTypeID, TmpString);
  sprintf (ValFlagByte, "%u", EditEntry.FlagByte);
  sprintf (ValParityByte, "%u", EditEntry.ParityByte);
  sprintf (ValPilotPulse, "%u", EditEntry.Timing.LengthPilotPulse);
  sprintf (ValPilotLength, "%u", EditEntry.Timing.NumPilotPulses);
  sprintf (ValSyncPulse1, "%u", EditEntry.Timing.LengthSyncPulse[0]);
  sprintf (ValSyncPulse2, "%u", EditEntry.Timing.LengthSyncPulse[1]);
  sprintf (ValZeroBitPulse, "%u", EditEntry.Timing.LengthZeroBitPulse);
  sprintf (ValOneBitPulse, "%u", EditEntry.Timing.LengthOneBitPulse);
  sprintf (ValLengthLastByte, "%u", EditEntry.Timing.NumBitsInLastByte);
  sprintf (ValPauseLength, "%u", EditEntry.PauseLength);
  SelectFont (DefaultFont, FONT_NORMAL);
  CWindow.StartY = (Screen.Height - (2 * NumLines * (CurrentFont->Height + 2))) / 2 - 30;
  CWindow.Height = CWindow.MinHeight = CWindow.MaxHeight = 2 * NumLines * (CurrentFont->Height + 2) + 60;
  PCancel.StartY = POk.StartY = 2 * NumLines * (CurrentFont->Height + 2) + 30;
  AddButton (&CWindow, &POk, TRUE);
  CurrentLine = 1;
  BBlType.StartY = BBlTypSel.StartY = CurrentLine * (CurrentFont->Height + 2);
  AddButton (&CWindow, &BBlType, FALSE);
  AddButton (&CWindow, &BBlTypSel, FALSE);
  CurrentLine += 2;
  if (EditEntry.BlockLength > 0)
  { BFlagByte.StartY = CurrentLine * (CurrentFont->Height + 2); AddButton (&CWindow, &BFlagByte, FALSE);  CurrentLine += 2; }
  if (EditEntry.BlockLength > 1)
  { BParByte.StartY = CurrentLine * (CurrentFont->Height + 2);  AddButton (&CWindow, &BParByte, FALSE);   CurrentLine += 2; }
  BPilotPul.StartY = CurrentLine * (CurrentFont->Height + 2);   AddButton (&CWindow, &BPilotPul, FALSE);  CurrentLine += 2;
  BPilotLen.StartY = CurrentLine * (CurrentFont->Height + 2);   AddButton (&CWindow, &BPilotLen, FALSE);  CurrentLine += 2;
  BSyncPul1.StartY = CurrentLine * (CurrentFont->Height + 2);   AddButton (&CWindow, &BSyncPul1, FALSE);  CurrentLine += 2;
  BSyncPul2.StartY = CurrentLine * (CurrentFont->Height + 2);   AddButton (&CWindow, &BSyncPul2, FALSE);  CurrentLine += 2;
  BZeroBit.StartY = CurrentLine * (CurrentFont->Height + 2);    AddButton (&CWindow, &BZeroBit, FALSE);   CurrentLine += 2;
  BOneBit.StartY = CurrentLine * (CurrentFont->Height + 2);     AddButton (&CWindow, &BOneBit, FALSE);    CurrentLine += 2;
  BLenLast.StartY = CurrentLine * (CurrentFont->Height + 2);    AddButton (&CWindow, &BLenLast, FALSE);
  BDropLst.StartY = CurrentLine * (CurrentFont->Height + 2);    AddButton (&CWindow, &BDropLst, FALSE);
  BDropEnd.StartY = CurrentLine * (CurrentFont->Height + 2);    AddButton (&CWindow, &BDropEnd, FALSE);   CurrentLine += 2;
  BPauseLen.StartY = CurrentLine * (CurrentFont->Height + 2);   AddButton (&CWindow, &BPauseLen, FALSE);  CurrentLine += 2;
  if (EditEntry.BlockTypeID == 0x10)
  {
    BPilotPul.Flags |= BF_GHOSTED;
    BPilotLen.Flags |= BF_GHOSTED;
    BSyncPul1.Flags |= BF_GHOSTED;
    BSyncPul2.Flags |= BF_GHOSTED;
    BZeroBit.Flags |= BF_GHOSTED;
    BOneBit.Flags |= BF_GHOSTED;
  }
  else if (EditEntry.BlockTypeID == 0x14)
  {
    BPilotPul.Flags |= BF_GHOSTED;
    BPilotLen.Flags |= BF_GHOSTED;
    BSyncPul1.Flags |= BF_GHOSTED;
    BSyncPul2.Flags |= BF_GHOSTED;
    BFlagByte.Flags |= BF_GHOSTED;
    BParByte.Flags |= BF_GHOSTED;
  }
  if (EditEntry.Timing.NumBitsInLastByte == 8)                                                         /* No loose end to strip ? */
  {
    BLenLast.Flags |= BF_GHOSTED;
    BDropEnd.Flags |= BF_GHOSTED;
  }
  if (EditEntry.BlockLength > 65535L)                                                       /* Cannot convert to Normal Speed, as */
  {                                                                                       /* such blocks have only 2 length bytes */
    BBlType.Flags |= BF_GHOSTED;
    BBlTypSel.Flags |= BF_GHOSTED;
  }
  AddButton (&CWindow, &PCancel, FALSE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  CurrentLine = 1;
  Wprintf (&CWindow, CurrentLine, 1, "{+1Block type}N");
  CurrentLine += 2;
  if (EditEntry.BlockLength > 0)
  { Wprintf (&CWindow, CurrentLine, 1, "{+1Flag byte}N"); CurrentLine += 2; }
  if (EditEntry.Flags.ParityByteUsed)
  {
    ParityLine = CurrentLine;
    Wprintf (&CWindow, CurrentLine, 1, "{+1Parity byte}N");
    Wprintf (&CWindow, CurrentLine, 25, "   [%u]", EditEntry.CalculatedParity);
    CurrentLine += 2;
  }
  Wprintf (&CWindow, CurrentLine, 1, "{+1PILOT pulse}N");
  Wprintf (&CWindow, CurrentLine, 25, "{+1T}N  [%4u]", NS_LENPILOT);
  CurrentLine += 2;
  Wprintf (&CWindow, CurrentLine, 1, "{+1PILOT length}N");
  Wprintf (&CWindow, CurrentLine, 25, "   [%4u|%4u]", NS_HDPILOT, NS_DTPILOT);
  CurrentLine += 2;
  Wprintf (&CWindow, CurrentLine, 1, "{+1SYNC pulse 1}N");
  Wprintf (&CWindow, CurrentLine, 25, "{+1T}N  [%4u]", NS_LENSYNC1);
  CurrentLine += 2;
  Wprintf (&CWindow, CurrentLine, 1, "{+1SYNC pulse 2}N");
  Wprintf (&CWindow, CurrentLine, 25, "{+1T}N  [%4u]", NS_LENSYNC2);
  CurrentLine += 2;
  Wprintf (&CWindow, CurrentLine, 1, "{+1ZERO bit pulse}N");
  Wprintf (&CWindow, CurrentLine, 25, "{+1T}N  [%4u]", NS_LEN0BIT);
  CurrentLine += 2;
  Wprintf (&CWindow, CurrentLine, 1, "{+1ONE bit pulse}N");
  Wprintf (&CWindow, CurrentLine, 25, "{+1T}N  [%4u]", NS_LEN1BIT);
  CurrentLine += 2;
  Wprintf (&CWindow, CurrentLine, 1, "{+1Last byte length}N");
  CurrentLine += 2;
  Wprintf (&CWindow, CurrentLine, 1, "{+1Pause length}N");
  Wprintf (&CWindow, CurrentLine, 25, "{+1ms.}N");
  while (!Done)
  {
    switch (HandleWindow (&CWindow, &ButtonP))
    {
      case WINH_CLOSE   : Done = TRUE; Cancelled = TRUE; break;
      case WINH_RNOHAND : switch (ButtonP->ReactNum)
                          {
                            case 0  : Done = TRUE; Cancelled = FALSE; break;                                              /* (Ok) */
                            case 1  : Done = TRUE; Cancelled = TRUE; break;                                           /* (Cancel) */
                            case 2  : switch (EditEntry.BlockTypeID)                                     /* Toggle the block type */
                                      {
                                        case 0x10 : EditEntry.BlockTypeID = 0x11; break;
                                        case 0x11 : EditEntry.BlockTypeID = 0x14; break;
                                        case 0x14 : if (EditEntry.BlockLength > 65537L)            /* Type 0x10 cannot handle it! */
                                                      EditEntry.BlockTypeID = 0x11;            /* (65537 = 65535 + flag + parity) */
                                                    else
                                                      EditEntry.BlockTypeID = 0x10;
                                                    break;
                                      }
                                      if (EditEntry.BlockTypeID == 0x10)
                                      {
                                        BFlagByte.Flags &= ~BF_GHOSTED;
                                        BParByte.Flags &= ~BF_GHOSTED;
                                        BPilotPul.Flags |= BF_GHOSTED;
                                        BPilotLen.Flags |= BF_GHOSTED;
                                        BSyncPul1.Flags |= BF_GHOSTED;
                                        BSyncPul2.Flags |= BF_GHOSTED;
                                        BZeroBit.Flags |= BF_GHOSTED;
                                        BOneBit.Flags |= BF_GHOSTED;
                                      }
                                      else if (EditEntry.BlockTypeID == 0x11)
                                      {
                                        BFlagByte.Flags &= ~BF_GHOSTED;
                                        BParByte.Flags &= ~BF_GHOSTED;
                                        BPilotPul.Flags &= ~BF_GHOSTED;
                                        BPilotLen.Flags &= ~BF_GHOSTED;
                                        BSyncPul1.Flags &= ~BF_GHOSTED;
                                        BSyncPul2.Flags &= ~BF_GHOSTED;
                                        BZeroBit.Flags &= ~BF_GHOSTED;
                                        BOneBit.Flags &= ~BF_GHOSTED;
                                      }
                                      else
                                      {
                                        BFlagByte.Flags |= BF_GHOSTED;
                                        BParByte.Flags |= BF_GHOSTED;
                                        BPilotPul.Flags |= BF_GHOSTED;
                                        BPilotLen.Flags |= BF_GHOSTED;
                                        BSyncPul1.Flags |= BF_GHOSTED;
                                        BSyncPul2.Flags |= BF_GHOSTED;
                                        BZeroBit.Flags &= ~BF_GHOSTED;
                                        BOneBit.Flags &= ~BF_GHOSTED;
                                      }
                                      GetBlockTypeInfo (EditEntry.BlockTypeID, NULL, NULL, &TmpString, NULL, NULL);
                                      strcpy (ValBlockTypeID, TmpString);
                                      if (EditEntry.BlockLength > 0)
                                        UpdateButtonText (&CWindow, &BFlagByte, BF_FITTXT);
                                      if (EditEntry.BlockLength > 1)
                                        UpdateButtonText (&CWindow, &BParByte, BF_FITTXT);
                                      UpdateButtonText (&CWindow, &BBlType, BF_FITTXT);
                                      UpdateButtonText (&CWindow, &BPilotPul, BF_FITTXT);
                                      UpdateButtonText (&CWindow, &BPilotLen, BF_FITTXT);
                                      UpdateButtonText (&CWindow, &BSyncPul1, BF_FITTXT);
                                      UpdateButtonText (&CWindow, &BSyncPul2, BF_FITTXT);
                                      UpdateButtonText (&CWindow, &BZeroBit, BF_FITTXT);
                                      UpdateButtonText (&CWindow, &BOneBit, BF_FITTXT);
                                      break;
                            case 12 : if (EditEntry.BlockTypeID == 0x14)                                               /* (Strip) */
                                        EditEntry.BlockLength --;                                   /* Strip from Pure Data block */
                                      EditEntry.Timing.NumBitsInLastByte = 8;
                                      sprintf (ValLengthLastByte, "%u", EditEntry.Timing.NumBitsInLastByte);
                                      BDropEnd.Flags |= BF_GHOSTED;                                      /* Can be done only once */
                                      BLenLast.Flags |= BF_GHOSTED;
                                      UpdateButtonText (&CWindow, &BDropEnd, 0);
                                      UpdateButtonText (&CWindow, &BLenLast, BF_FITTXT);
                                      break;
                            case 14 : if (EditEntry.BlockTypeID == 0x14)                                      /* (Drop last byte) */
                                      {
                                        if (EditEntry.BlockLength > 0)
                                          EditEntry.BlockLength --;                       /* Strip last byte from Pure Data Block */
                                        if (EditEntry.Timing.NumBitsInLastByte < 8)                     /* Incomplete last byte ? */
                                        {                                                  /* Perform the 'Strip' button's action */
                                          EditEntry.Timing.NumBitsInLastByte = 8;
                                          sprintf (ValLengthLastByte, "%u", EditEntry.Timing.NumBitsInLastByte);
                                        }
                                        else
                                        {
                                          EditEntry.CalculatedParity ^= EditEntry.ParityByte;
                                          XMSIndex = GetTapeBlockXMSIndex (WhichWindow, Entry);
                                          DXMSSize = (EditEntry.BlockLength & 0xFFFFFFFC) + (EditEntry.BlockLength % 4 ? 4 : 0);
                                          CopyFromXMS (4, TapeWindow[WhichWindow].Tape.XMSHandle,
                                                       XMSIndex + DXMSSize - 4, WorkBuffer);
                                          EditEntry.ParityByte = *(WorkBuffer + EditEntry.BlockLength + 3 - DXMSSize);
                                        }
                                      }
                                      else
                                      {
                                        if (EditEntry.Timing.NumBitsInLastByte < 8)
                                        {
                                          EditEntry.Timing.NumBitsInLastByte = 8;
                                          sprintf (ValLengthLastByte, "%u", EditEntry.Timing.NumBitsInLastByte);
                                        }
                                        else
                                          if (EditEntry.BlockLength > 0)
                                          {
                                            XMSIndex = GetTapeBlockXMSIndex (WhichWindow, Entry);
                                            DXMSSize = (EditEntry.BlockLength & 0xFFFFFFFC) + (EditEntry.BlockLength % 4 ? 4 : 0);
                                            CopyFromXMS (4, TapeWindow[WhichWindow].Tape.XMSHandle,
                                                         XMSIndex + DXMSSize - 4, WorkBuffer);
                                            /* Move the last byte (which we now strip) into the parity byte */
                                            EditEntry.CalculatedParity = *(WorkBuffer + EditEntry.BlockLength + 3 - DXMSSize);
                                            EditEntry.BlockLength --;
                                          }
                                          else                                                            /* (Nothing to drop...) */
                                            SoundErrorBeeper (BEEP_SHORT);
                                      }
                                      BDropEnd.Flags |= BF_GHOSTED;                                      /* Can be done only once */
                                      BLenLast.Flags |= BF_GHOSTED;
                                      BDropLst.Flags |= BF_GHOSTED;
                                      UpdateButtonText (&CWindow, &BDropEnd, 0);
                                      UpdateButtonText (&CWindow, &BLenLast, BF_FITTXT);
                                      UpdateButtonText (&CWindow, &BDropLst, 0);
                                      Wprintf (&CWindow, ParityLine, 28, "[%u]  ", EditEntry.CalculatedParity);
                                      break;
                          }
                          break;
      case WINH_EDITEND : if (ButtonP->Flags & BF_FITINPUT)                                            /* Contents were changed ? */
                            switch (ButtonP->ReactNum)
                            {
                              case 3  : EditEntry.ParityByte = atoi (ValParityByte);
                                        sprintf (ValParityByte, "%u", EditEntry.ParityByte);
                                        UpdateButtonText (&CWindow, &BParByte, BF_FITTXT);
                                        break;
                              case 4  : EditEntry.Timing.LengthPilotPulse = atoi (ValPilotPulse);
                                        sprintf (ValPilotPulse, "%u", EditEntry.Timing.LengthPilotPulse);
                                        UpdateButtonText (&CWindow, &BPilotPul, BF_FITTXT);
                                        break;
                              case 5  : EditEntry.Timing.NumPilotPulses = atoi (ValPilotLength);
                                        sprintf (ValPilotLength, "%u", EditEntry.Timing.NumPilotPulses);
                                        UpdateButtonText (&CWindow, &BPilotLen, BF_FITTXT);
                                        break;
                              case 6  : EditEntry.Timing.LengthSyncPulse[0] = atoi (ValSyncPulse1);
                                        sprintf (ValSyncPulse1, "%u", EditEntry.Timing.LengthSyncPulse[0]);
                                        UpdateButtonText (&CWindow, &BSyncPul1, BF_FITTXT);
                                        break;
                              case 7  : EditEntry.Timing.LengthSyncPulse[1] = atoi (ValSyncPulse2);
                                        sprintf (ValSyncPulse2, "%u", EditEntry.Timing.LengthSyncPulse[1]);
                                        UpdateButtonText (&CWindow, &BSyncPul2, BF_FITTXT);
                                        break;
                              case 8  : EditEntry.Timing.LengthZeroBitPulse = atoi (ValZeroBitPulse);
                                        sprintf (ValZeroBitPulse, "%u", EditEntry.Timing.LengthZeroBitPulse);
                                        UpdateButtonText (&CWindow, &BZeroBit, BF_FITTXT);
                                        break;
                              case 9  : EditEntry.Timing.LengthOneBitPulse = atoi (ValOneBitPulse);
                                        sprintf (ValOneBitPulse, "%u", EditEntry.Timing.LengthOneBitPulse);
                                        UpdateButtonText (&CWindow, &BOneBit, BF_FITTXT);
                                        break;
                              case 10 : EditEntry.Timing.NumBitsInLastByte = atoi (ValLengthLastByte);
                                        sprintf (ValLengthLastByte, "%u", EditEntry.Timing.NumBitsInLastByte);
                                        UpdateButtonText (&CWindow, &BLenLast, BF_FITTXT);
                                        break;
                              case 11 : EditEntry.PauseLength = atoi (ValPauseLength);
                                        sprintf (ValPauseLength, "%u", EditEntry.PauseLength);
                                        UpdateButtonText (&CWindow, &BPauseLen, BF_FITTXT);
                                        break;
                              case 13 : EditEntry.CalculatedParity ^= EditEntry.FlagByte;
                                        EditEntry.FlagByte = atoi (ValFlagByte);
                                        EditEntry.CalculatedParity ^= EditEntry.FlagByte;             /* Update parity match too! */
                                        sprintf (ValFlagByte, "%u", EditEntry.FlagByte);
                                        UpdateButtonText (&CWindow, &BFlagByte, BF_FITTXT);
                                        Wprintf (&CWindow, ParityLine, 28, "[%u]  ", EditEntry.CalculatedParity);
                                        break;
                            }
                          break;
    }
    if (GlobalKey == K_ESC)  { Done = TRUE; Cancelled = TRUE; }
    MouseStatus ();
  }
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
  if (Cancelled)
    return (FALSE);
  Changed = memcmp (Entry, &EditEntry, sizeof (struct TapeEntry_s));                                      /* Was there a change ? */
  if (Changed)
  {
    memcpy (Entry, &EditEntry, sizeof (struct TapeEntry_s));                                                  /* Save the changes */
    if (Entry->BlockTypeID == 0x10 ||                                                                  /* Is (now) Normal Speed ? */
        (Entry->Timing.LengthPilotPulse == NS_LENPILOT &&
         (Entry->Timing.NumPilotPulses == NS_HDPILOT || Entry->Timing.NumPilotPulses == NS_DTPILOT) &&
         Entry->Timing.LengthSyncPulse[0] == NS_LENSYNC1 &&
         Entry->Timing.LengthSyncPulse[1] == NS_LENSYNC2 &&
         Entry->Timing.LengthZeroBitPulse == NS_LEN0BIT &&
         Entry->Timing.LengthOneBitPulse == NS_LEN1BIT))
      Entry->Flags.IsNormalSpeed = TRUE;
    else
      Entry->Flags.IsNormalSpeed = FALSE;
    if (OldBlockTypeID == 0x14 && Entry->BlockTypeID != 0x14)                                         /* Was 'Pure Data' before ? */
    {                                                                          /* Then we need to adjust the flag/parity behavior */
      if (Entry->BlockLength > 0)
      {
        if (Entry->BlockLength < 49152L)                                                            /* Can be handled in one go ? */
        {
          XMSIndex = GetTapeBlockXMSIndex (WhichWindow, Entry);
          CopyFromXMS (Entry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
          for (Cnt = 1 ; Cnt < (word)Entry->BlockLength ; Cnt ++)
            *(WorkBuffer + Cnt - 1) = *(WorkBuffer + Cnt);                                         /* Shift the block 1 byte back */
          CopyToXMS (Entry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, XMSIndex);
        }
        Entry->Flags.FlagByteUsed = TRUE;
        Entry->BlockLength --;
      }
      if (Entry->BlockLength > 1)
      {
        Entry->Flags.ParityByteUsed = TRUE;
        Entry->BlockLength --;
      }
    }
    else if (OldBlockTypeID != 0x14 && Entry->BlockTypeID == 0x14)                                        /* Is now 'Pure Data' ? */
    {
      if (Entry->Flags.FlagByteUsed)
      {
        Entry->Flags.FlagByteUsed = FALSE;
        Entry->BlockLength ++;
        if (Entry->BlockLength < 49152L)                                                            /* Can be handled in one go ? */
        {
          XMSIndex = GetTapeBlockXMSIndex (WhichWindow, Entry);
          CopyFromXMS (Entry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);
          for (Cnt = Entry->BlockLength - 2 ; Cnt >= 0 ; Cnt --)
            *(WorkBuffer + Cnt + 1) = *(WorkBuffer + Cnt);                                      /* Shift the block 1 byte forward */
          *WorkBuffer = Entry->FlagByte;
          if (Entry->Flags.ParityByteUsed)
          {
            Entry->Flags.ParityByteUsed = FALSE;
            *(WorkBuffer + Entry->BlockLength) = Entry->ParityByte;
            Entry->BlockLength ++;
          }
          CopyToXMS (Entry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, XMSIndex);
        }
      }
    }
    if ((Entry->BlockTypeID == 0x10 || Entry->BlockTypeID == 0x11) && Entry->BlockLength == 17 && Entry->FlagByte == 0x00)
      Entry->Flags.IsHeader = TRUE;
    else
      Entry->Flags.IsHeader = FALSE;
  }
  return (Changed);
}

bool EditPureToneBlock (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : The entry has been adjusted by the user and saved in the tape directory.                                               */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct WindowInfo CWindow   = {193, 225, 256, 90, 256, 90, 256, 90, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                 WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "Edit Pure Tone Block"};
  struct ButtonInfo PCancel   = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 63, 60, 8, DBLACK, DWHITE, NULL, 1, "@Cancel", NULL};
  struct ButtonInfo POk       = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 137, 60, 8, DBLACK, DWHITE, NULL, 0, "@Ok", NULL};
  struct ButtonInfo BPilotPul = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 112, 10, 5, DBLACK, DWHITE, NULL, 2,
                                 NULL, NULL};
  struct ButtonInfo BPilotLen = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 112, 30, 5, DBLACK, DWHITE, NULL, 3,
                                 NULL, NULL};
  struct ButtonInfo *ButtonP;
  char              ValPilotPulse[6];
  char              ValPilotLength[6];
  word              ELengthPilotPulse;
  word              ENumPilotPulses;
  bool              Done      = FALSE;
  bool              Cancelled = FALSE;
  bool              Changed   = FALSE;

  BPilotPul.Contents = ValPilotPulse;
  BPilotLen.Contents = ValPilotLength;
  ELengthPilotPulse = Entry->Timing.LengthPilotPulse;
  ENumPilotPulses = Entry->Timing.NumPilotPulses;
  sprintf (ValPilotPulse, "%u", ELengthPilotPulse);
  sprintf (ValPilotLength, "%u", ENumPilotPulses);
  AddButton (&CWindow, &POk, TRUE);
  AddButton (&CWindow, &BPilotPul, FALSE);
  AddButton (&CWindow, &BPilotLen, FALSE);
  AddButton (&CWindow, &PCancel, FALSE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  Wprintf (&CWindow, 1, 1, "{+1Pulse length}N");
  Wprintf (&CWindow, 1, 25, "{+1T}N");
  Wprintf (&CWindow, 3, 1, "{+1Tone length}N");
  Wprintf (&CWindow, 3, 25, "{+1pulses}N");
  while (!Done)
  {
    switch (HandleWindow (&CWindow, &ButtonP))
    {
      case WINH_CLOSE   : Done = TRUE; Cancelled = TRUE; break;
      case WINH_RNOHAND : switch (ButtonP->ReactNum)
                          {
                            case 0  : Done = TRUE; Cancelled = FALSE; break;                                              /* (Ok) */
                            case 1  : Done = TRUE; Cancelled = TRUE; break;                                           /* (Cancel) */
                          }
                          break;
      case WINH_EDITEND : if (ButtonP->Flags & BF_FITINPUT)                                            /* Contents were changed ? */
                            switch (ButtonP->ReactNum)
                            {
                              case 2  : ELengthPilotPulse = atol (ValPilotPulse);
                                        sprintf (ValPilotPulse, "%u", ELengthPilotPulse);
                                        UpdateButtonText (&CWindow, &BPilotPul, BF_FITTXT);
                                        break;
                              case 3  : ENumPilotPulses = atol (ValPilotLength);
                                        sprintf (ValPilotLength, "%u", ENumPilotPulses);
                                        UpdateButtonText (&CWindow, &BPilotLen, BF_FITTXT);
                                        break;
                            }
                          break;
    }
    if (GlobalKey == K_ESC)  { Done = TRUE; Cancelled = TRUE; }
    MouseStatus ();
  }
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
  Changed = (ELengthPilotPulse != Entry->Timing.LengthPilotPulse || ENumPilotPulses != Entry->Timing.NumPilotPulses);
  if (Cancelled || !Changed)
  {
    if (Entry->Timing.LengthPilotPulse == 0 && Entry->Timing.NumPilotPulses == 0)                   /* Came from insert handler ? */
      DeleteTapeBlock (WhichWindow, Entry, FALSE, FALSE);                         /* Then we don't have to keep the entry anymore */
    _Cancelled = TRUE;
    return (FALSE);
  }
  Entry->Timing.LengthPilotPulse = ELengthPilotPulse;
  Entry->Timing.NumPilotPulses = ENumPilotPulses;
  return (TRUE);
}

bool EditRowOfPulsesBlock (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : The entry has been adjusted by the user and saved in the tape directory.                                               */
/*         Note: when called from ViewTapeBlock, there is at least 1 pulse left on exit. When called from InsertTapeEntry, if the */
/*         user hit 'Cancel' without inserting any pulses, the entry has been removed from the tape again.                        */
/* Import: ResizeTapeBlock.                                                                                                       */
/**********************************************************************************************************************************/

{
  struct WindowInfo CWindow  = {230, 150, 175, 211, 175, 211, 175, 211, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "Edit Row Of Pulses Block"};
  struct ButtonInfo PCancel  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 23, 185, 8, DBLACK, DWHITE, NULL, 0, "@Cancel", NULL};
  struct ButtonInfo POk      = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 102, 185, 8, DBLACK, DWHITE, NULL, 1, "@Ok", NULL};
  struct ButtonInfo PInsert  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 107, 40, 8, DBLACK, DWHITE, NULL, 2, "@Insert", NULL};
  struct ButtonInfo PAppend  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 107, 60, 8, DBLACK, DWHITE, NULL, 3, "@Append", NULL};
  struct ButtonInfo PRemove  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 107, 80, 8, DBLACK, DWHITE, NULL, 4, "@Remove", NULL};

  struct ComboInfo        TWindow;
  struct ComboEntry       TEntry;
  struct ButtonInfo      *ButtonP;
  comboent                SelectedLine;
  dword                   XMSIndex;
  short         register  Cnt;
  word                    Pulse;
  bool                    Done = FALSE;
  bool                    Cancelled = FALSE;
  bool                    Accepted = FALSE;
  short                   NumPulses;
  short                   InitialNumPulses;
  char                    StrPulse[6];
  char                    WHeader[30];
  word                    ValPulse;

  InitialNumPulses = NumPulses = Entry->BlockLength / 2;
  if (NumPulses == 0)                                                                   /* Empty list (just inserted the entry) ? */
  {
    PRemove.Flags |= BF_GHOSTED;
    POk.Flags |= BF_GHOSTED;
  }
  AddButton (&CWindow, &POk, FALSE);
  AddButton (&CWindow, &PInsert, FALSE);
  AddButton (&CWindow, &PAppend, FALSE);
  AddButton (&CWindow, &PRemove, FALSE);
  AddButton (&CWindow, &PCancel, TRUE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  Wprintf (&CWindow, 1, 2, "{+1Number of pulses}N");
  Wprintf (&CWindow, 1, 20, "%u ", NumPulses);
  MakeComboWindow (&TWindow, &CWindow, COMT_SELECT, COMF_NOBUFFER | COMF_VSLIDER | COMF_SYNCSLIDER | COMF_KEYSELECT,
                   CWindow.StartX + 12, CWindow.StartY + 38, 0, 0, 12, 10, 12, NULL, FALSE, 0, NULL, DBLACK, DWHITE, 0, 0);
  TEntry.Selected = 0x00;
  if (Entry->AllocLength > 0)
  {
    XMSIndex = GetTapeBlockXMSIndex (WhichWindow, Entry);
    CopyFromXMS (Entry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);              /* Fetch old row */
    for (Cnt = 0 ; Cnt < NumPulses ; Cnt ++)                                           /* Insert all pulses into the combo window */
    {
      Pulse = *((word *)WorkBuffer + Cnt);
      TEntry.Identifier = (dword)Pulse;
      sprintf (TEntry.Text, "%u", Pulse);
      if (!InsertComboEntry (&TWindow, &TEntry, NULL, -1))
        FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
    }
  }
  PrintComboEntries (&TWindow);
  SelectEntry (&TWindow, 0);
  while (!Done)
  {
    switch (HandleWindow (&CWindow, &ButtonP))
    {
      case WINH_CLOSE   : Done = TRUE; Cancelled = TRUE; break;
      case WINH_RNOHAND : switch (ButtonP->ReactNum)
                          {
                            case 0 : Done = TRUE; Cancelled = TRUE; break;                                            /* (Cancel) */
                            case 1 : Done = TRUE; Cancelled = FALSE; break;                                               /* (Ok) */
                            case 2 :                                                                                  /* (Insert) */
                            case 3 : Accepted = FALSE;                                                                /* (Append) */
                                     StrPulse[0] = '\0';
                                     sprintf (WHeader, "%s Pulse Value", ButtonP->ReactNum == 2 ? "Insert" : "Append");
                                     while (!Accepted)
                                     {
                                       if (InputRequester (-1, -1, NULL, NULL, StrPulse, 6, FALSE, FALSE, WHeader,
                                                           "{CGive the new pulse length in T states}C"))
                                         if (StrPulse[0] != '\0' && atol (StrPulse) < 65536L)
                                         {
                                           Accepted = TRUE;
                                           ValPulse = (word)atol (StrPulse);
                                           TEntry.Identifier = (dword)ValPulse;
                                           sprintf (TEntry.Text, "%u", ValPulse);
                                           SelectedLine = TWindow.LastSelected + (ButtonP->ReactNum == 2 ? 0 : 1);
                                           if (!InsertComboEntry (&TWindow, &TEntry, NULL, SelectedLine))
                                             FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                                           SelectEntry (&TWindow, SelectedLine);
                                           NumPulses ++;
                                           Wprintf (&CWindow, 1, 20, "%u ", NumPulses);
                                           if (NumPulses == 1)                                       /* First pulse in the list ? */
                                           {
                                             PRemove.Flags &= ~BF_GHOSTED;
                                             POk.Flags &= ~BF_GHOSTED;
                                             UpdateButtonText (&CWindow, &PRemove, 0);
                                             UpdateButtonText (&CWindow, &POk, 0);
                                             SelectEntry (&TWindow, 0);
                                           }
                                           else if (NumPulses == 255)                                          /* List full now ? */
                                           {
                                             PInsert.Flags |= BF_GHOSTED;
                                             PAppend.Flags |= BF_GHOSTED;
                                             UpdateButtonText (&CWindow, &PInsert, 0);
                                             UpdateButtonText (&CWindow, &PAppend, 0);
                                           }
                                         }
                                         else
                                           ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - The value is out of range\n"
                                                           "Use values 0 through 65535");
                                       else
                                         Accepted = TRUE;
                                     }
                                     break;
                            case 4 : if (TWindow.LastSelected >= 0 && TWindow.LastSelected <= TWindow.NumEntries)     /* (Remove) */
                                     {
                                       DeleteComboEntry (&TWindow, TWindow.LastSelected);
                                       if (TWindow.LastSelected > TWindow.NumEntries)                     /* Removed last entry ? */
                                         TWindow.LastSelected = TWindow.NumEntries;
                                       SelectEntry (&TWindow, TWindow.LastSelected);
                                       NumPulses --;
                                       Wprintf (&CWindow, 1, 20, "%u ", NumPulses);
                                       if (NumPulses == 0)                                        /* No more pulses in the list ? */
                                       {
                                         PRemove.Flags |= BF_GHOSTED;
                                         POk.Flags |= BF_GHOSTED;
                                         UpdateButtonText (&CWindow, &PRemove, 0);
                                         UpdateButtonText (&CWindow, &POk, 0);
                                       }
                                       else if (NumPulses == 254)                                /* Insert buttons were ghosted ? */
                                       {
                                         PInsert.Flags &= ~BF_GHOSTED;
                                         PAppend.Flags &= ~BF_GHOSTED;
                                         UpdateButtonText (&CWindow, &PInsert, 0);
                                         UpdateButtonText (&CWindow, &PAppend, 0);
                                       }
                                     }
                                     break;
                          }
                          break;
    }
    switch (HandleComboWindow (&TWindow, &SelectedLine))
    {
      case WINH_LINESEL : Accepted = FALSE;
                          GetComboEntry (&TWindow, &TEntry, TWindow.LastSelected);
                          strcpy (StrPulse, TEntry.Text);
                          while (!Accepted)
                          {
                            if (InputRequester (-1, -1, NULL, NULL, StrPulse, 6, FALSE, FALSE, "Edit Pulse Value",
                                                "{CGive the new pulse length in T states}C"))
                              if (StrPulse[0] != '\0' && atol (StrPulse) < 65536L)
                              {
                                Accepted = TRUE;
                                ValPulse = (word)atol (StrPulse);
                                TEntry.Identifier = (dword)ValPulse;
                                sprintf (TEntry.Text, "%u", ValPulse);
                                PutComboEntry (&TWindow, &TEntry, TWindow.LastSelected, TRUE);
                              }
                              else
                                ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - The value is out of range\n"
                                                "Use values 0 through 65535");
                            else
                              Accepted = TRUE;
                          }
                          break;
    }
    if (GlobalKey == K_ESC)  { Done = TRUE; Cancelled = TRUE; }
    MouseStatus ();
  }
  if (!Cancelled)
  {
    for (Cnt = 0 ; Cnt < NumPulses ; Cnt ++)                                               /* Write all pulses back to the buffer */
    {
      GetComboEntry (&TWindow, &TEntry, Cnt);
      *((word *)WorkBuffer + Cnt) = (word)TEntry.Identifier;
    }
    XMSIndex = ResizeTapeBlock (WhichWindow, Entry, (dword)NumPulses * 2, 0);                  /* Resize the data block if needed */
    CopyToXMS (Entry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, XMSIndex); /* Insert into the virtual tape */
  }
  DestroyComboWindow (&TWindow, 0);
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
  if (Cancelled && InitialNumPulses == 0)                                                     /* No pulses (only after Insert!) ? */
  {
    DeleteTapeBlock (WhichWindow, Entry, FALSE, FALSE);                           /* Then we don't have to keep the entry anymore */
    _Cancelled = TRUE;
  }
  return (!Cancelled);
}

bool EditPauseBlock (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : The entry has been adjusted by the user and saved in the tape directory.                                               */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  char NumBuffer[6];
  bool Accepted    = FALSE;
  word OldValue;

  OldValue = Entry->PauseLength;
  sprintf (NumBuffer, "%u", Entry->PauseLength);
  while (!Accepted)
  {
    if (InputRequester (-1, -1, NULL, NULL, NumBuffer, 6, FALSE, FALSE, "Edit Pause Block",
                        "{CGive the value for the pause length in ms.}C\n{CUse \'0\' for an infinite pause}C"))
      if (NumBuffer[0] != '\0' && atol (NumBuffer) < 65536L)
      {
        Accepted = TRUE;
        Entry->PauseLength = (word)atoi (NumBuffer);
      }
      else
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - The value is out of range\n        Use values 0 through 65535");
    else                                                                                                           /* (Cancelled) */
    {
      if (_EntryNumber != -1)                                                                   /* Came from the insert handler ? */
        DeleteTapeBlock (WhichWindow, Entry, FALSE, FALSE);                       /* Then we don't have to keep the entry anymore */
      _Cancelled = TRUE;
      return (FALSE);
    }
  }
  return ((bool)(Entry->PauseLength != OldValue));
}

bool EditGroupStrtBlock (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : The entry has been adjusted by the user and saved in the tape directory.                                               */
/* Import: ResizeTapeBlock.                                                                                                       */
/**********************************************************************************************************************************/

{
  char  TextBuffer[257];
  bool  Accepted    = FALSE;
  short StrLength;
  dword XMSIndex;
  bool  Changed     = FALSE;

  if (Entry->BlockLength > 0)
  {
    XMSIndex = GetTapeBlockXMSIndex (WhichWindow, Entry);
    CopyFromXMS (Entry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);           /* Fetch old string */
    *(WorkBuffer + Entry->BlockLength) = '\0';                                                               /* Force it a string */
    strcpy (TextBuffer, WorkBuffer);
  }
  else
    TextBuffer[0] = '\0';
  while (!Accepted)
  {
    if (InputRequester (-1, -1, NULL, NULL, TextBuffer, 256, TRUE, FALSE, "Edit Group Start Block",
                        "{CGive the group name}C"))
    {
      Accepted = TRUE;
      if (strcmp (TextBuffer, WorkBuffer))                                                  /* String has actually been changed ? */
      {
        StrLength = strlen (TextBuffer);
        XMSIndex = ResizeTapeBlock (WhichWindow, Entry, (dword)StrLength, 0);                  /* Resize the data block if needed */
        CopyToXMS (Entry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, TextBuffer, XMSIndex);         /* Write new string */
        Changed = TRUE;
      }
    }
    else                                                                                                           /* (Cancelled) */
    {
      if (_EntryNumber != -1)                                                                   /* Came from the insert handler ? */
        DeleteTapeBlock (WhichWindow, Entry, FALSE, FALSE);                       /* Then we don't have to keep the entry anymore */
      _Cancelled = TRUE;
      _MakeGroupCancelled = TRUE;
      return (FALSE);
    }
  }
  return (Changed);
}

bool EditJumpToBlock (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : The entry has been adjusted by the user and saved in the tape directory.                                               */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct ComboEntry   TEntry;
  struct TapeEntry_s *DstDirP;
  comboent            SrcIndex;
  comboent            DstIndex;
  char                NumBuffer[6];
  bool                Accepted    = FALSE;
  word                OldValue;
  bool                StillInUse  = FALSE;

  if (_EntryNumber == -1)                                                                 /* Not coming from the insert handler ? */
  {
    GetComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &TEntry, TapeWindow[WhichWindow].SelectedLine);
    SrcIndex = GetBlockNumber (TEntry.Text);
  }
  else
    SrcIndex = _EntryNumber + 1;                                                                   /* Don't want an infinite loop */
  OldValue = Entry->PauseLength;
  if (SrcIndex + (short)Entry->PauseLength > 0)
    if (SrcIndex + (short)Entry->PauseLength > TapeWindow[WhichWindow].NumEntries)
      sprintf (NumBuffer, "%u", TapeWindow[WhichWindow].NumEntries);                         /* Cannot jump to after the tape end */
    else
      sprintf (NumBuffer, "%u", SrcIndex + (short)Entry->PauseLength);
  else
    strcpy (NumBuffer, "0");                                                              /* Cannot jump to before the tape start */
  while (!Accepted)
  {
    if (InputRequester (-1, -1, NULL, NULL, NumBuffer, 6, FALSE, FALSE, "Edit Jump To Entry Block",
                        "{CGive the value for the destination entry}C"))
      if (NumBuffer[0] != '\0' && atoi (NumBuffer) == SrcIndex)                                            /* Jumping to itself ? */
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Don't be a wally!");
      else if (NumBuffer[0] != '\0' && atoi (NumBuffer) <= TapeWindow[WhichWindow].NumEntries)
      {
        Accepted = TRUE;
        Entry->PauseLength = (word)((short)atoi (NumBuffer) - SrcIndex);
        if (Entry->SrcJumpCode == 0x0000)                                                                 /* Entry just created ? */
        {
          if (++ TapeWindow[WhichWindow].NextJumpCode == 0x0000)                            /* Overflow - more than 65535 jumps ? */
            ;
          Entry->SrcJumpCode = TapeWindow[WhichWindow].NextJumpCode;                                        /* Set source jump ID */
          DstIndex = (short)Entry->PauseLength + SrcIndex;                                         /* Determine destination block */
          DstDirP = TapeWindow[WhichWindow].Directory;
          if (DstIndex > 0)                                                                                     /* (Sanity check) */
            while (-- DstIndex >= 0 && DstDirP != NULL)                                                 /* Find destination entry */
              DstDirP = DstDirP->Next;
          if (DstDirP != NULL)                                                                                  /* (Sanity check) */
            if (DstDirP->DstJumpCode != 0x0000)                                              /* This entry is jumped to already ? */
            {
              Entry->SrcJumpCode = DstDirP->DstJumpCode;                                     /* Use that jump code for the source */
              TapeWindow[WhichWindow].NextJumpCode --;                                               /* Give the 'next code' back */
            }
            else
              DstDirP->DstJumpCode = TapeWindow[WhichWindow].NextJumpCode;                             /* Set destination jump ID */
        }
        else if (Entry->PauseLength != OldValue)                                              /* (Editing an existing Jump entry) */
        {
          DstDirP = TapeWindow[WhichWindow].Directory;
          while (DstDirP != NULL && !StillInUse)                            /* Check if the jump code is also used by other jumps */
          {
            if (DstDirP->SrcJumpCode == Entry->SrcJumpCode && DstDirP != Entry)
              StillInUse = TRUE;
            DstDirP = DstDirP->Next;
          }
          if (!StillInUse)                                                    /* (This is the only jump that uses this jump code) */
          {
            DstIndex = (short)OldValue + SrcIndex;                                             /* Determine old destination block */
            DstDirP = TapeWindow[WhichWindow].Directory;
            if (DstIndex > 0)                                                                                   /* (Sanity check) */
              while (-- DstIndex >= 0 && DstDirP != NULL)                                               /* Find destination entry */
                DstDirP = DstDirP->Next;
            if (DstDirP != NULL)                                                                                /* (Sanity check) */
              DstDirP->DstJumpCode = 0x0000;                                                         /* Clear the old destination */
          }
          DstIndex = (short)Entry->PauseLength + SrcIndex;                                         /* Determine destination block */
          DstDirP = TapeWindow[WhichWindow].Directory;
          if (DstIndex > 0)                                                                                     /* (Sanity check) */
            while (-- DstIndex >= 0 && DstDirP != NULL)                                                 /* Find destination entry */
              DstDirP = DstDirP->Next;
          if (DstDirP != NULL)                                                                                  /* (Sanity check) */
            if (DstDirP->DstJumpCode != 0x0000)                                              /* This entry is jumped to already ? */
              Entry->SrcJumpCode = DstDirP->DstJumpCode;                                     /* Use that jump code for the source */
            else if (!StillInUse)
              DstDirP->DstJumpCode = Entry->SrcJumpCode;                                               /* Set destination jump ID */
            else                                                                                     /* (We need a new jump code) */
            {
              if (++ TapeWindow[WhichWindow].NextJumpCode == 0x0000)                        /* Overflow - more than 65535 jumps ? */
                ;
              Entry->SrcJumpCode = DstDirP->DstJumpCode = TapeWindow[WhichWindow].NextJumpCode;             /* Set source jump ID */
            }
        }
      }
      else
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - The value is out of range\n        Use values 0 through %u",
                        TapeWindow[WhichWindow].NumEntries);
    else                                                                                                           /* (Cancelled) */
    {
      if (_EntryNumber != -1)                                                                   /* Came from the insert handler ? */
        DeleteTapeBlock (WhichWindow, Entry, FALSE, FALSE);                       /* Then we don't have to keep the entry anymore */
      _Cancelled = TRUE;
      return (FALSE);
    }
  }
  return ((bool)(Entry->PauseLength != OldValue));
}

bool EditLoopStrtBlock (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : The entry has been adjusted by the user and saved in the tape directory.                                               */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  char NumBuffer[6];
  bool Accepted    = FALSE;
  word OldValue;

  if (_EntryNumber == -1)
    OldValue = Entry->PauseLength;
  else
    OldValue = 1;
  sprintf (NumBuffer, "%u", Entry->PauseLength);
  while (!Accepted)
  {
    if (InputRequester (-1, -1, NULL, NULL, NumBuffer, 6, FALSE, FALSE, "Edit Loop Start Block",
                        "{CGive the number of repetitions}C"))
      if (NumBuffer[0] != '\0' && atol (NumBuffer) > 1 && atol (NumBuffer) < 65536L)
      {
        Accepted = TRUE;
        Entry->PauseLength = (word)atoi (NumBuffer);
      }
      else
        ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - The value is out of range\n        Use values 2 through 65535");
    else                                                                                                           /* (Cancelled) */
    {
      if (_EntryNumber != -1)                                                                   /* Came from the insert handler ? */
        DeleteTapeBlock (WhichWindow, Entry, FALSE, FALSE);                       /* Then we don't have to keep the entry anymore */
      _Cancelled = TRUE;
      _MakeGroupCancelled = TRUE;
      return (FALSE);
    }
  }
  return ((bool)(Entry->PauseLength != OldValue));
}

bool EditCallSequenceBlock (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : The entry has been adjusted by the user and saved in the tape directory.                                               */
/*         Note: when called from ViewTapeBlock, there is at least 1 call left on exit. When called from InsertTapeEntry, if the  */
/*         user hit 'Cancel' without inserting any calls, the entry has been removed from the tape again.                         */
/* Import: ResizeTapeBlock.                                                                                                       */
/**********************************************************************************************************************************/

{
  struct WindowInfo CWindow  = {230, 150, 175, 211, 175, 211, 175, 211, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "Edit Call Sequence Block"};
  struct ButtonInfo PCancel  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 23, 185, 8, DBLACK, DWHITE, NULL, 0, "@Cancel", NULL};
  struct ButtonInfo POk      = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 102, 185, 8, DBLACK, DWHITE, NULL, 1, "@Ok", NULL};
  struct ButtonInfo PInsert  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 107, 40, 8, DBLACK, DWHITE, NULL, 2, "@Insert", NULL};
  struct ButtonInfo PAppend  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 107, 60, 8, DBLACK, DWHITE, NULL, 3, "@Append", NULL};
  struct ButtonInfo PRemove  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 107, 80, 8, DBLACK, DWHITE, NULL, 4, "@Remove", NULL};

  struct ComboInfo        TWindow;
  struct ComboEntry       TEntry;
  struct ButtonInfo      *ButtonP;
  comboent                SelectedLine;
  dword                   XMSIndex;
  short         register  Cnt;
  comboent                CallValue;
  bool                    Done = FALSE;
  bool                    Cancelled = FALSE;
  bool                    Accepted = FALSE;
  short                   NumCalls;
  short                   InitialNumCalls;
  char                    StrCall[6];
  char                    WHeader[30];
  comboent                ValCall;
  comboent                SrcIndex;                                                      /* (Absolute block number of this block) */

  InitialNumCalls = NumCalls = Entry->BlockLength / 2;
  if (NumCalls == 0)                                                                    /* Empty list (just inserted the entry) ? */
  {
    SetupNewData (WhichWindow);                          /* Insert into the list, so the user can see the resulting block numbers */
                                                                       /* This also makes sure that all jump values are corrected */
    PRemove.Flags |= BF_GHOSTED;
    POk.Flags |= BF_GHOSTED;
    SrcIndex = _EntryNumber;
  }
  else                                                                                    /* Not coming from the insert handler ? */
  {
    GetComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &TEntry, TapeWindow[WhichWindow].SelectedLine);
    SrcIndex = GetBlockNumber (TEntry.Text);
  }
  AddButton (&CWindow, &POk, TRUE);
  AddButton (&CWindow, &PInsert, FALSE);
  AddButton (&CWindow, &PAppend, FALSE);
  AddButton (&CWindow, &PRemove, FALSE);
  AddButton (&CWindow, &PCancel, FALSE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  Wprintf (&CWindow, 1, 2, "{+1Number of calls}N");
  Wprintf (&CWindow, 1, 20, "%u ", NumCalls);
  MakeComboWindow (&TWindow, &CWindow, COMT_SELECT, COMF_NOBUFFER | COMF_VSLIDER | COMF_SYNCSLIDER | COMF_KEYSELECT,
                   CWindow.StartX + 12, CWindow.StartY + 38, 0, 0, 12, 10, 12, NULL, FALSE, 0, NULL, DBLACK, DWHITE, 0, 0);
  TEntry.Selected = 0x00;
  if (Entry->AllocLength > 0)
  {
    XMSIndex = GetTapeBlockXMSIndex (WhichWindow, Entry);
    CopyFromXMS (Entry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);              /* Fetch old row */
    for (Cnt = 0 ; Cnt < NumCalls ; Cnt ++)                                             /* Insert all calls into the combo window */
    {
      CallValue = *((short *)WorkBuffer + Cnt);
      TEntry.Identifier = (dword)CallValue;
      sprintf (TEntry.Text, "%d", SrcIndex + CallValue);                                     /* Make relative jump value absolute */
      if (!InsertComboEntry (&TWindow, &TEntry, NULL, -1))
        FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
    }
  }
  PrintComboEntries (&TWindow);
  SelectEntry (&TWindow, 0);
  while (!Done)
  {
    switch (HandleWindow (&CWindow, &ButtonP))
    {
      case WINH_CLOSE   : Done = TRUE; Cancelled = TRUE; break;
      case WINH_RNOHAND : switch (ButtonP->ReactNum)
                          {
                            case 0 : Done = TRUE; Cancelled = TRUE; break;                                            /* (Cancel) */
                            case 1 : Done = TRUE; Cancelled = FALSE; break;                                               /* (Ok) */
                            case 2 :                                                                                  /* (Insert) */
                            case 3 : Accepted = FALSE;                                                                /* (Append) */
                                     StrCall[0] = '\0';
                                     sprintf (WHeader, "%s Call Value", ButtonP->ReactNum == 2 ? "Insert" : "Append");
                                     while (!Accepted)
                                     {
                                       if (InputRequester (-1, -1, NULL, NULL, StrCall, 6, FALSE, FALSE, WHeader,
                                                           "{CGive the new CALL destination block number}C"))
                                         if (StrCall[0] != '\0' && atoi (StrCall) == SrcIndex)                /* Calling inself ? */
                                           ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Don't be a wally!");
                                         else if (StrCall[0] != '\0' && atol (StrCall) <= TapeWindow[WhichWindow].NumEntries)
                                         {
                                           Accepted = TRUE;
                                           ValCall = (short)atol (StrCall);
                                           CallValue = ValCall - SrcIndex;
                                           TEntry.Identifier = (dword)CallValue;
                                           sprintf (TEntry.Text, "%d", ValCall);
                                           SelectedLine = TWindow.LastSelected + (ButtonP->ReactNum == 2 ? 0 : 1);
                                           if (!InsertComboEntry (&TWindow, &TEntry, NULL, SelectedLine))
                                             FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                                           SelectEntry (&TWindow, SelectedLine);
                                           NumCalls ++;
                                           Wprintf (&CWindow, 1, 20, "%u ", NumCalls);
                                           if (NumCalls == 1)                                         /* First call in the list ? */
                                           {
                                             PRemove.Flags &= ~BF_GHOSTED;
                                             POk.Flags &= ~BF_GHOSTED;
                                             UpdateButtonText (&CWindow, &PRemove, 0);
                                             UpdateButtonText (&CWindow, &POk, 0);
                                             SelectEntry (&TWindow, 0);
                                           }
                                           else if (NumCalls == 255)                                           /* List full now ? */
                                           {
                                             PInsert.Flags |= BF_GHOSTED;
                                             PAppend.Flags |= BF_GHOSTED;
                                             UpdateButtonText (&CWindow, &PInsert, 0);
                                             UpdateButtonText (&CWindow, &PAppend, 0);
                                           }
                                         }
                                         else
                                           ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - The value is out of range\n"
                                                           "Use values 0 through %d", TapeWindow[WhichWindow].NumEntries);
                                       else
                                         Accepted = TRUE;
                                     }
                                     break;
                            case 4 : if (TWindow.LastSelected >= 0 && TWindow.LastSelected <= TWindow.NumEntries)     /* (Remove) */
                                     {
                                       DeleteComboEntry (&TWindow, TWindow.LastSelected);
                                       if (TWindow.LastSelected > TWindow.NumEntries)                     /* Removed last entry ? */
                                         TWindow.LastSelected = TWindow.NumEntries;
                                       SelectEntry (&TWindow, TWindow.LastSelected);
                                       NumCalls --;
                                       Wprintf (&CWindow, 1, 20, "%d ", NumCalls);
                                       if (NumCalls == 0)                                          /* No more calls in the list ? */
                                       {
                                         PRemove.Flags |= BF_GHOSTED;
                                         POk.Flags |= BF_GHOSTED;
                                         UpdateButtonText (&CWindow, &PRemove, 0);
                                         UpdateButtonText (&CWindow, &POk, 0);
                                       }
                                       else if (NumCalls == 254)                                 /* Insert buttons were ghosted ? */
                                       {
                                         PInsert.Flags &= ~BF_GHOSTED;
                                         PAppend.Flags &= ~BF_GHOSTED;
                                         UpdateButtonText (&CWindow, &PInsert, 0);
                                         UpdateButtonText (&CWindow, &PAppend, 0);
                                       }
                                     }
                                     break;
                          }
                          break;
    }
    switch (HandleComboWindow (&TWindow, &SelectedLine))
    {
      case WINH_LINESEL : Accepted = FALSE;
                          GetComboEntry (&TWindow, &TEntry, TWindow.LastSelected);
                          strcpy (StrCall, TEntry.Text);
                          while (!Accepted)
                          {
                            if (InputRequester (-1, -1, NULL, NULL, StrCall, 6, FALSE, FALSE, "Edit Call Value",
                                                "{CGive the new CALL destination block number}C"))
                              if (StrCall[0] != '\0' && atoi (StrCall) == SrcIndex)                           /* Calling inself ? */
                                ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Don't be a wally!");
                              else if (StrCall[0] != '\0' && atol (StrCall) < TapeWindow[WhichWindow].NumEntries)
                              {
                                Accepted = TRUE;
                                ValCall = (short)atol (StrCall);
                                CallValue = ValCall - SrcIndex;
                                TEntry.Identifier = (dword)CallValue;
                                sprintf (TEntry.Text, "%d", ValCall);
                                PutComboEntry (&TWindow, &TEntry, TWindow.LastSelected, TRUE);
                              }
                              else
                                ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - The value is out of range\n"
                                                "Use values 0 through %d", TapeWindow[WhichWindow].NumEntries);
                            else
                              Accepted = TRUE;
                          }
                          break;
    }
    if (GlobalKey == K_ESC)  { Done = TRUE; Cancelled = TRUE; }
    MouseStatus ();
  }
  if (!Cancelled)
  {
    for (Cnt = 0 ; Cnt < NumCalls ; Cnt ++)                                                 /* Write all calls back to the buffer */
    {
      GetComboEntry (&TWindow, &TEntry, Cnt);
      *((short *)WorkBuffer + Cnt) = (short)TEntry.Identifier;
    }
    /* Resize the data block if needed - double size, for the jump codes */
    XMSIndex = ResizeTapeBlock (WhichWindow, Entry, (dword)NumCalls * 2, (dword)NumCalls * 2);
    CopyToXMS (Entry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, XMSIndex); /* Insert into the virtual tape */
  }
  DestroyComboWindow (&TWindow, 0);
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
  if (Cancelled && InitialNumCalls == 0)                                                       /* No calls (only after Insert!) ? */
  {
    DeleteTapeBlock (WhichWindow, Entry, TRUE, FALSE);                            /* Then we don't have to keep the entry anymore */
    _Cancelled = TRUE;
  }
  else
    _MustInitializeJumpCodes = TRUE;                                                      /* Signal: re-initialize all jump codes */
  return (!Cancelled);
}

bool EditRetSequenceBlock (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : There's nothing to change, as this block has no body.                                                                  */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  WhichWindow = WhichWindow;
  Entry = Entry;
  return (TRUE);
}

bool EditSelectBlock (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : The entry has been adjusted by the user and saved in the tape directory.                                               */
/*         Note: when called from ViewTapeBlock, there is at least 1 call left on exit. When called from InsertTapeEntry, if the  */
/*         user hit 'Cancel' without inserting any calls, the entry has been removed from the tape again.                         */
/* Import: ResizeTapeBlock.                                                                                                       */
/**********************************************************************************************************************************/

{
  struct WindowInfo CWindow  = {173, 160, 296, 150, 296, 150, 296, 150, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "Edit Select Block"};
  struct ButtonInfo PCancel  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 23, 125, 8, DBLACK, DWHITE, NULL, 0, "@Cancel", NULL};
  struct ButtonInfo POk      = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 102, 125, 8, DBLACK, DWHITE, NULL, 1, "@Ok", NULL};
  struct ButtonInfo PInsert  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 234, 11, 8, DBLACK, DWHITE, NULL, 2, "@Insert", NULL};
  struct ButtonInfo PAppend  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 234, 31, 8, DBLACK, DWHITE, NULL, 3, "@Append", NULL};
  struct ButtonInfo PRemove  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 234, 51, 8, DBLACK, DWHITE, NULL, 4, "@Remove", NULL};

  struct ComboInfo        TWindow;
  struct ComboEntry       TEntry;
  struct ButtonInfo      *ButtonP;
  comboent                SelectedLine;
  dword                   XMSIndex;
  short         register  Cnt;
  comboent                SelectValue;
  bool                    Done = FALSE;
  bool                    Cancelled = FALSE;
  bool                    Accepted = FALSE;
  bool                    Changed = FALSE;
  short                   NumSelections;
  short                   InitialNumSelections;
  char                    StrSelect[31];
  comboent                ValSelect;
  comboent                SrcIndex;                                                      /* (Absolute block number of this block) */
  byte                   *BTable;

  InitialNumSelections = NumSelections = Entry->BlockLength;                                           /* (Just to see if it's 0) */
  if (NumSelections == 0)                                                               /* Empty list (just inserted the entry) ? */
  {
    SetupNewData (WhichWindow);                          /* Insert into the list, so the user can see the resulting block numbers */
                                                                       /* This also makes sure that all jump values are corrected */
    PRemove.Flags |= BF_GHOSTED;
    POk.Flags |= BF_GHOSTED;
    SrcIndex = _EntryNumber;
  }
  else                                                                                    /* Not coming from the insert handler ? */
  {
    GetComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &TEntry, TapeWindow[WhichWindow].SelectedLine);
    SrcIndex = GetBlockNumber (TEntry.Text);
  }
  AddButton (&CWindow, &POk, TRUE);
  AddButton (&CWindow, &PInsert, FALSE);
  AddButton (&CWindow, &PAppend, FALSE);
  AddButton (&CWindow, &PRemove, FALSE);
  AddButton (&CWindow, &PCancel, FALSE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  MakeComboWindow (&TWindow, &CWindow, COMT_SELECT, COMF_NOBUFFER | COMF_VSLIDER | COMF_SYNCSLIDER | COMF_KEYSELECT,
                   CWindow.StartX + 8, CWindow.StartY + 8, 0, 0, 44, 34, 10, NULL, FALSE, 0, NULL, DBLACK, DWHITE, 0, 0);
  TEntry.Selected = 0x00;
  if (Entry->AllocLength > 0)
  {
    XMSIndex = GetTapeBlockXMSIndex (WhichWindow, Entry);
    CopyFromXMS (Entry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);              /* Fetch old row */
    InitialNumSelections = NumSelections = *WorkBuffer;                                        /* Get actual number of selections */
    BTable = WorkBuffer + 1;
    for (Cnt = 0 ; Cnt < NumSelections ; Cnt ++)                                   /* Insert all selections into the combo window */
    {
      SelectValue = *((short *)BTable);
      TEntry.Identifier = (dword)SelectValue;
      sprintf (TEntry.Text, "{+1%3d{++ ", SrcIndex + SelectValue);                           /* Make relative jump value absolute */
      strncpy (TEntry.Text + 10, (char *)BTable + 3, *(BTable + 2));                                          /* Add string value */
      TEntry.Text[*(BTable + 2) + 10] = '\0';
      if (!InsertComboEntry (&TWindow, &TEntry, NULL, -1))
        FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
      BTable += *(BTable + 2) + 3;
    }
    if (NumSelections >= 10)                                                                                       /* Full menu ? */
    {
      PInsert.Flags |= BF_GHOSTED;
      PAppend.Flags |= BF_GHOSTED;
      UpdateButtonText (&CWindow, &PInsert, 0);
      UpdateButtonText (&CWindow, &PAppend, 0);
    }
  }
  PrintComboEntries (&TWindow);
  SelectEntry (&TWindow, 0);
  while (!Done)
  {
    switch (HandleWindow (&CWindow, &ButtonP))
    {
      case WINH_CLOSE   : Done = TRUE; Cancelled = TRUE; break;
      case WINH_RNOHAND : switch (ButtonP->ReactNum)
                          {
                            case 0 : Done = TRUE; Cancelled = TRUE; break;                                            /* (Cancel) */
                            case 1 : Done = TRUE; Cancelled = FALSE; break;                                               /* (Ok) */
                            case 2 :                                                                                  /* (Insert) */
                            case 3 : Accepted = FALSE;                                                                /* (Append) */
                                     StrSelect[0] = '\0';
                                     ValSelect = SrcIndex + 1;
                                     if (_EditSelectValue (StrSelect, &ValSelect, TapeWindow[WhichWindow].NumEntries, SrcIndex))
                                     {
                                       SelectValue = ValSelect - SrcIndex;
                                       TEntry.Identifier = (dword)SelectValue;
                                       SelectedLine = TWindow.LastSelected + (ButtonP->ReactNum == 2 ? 0 : 1);
                                       sprintf (TEntry.Text, "{+1%3d{++ %s", SrcIndex + SelectValue, StrSelect);
                                       if (!InsertComboEntry (&TWindow, &TEntry, NULL, SelectedLine))
                                         FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
                                       SelectEntry (&TWindow, SelectedLine);
                                       NumSelections ++;
                                       if (NumSelections == 1)                                    /* First call in the list ? */
                                       {
                                         PRemove.Flags &= ~BF_GHOSTED;
                                         POk.Flags &= ~BF_GHOSTED;
                                         UpdateButtonText (&CWindow, &PRemove, 0);
                                         UpdateButtonText (&CWindow, &POk, 0);
                                         SelectEntry (&TWindow, 0);
                                       }
                                       else if (NumSelections >= 10)                                           /* List full now ? */
                                       {
                                         PInsert.Flags |= BF_GHOSTED;
                                         PAppend.Flags |= BF_GHOSTED;
                                         UpdateButtonText (&CWindow, &PInsert, 0);
                                         UpdateButtonText (&CWindow, &PAppend, 0);
                                       }
                                       Changed = TRUE;
                                     }
                                     break;
                            case 4 : if (TWindow.LastSelected >= 0 && TWindow.LastSelected <= TWindow.NumEntries)     /* (Remove) */
                                     {
                                       Changed = TRUE;
                                       DeleteComboEntry (&TWindow, TWindow.LastSelected);
                                       if (TWindow.LastSelected > TWindow.NumEntries)                     /* Removed last entry ? */
                                         TWindow.LastSelected = TWindow.NumEntries;
                                       SelectEntry (&TWindow, TWindow.LastSelected);
                                       NumSelections --;
                                       if (NumSelections == 0)                                     /* No more calls in the list ? */
                                       {
                                         PRemove.Flags |= BF_GHOSTED;
                                         POk.Flags |= BF_GHOSTED;
                                         UpdateButtonText (&CWindow, &PRemove, 0);
                                         UpdateButtonText (&CWindow, &POk, 0);
                                       }
                                       else if (NumSelections == 9)                              /* Insert buttons were ghosted ? */
                                       {
                                         PInsert.Flags &= ~BF_GHOSTED;
                                         PAppend.Flags &= ~BF_GHOSTED;
                                         UpdateButtonText (&CWindow, &PInsert, 0);
                                         UpdateButtonText (&CWindow, &PAppend, 0);
                                       }
                                     }
                                     break;
                          }
                          break;
    }
    switch (HandleComboWindow (&TWindow, &SelectedLine))
    {
      case WINH_LINESEL : GetComboEntry (&TWindow, &TEntry, TWindow.LastSelected);
                          strcpy (StrSelect, TEntry.Text + 10);
                          ValSelect = SrcIndex + (short)(TEntry.Identifier);
                          if (_EditSelectValue (StrSelect, &ValSelect, TapeWindow[WhichWindow].NumEntries, SrcIndex))
                          {
                            SelectValue = ValSelect - SrcIndex;
                            TEntry.Identifier = (dword)SelectValue;
                            sprintf (TEntry.Text, "{+1%3d{++ %s", SrcIndex + SelectValue, StrSelect);
                            PutComboEntry (&TWindow, &TEntry, TWindow.LastSelected, TRUE);
                            Changed = TRUE;
                          }
                          break;
    }
    if (GlobalKey == K_ESC)  { Done = TRUE; Cancelled = TRUE; }
    MouseStatus ();
  }
  if (!Cancelled && Changed)
  {
    *WorkBuffer = NumSelections;
    BTable = WorkBuffer + 1;
    for (Cnt = 0 ; Cnt < NumSelections ; Cnt ++)                                       /* Write all selections back to the buffer */
    {
      GetComboEntry (&TWindow, &TEntry, Cnt);
      *((short *)BTable) = (short)TEntry.Identifier;
      *(BTable + 2) = strlen (TEntry.Text + 10);
      strncpy (BTable + 3, TEntry.Text + 10, *(BTable + 2));
      BTable += *(BTable + 2) + 3;
    }
    /* Resize the data block if needed - double size, for the jump codes */
    XMSIndex = ResizeTapeBlock (WhichWindow, Entry, (dword)(BTable - WorkBuffer), (dword)NumSelections * 2);
    CopyToXMS (Entry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, XMSIndex); /* Insert into the virtual tape */
  }
  DestroyComboWindow (&TWindow, 0);
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
  if (Cancelled && InitialNumSelections == 0)                                             /* No selections (only after Insert!) ? */
  {
    DeleteTapeBlock (WhichWindow, Entry, TRUE, FALSE);                            /* Then we don't have to keep the entry anymore */
    _Cancelled = TRUE;
  }
  else if (Changed)
    _MustInitializeJumpCodes = TRUE;                                                      /* Signal: re-initialize all jump codes */
  return (!Cancelled && Changed);
}

bool EditTextDescrBlock (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : The entry has been adjusted by the user and saved in the tape directory.                                               */
/* Import: ResizeTapeBlock.                                                                                                       */
/**********************************************************************************************************************************/

{
  char  TextBuffer[257];
  short StrLength;
  dword XMSIndex;
  bool  Changed     = FALSE;
  bool  Cancelled   = FALSE;

  if (Entry->BlockLength > 0)
  {
    XMSIndex = GetTapeBlockXMSIndex (WhichWindow, Entry);
    CopyFromXMS (Entry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);           /* Fetch old string */
    *(WorkBuffer + Entry->BlockLength) = '\0';                                                               /* Force it a string */
    strcpy (TextBuffer, WorkBuffer);
  }
  else
    TextBuffer[0] = '\0';
  if (InputRequester (-1, -1, NULL, NULL, TextBuffer, 256, TRUE, FALSE, "Edit Text Description Block",
                      "{CGive the text description string}C"))
  {
    if (strcmp (TextBuffer, WorkBuffer))                                                    /* String has actually been changed ? */
    {
      StrLength = strlen (TextBuffer);
      XMSIndex = ResizeTapeBlock (WhichWindow, Entry, (dword)StrLength, 0);                    /* Resize the data block if needed */
      CopyToXMS (Entry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, TextBuffer, XMSIndex);           /* Write new string */
      Changed = TRUE;
    }
  }
  else                                                                                                             /* (Cancelled) */
    Cancelled = TRUE;
  if (Cancelled || !Changed)
  {
    if (_EntryNumber != -1)                                                                     /* Came from the insert handler ? */
      DeleteTapeBlock (WhichWindow, Entry, FALSE, FALSE);                         /* Then we don't have to keep the entry anymore */
    _Cancelled = TRUE;
    return (FALSE);
  }
  return (TRUE);
}

bool EditMessageBlock (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : The entry has been adjusted by the user and saved in the tape directory.                                               */
/* Import: ResizeTapeBlock.                                                                                                       */
/**********************************************************************************************************************************/

{
  struct WindowInfo   CWindow = {220, 150, 200, 170, 200, 170, 200, 170, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                 WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "Edit Message Block"};
  struct ButtonInfo   PCancel = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 33, 145, 8, DBLACK, DWHITE, NULL, 0, "@Cancel", NULL};
  struct ButtonInfo   POk     = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 112, 145, 8, DBLACK, DWHITE, NULL, 1, "@Ok", NULL};
  struct ButtonInfo   BTimer  = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 102, 10, 5, DBLACK, DWHITE, NULL, 2,
                                 NULL, NULL};
  struct InputBoxInfo BIBox   = {8, 46, 30, 8, DBLACK, DWHITE, NULL, 3, NULL, BF_NOSURROUND | BF_KEYSELECT, 255, 20};
  struct ButtonInfo  *ButtonP;
  short      register Cnt;
  short      register Cnt2;
  char                ValTimer[6];
  word                OldTimer;
  bool                Done                  = FALSE;
  bool                Cancelled             = FALSE;
  bool                Changed               = FALSE;
  short               StrLength;
  dword               XMSIndex;

  BTimer.Contents = ValTimer;
  BIBox.Contents = WorkBuffer + 256;
  *WorkBuffer = '\0';
  if (Entry->BlockLength > 0)
  {
    XMSIndex = GetTapeBlockXMSIndex (WhichWindow, Entry);
    CopyFromXMS (Entry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);           /* Fetch old string */
    *(WorkBuffer + Entry->BlockLength) = '\0';                                                               /* Force it a string */
  }
  for (Cnt = 0, Cnt2 = 256 ; Cnt < Entry->BlockLength ; Cnt ++, Cnt2 ++)
    if (*(WorkBuffer + Cnt) != 0x0A)                                                                /* Remove linefeed characters */
      *(WorkBuffer + Cnt2) = *(WorkBuffer + Cnt);
  *(WorkBuffer + Cnt2) = '\0';
  OldTimer = Entry->PauseLength;
  sprintf (ValTimer, "%u", OldTimer);
  AddButton (&CWindow, &POk, TRUE);
  AddButton (&CWindow, &BTimer, FALSE);
  AddInputBox (&CWindow, &BIBox, FALSE);
  AddButton (&CWindow, &PCancel, FALSE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  Wprintf (&CWindow, 1, 1, "{+1Display time}N");
  Wprintf (&CWindow, 1, 23, "{+1seconds}N");
  Wprintf (&CWindow, 3, 1, "{+1Message}N");
  while (!Done)
  {
    switch (HandleWindow (&CWindow, &ButtonP))
    {
      case WINH_CLOSE   : Done = TRUE; Cancelled = TRUE; break;
      case WINH_RNOHAND : switch (ButtonP->ReactNum)
                          {
                            case 0 : Done = TRUE; Cancelled = TRUE; break;                                            /* (Cancel) */
                            case 1 : Done = TRUE; Cancelled = FALSE; break;                                               /* (Ok) */
                          }
      case WINH_EDITEND : if (ButtonP->Flags & BF_FITINPUT)                                            /* Contents were changed ? */
                          {
                            Changed = TRUE;
                            switch (ButtonP->ReactNum)
                            {
                              case 2 : OldTimer = atol (ValTimer);
                                       sprintf (ValTimer, "%u", OldTimer);
                                       UpdateButtonText (&CWindow, &BTimer, BF_FITTXT);
                                       break;
                            }
                          }
                          break;
    }
    if (GlobalKey == K_ESC)  { Done = TRUE; Cancelled = TRUE; }
    MouseStatus ();
  }
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
  if (Cancelled || !Changed)
  {
    if (_EntryNumber != -1)                                                                     /* Came from the insert handler ? */
      DeleteTapeBlock (WhichWindow, Entry, FALSE, FALSE);                         /* Then we don't have to keep the entry anymore */
    _Cancelled = TRUE;
    return (FALSE);
  }
  StrLength = strlen (WorkBuffer + 256);
  XMSIndex = ResizeTapeBlock (WhichWindow, Entry, (dword)StrLength, 0);                        /* Resize the data block if needed */
  CopyToXMS (Entry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer + 256, XMSIndex);         /* Write new string */
  Entry->PauseLength = OldTimer;
  return (TRUE);
}

bool EditArchiveInfoBlock (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : The entry has been adjusted by the user and saved in the tape directory.                                               */
/* Import: ResizeTapeBlock.                                                                                                       */
/**********************************************************************************************************************************/

{
  struct WindowInfo   CWindow   = {58, 93, 523, 295, 523, 295, 523, 295, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                   WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "Edit Archive Info Block"};
  struct ButtonInfo   PCancel   = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 180, 270, 8, DBLACK, DWHITE, NULL, 0, "@Cancel", NULL};
  struct ButtonInfo   POk       = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 290, 270, 8, DBLACK, DWHITE, NULL, 1, "@Ok", NULL};
  struct ButtonInfo   BTFullNam = {BT_INPUT, BC_TEXT, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 10, 24, 40, DBLACK, DWHITE, NULL, 2,
                                   NULL, NULL};
  struct ButtonInfo   BTType    = {BT_INPUT, BC_TEXT, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 268, 24, 40, DBLACK, DWHITE, NULL, 8,
                                   NULL, NULL};
  struct ButtonInfo   BTPublish = {BT_INPUT, BC_TEXT, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 10, 64, 40, DBLACK, DWHITE, NULL, 3,
                                   NULL, NULL};
  struct ButtonInfo   BTYear    = {BT_INPUT, BC_TEXT, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 268, 64, 40, DBLACK, DWHITE, NULL, 5,
                                   NULL, NULL};
  struct ButtonInfo   BTLang    = {BT_INPUT, BC_TEXT, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 10, 104, 40, DBLACK, DWHITE, NULL, 6,
                                   NULL, NULL};
  struct ButtonInfo   BTScheme  = {BT_INPUT, BC_TEXT, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 268, 104, 40, DBLACK, DWHITE, NULL, 10,
                                   NULL, NULL};
  struct ButtonInfo   BTPrice   = {BT_INPUT, BC_TEXT, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 10, 144, 40, DBLACK, DWHITE, NULL, 9,
                                   NULL, NULL};
  struct ButtonInfo   BTOrigin  = {BT_INPUT, BC_TEXT, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 268, 144, 40, DBLACK, DWHITE, NULL, 11,
                                   NULL, NULL};
  struct InputBoxInfo BTAuthors = {10, 184, 40, 6, DBLACK, DWHITE, NULL, 4, NULL, BF_NOSURROUND | BF_KEYSELECT, 255, 10};
  struct InputBoxInfo BTComment = {268, 184, 40, 6, DBLACK, DWHITE, NULL, 7, NULL, BF_NOSURROUND | BF_KEYSELECT, 255, 10};
  struct ButtonInfo  *ButtonP;
  char                TextString[10][256] = {{"\0"}, {"\0"}, {"\0"}, {"\0"}, {"\0"}, {"\0"}, {"\0"}, {"\0"}, {"\0"}, {"\0"}};
  dword               XMSIndex;
  bool                Done      = FALSE;
  bool                Cancelled = FALSE;
  bool                Changed   = FALSE;
  byte               *BlockP;
  byte                NumFields;
  byte                FieldType;
  byte                FieldLength;
  byte      register  Cnt;
  dword               NewSize;

  BTFullNam.Contents = TextString[0];
  BTPublish.Contents = TextString[1];
  BTAuthors.Contents = TextString[2];
  BTYear.Contents    = TextString[3];
  BTLang.Contents    = TextString[4];
  BTType.Contents    = TextString[5];
  BTPrice.Contents   = TextString[6];
  BTScheme.Contents  = TextString[7];
  BTOrigin.Contents  = TextString[8];
  BTComment.Contents = TextString[9];
  if (Entry->BlockLength > 0)                                                                      /* Editing an existing entry ? */
  {
    XMSIndex = GetTapeBlockXMSIndex (WhichWindow, Entry);
    CopyFromXMS (Entry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, WorkBuffer);            /* Fetch old block */
    NumFields = *WorkBuffer;
    BlockP = WorkBuffer + 1;                                                                            /* Start at the beginning */
    while (NumFields > 0)
    {
      switch (*(BlockP ++))                                                               /* Determine what type of field this is */
      {
        case 0x00 : FieldType = 0x00; break;                                                                      /* (Full Title) */
        case 0x01 : FieldType = 0x01; break;                                                        /* (Software House/Publisher) */
        case 0x02 : FieldType = 0x02; break;                                                                       /* (Author(s)) */
        case 0x03 : FieldType = 0x03; break;                                                             /* (Year of Publication) */
        case 0x04 : FieldType = 0x04; break;                                                                        /* (Language) */
        case 0x05 : FieldType = 0x05; break;                                                               /* (Game/Utility Type) */
        case 0x06 : FieldType = 0x06; break;                                                                           /* (Price) */
        case 0x07 : FieldType = 0x07; break;                                                        /* (Protection Scheme/Loader) */
        case 0x08 : FieldType = 0x08; break;                                                                          /* (Origin) */
        default   : FieldType = 0x09; break;                                                              /* (Comment(s)/Unknown) */
      }
      FieldLength = *(BlockP ++);
      for (Cnt = 0 ; Cnt < FieldLength ; Cnt ++)
        TextString[FieldType][Cnt] = *(BlockP ++);
      TextString[FieldType][FieldLength] = '\0';                                               /* (Stick an endmarker at the end) */
      NumFields --;
    }
  }
  AddButton (&CWindow, &POk, TRUE);
  AddButton (&CWindow, &BTFullNam, FALSE);
  AddButton (&CWindow, &BTType, FALSE);
  AddButton (&CWindow, &BTPublish, FALSE);
  AddButton (&CWindow, &BTYear, FALSE);
  AddButton (&CWindow, &BTLang, FALSE);
  AddButton (&CWindow, &BTScheme, FALSE);
  AddButton (&CWindow, &BTPrice, FALSE);
  AddButton (&CWindow, &BTOrigin, FALSE);
  AddInputBox (&CWindow, &BTAuthors, FALSE);
  AddInputBox (&CWindow, &BTComment, FALSE);
  BTFullNam.MaxInputLength = BTPublish.MaxInputLength = BTYear.MaxInputLength = BTLang.MaxInputLength = 
  BTType.MaxInputLength = BTPrice.MaxInputLength = BTScheme.MaxInputLength = BTOrigin.MaxInputLength = 255;
  AddButton (&CWindow, &PCancel, FALSE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  Wprintf (&CWindow,  1,  1, "{+1Full Title}N");
  Wprintf (&CWindow,  1, 44, "{+1Game/Utility Type}N");
  Wprintf (&CWindow,  5,  1, "{+1Software House/Publisher}N");
  Wprintf (&CWindow,  5, 44, "{+1Year of Publication}N");
  Wprintf (&CWindow,  9,  1, "{+1Language}N");
  Wprintf (&CWindow,  9, 44, "{+1Protection Scheme/Loader}N");
  Wprintf (&CWindow, 13,  1, "{+1Price}N");
  Wprintf (&CWindow, 13, 44, "{+1Origin}N");
  Wprintf (&CWindow, 17,  1, "{+1Author(s)}N");
  Wprintf (&CWindow, 17, 44, "{+1Comment(s)}N");
  while (!Done)
  {
    switch (HandleWindow (&CWindow, &ButtonP))
    {
      case WINH_CLOSE   : Done = TRUE; Cancelled = TRUE; break;
      case WINH_RNOHAND : switch (ButtonP->ReactNum)
                          {
                            case 0  : Done = TRUE; Cancelled = TRUE; break;                                           /* (Cancel) */
                            case 1  : Done = TRUE; Cancelled = FALSE; break;                                              /* (Ok) */
                          }
                          break;
      case WINH_EDITEND : if (ButtonP->Flags & BF_FITINPUT)                                            /* Contents were changed ? */
                            Changed = TRUE;
    }
    if (GlobalKey == K_ESC)  { Done = TRUE; Cancelled = TRUE; }
    MouseStatus ();
  }
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
  if (!Changed || Cancelled)
  {
    if (_EntryNumber != -1)                                                                     /* Came from the insert handler ? */
      DeleteTapeBlock (WhichWindow, Entry, FALSE, FALSE);                         /* Then we don't have to keep the entry anymore */
    _Cancelled = TRUE;
    return (FALSE);
  }
  NumFields = 0;
  BlockP = WorkBuffer + 1;                                                                /* (After the number-of-fields counter) */
  for (FieldType = 0 ; FieldType < 10 ; FieldType ++)
    if (TextString[FieldType][0] != '\0')                                                                    /* Field filled in ? */
    {
      NumFields ++;
      FieldLength = strlen (TextString[FieldType]);
      switch (FieldType)
      {
        case 0x00 : *(BlockP ++) = 0x00; break;
        case 0x01 : *(BlockP ++) = 0x01; break;
        case 0x02 : *(BlockP ++) = 0x02; break;
        case 0x03 : *(BlockP ++) = 0x03; break;
        case 0x04 : *(BlockP ++) = 0x04; break;
        case 0x05 : *(BlockP ++) = 0x05; break;
        case 0x06 : *(BlockP ++) = 0x06; break;
        case 0x07 : *(BlockP ++) = 0x07; break;
        case 0x08 : *(BlockP ++) = 0x08; break;
        case 0x09 : *(BlockP ++) = 0xFF; break;
      }
      *(BlockP ++) = FieldLength;
      for (Cnt = 0 ; Cnt < FieldLength ; Cnt ++)
        *(BlockP ++) = TextString[FieldType][Cnt];
    }
  *WorkBuffer = NumFields;                                                                     /* Store the number of used fields */
  NewSize = BlockP - WorkBuffer;
  XMSIndex = ResizeTapeBlock (WhichWindow, Entry, NewSize, 0);                                 /* Resize the data block if needed */
  CopyToXMS (Entry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, XMSIndex);                /* Write new block */
  return (TRUE);                                                                                               /* Signal: changed */
}

bool EditHardwareTypeBlock (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : The entry has been adjusted by the user and saved in the tape directory.                                               */
/*         Note: when called from ViewTapeBlock, there is at least 1 entry left on exit. When called from InsertTapeEntry, if the */
/*         user hit 'Cancel' without inserting any entries, the entry has been removed from the tape again.                       */
/* Import: ResizeTapeBlock, GetHardwareTable.                                                                                     */
/**********************************************************************************************************************************/

{
  struct WindowInfo CWindow  = {36, 60, 564, 360, 564, 360, 564, 360, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "Edit Hardware Type Block"};
  struct ButtonInfo PCancel  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 254, 302, 8, DBLACK, DWHITE, NULL, 0, "@Cancel", NULL};
  struct ButtonInfo POk      = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 254, 322, 8, DBLACK, DWHITE, NULL, 1, "@Ok", NULL};
  struct ButtonInfo BAdd     = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 254, 55, 8, DBLACK, DWHITE, NULL, 2, "@Add", NULL};
  struct ButtonInfo BDelete  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 254, 75, 8, DBLACK, DWHITE, NULL, 3, "@Delete", NULL};
  struct ButtonInfo THow     = {BT_TEXT, BC_TEXT, BF_FWIDTH | BF_FITTXT, 393, 30, 19, DBLACK, DWHITE, NULL, 4, NULL, NULL};
  struct ButtonInfo BHow     = {BT_PUSH, BC_GADGET, BF_FWIDTH | BF_KEYSELECT, 514, 30, 1, DBLACK, DWHITE, NULL, 5, "\007", NULL};

  struct ComboInfo    TWindow;
  struct ComboInfo    IWindow;
  struct ComboEntry   TEntry;
  struct ButtonInfo  *ButtonP;
  char                HValue[20] = "Runs & could use";
  dword               XMSIndex;
  dword              *DTable;
  byte               *HWIndex;
  byte               *HWName;
  byte                Bt1;
  byte                Bt2;
  byte                Bt3;
  byte                LastValue = 0xFF;
  byte                LastType = 0xFF;
  comboent            SelectedTLine;
  comboent            SelectedILine;
  byte      register  Cnt;
  byte      register  Cnt2;
  byte                HNum        = 0;
  bool                Done        = FALSE;
  bool                Cancelled   = FALSE;
  bool                Redraw      = TRUE;
  bool                First       = TRUE;
  bool                DoAdd       = FALSE;
  bool                DoDelete    = FALSE;
  bool                Changed     = FALSE;
  word                NumHWTypes;
  word                InitialNumHWTypes;
  dword               NewField;

  THow.Contents = HValue;
  HWIndex = WorkBuffer + (dword)49152 - HWTableLength;                                           /* Use the end of the WorkBuffer */
  DTable = (dword *)WorkBuffer;
  CopyFromXMS (HWTableLength, HWTableXMSHandle, 0, HWIndex);                                          /* Read hardware type table */
  InitialNumHWTypes = NumHWTypes = (word)(Entry->BlockLength / 3);
  if (Entry->BlockLength > 0)
    GetHardwareTable (WhichWindow, Entry);                                                           /* Get the table and sort it */
  else
    BDelete.Flags |= BF_GHOSTED;
  AddButton (&CWindow, &POk, TRUE);
  AddButton (&CWindow, &BAdd, FALSE);
  AddButton (&CWindow, &BDelete, FALSE);
  AddButton (&CWindow, &THow, FALSE);
  AddButton (&CWindow, &BHow, FALSE);
  AddButton (&CWindow, &PCancel, FALSE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  WDrawShadowBox (&CWindow, 1, 1, 34, 40, "Hardware Types");
  WDrawShadowBox (&CWindow, 1, 52, 34, 91, "Selected Types");
  Wprintf (&CWindow, 3, 57, "{+1Add as}N");
  MakeComboWindow (&TWindow, &CWindow, COMT_SELECT, COMF_NOBUFFER | COMF_VSLIDER | COMF_SYNCSLIDER, CWindow.StartX + 18,
                   CWindow.StartY + 33, 0, 0, 40, 34, 30, NULL, FALSE, 0, NULL, DBLACK, DWHITE, 0, 0);
  MakeComboWindow (&IWindow, &CWindow, COMT_SELECT, COMF_NOBUFFER | COMF_VSLIDER | COMF_SYNCSLIDER, CWindow.StartX + 324,
                   CWindow.StartY + 53, 0, 0, 40, 34, 28, NULL, FALSE, 0, NULL, DBLACK, DWHITE, 0, 0);
  TEntry.Selected = 0;
  for (Cnt = 0 ; Cnt < *HWIndex ; Cnt ++)                                                                        /* Do all groups */
  {
    sprintf (TEntry.Text, "{+1%s}N", HWIndex + 4 + (dword)*(HWIndex + 2) * Cnt);                             /* Insert group name */
    TEntry.Selected = ENT_NOSELECT;
    if (!InsertComboEntry (&TWindow, &TEntry, NULL, -1))
      FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
    Done = FALSE;
    for (Cnt2 = 0 ; !Done && Cnt2 < *(HWIndex + 1) ; Cnt2 ++)                                        /* Do all objects in a group */
    {
      HWName = HWIndex + 4 + ((dword)*(HWIndex + 2) * *HWIndex) +
                             ((dword)*(HWIndex + 2) * *(HWIndex + 1) * Cnt) +
                             ((dword)*(HWIndex + 2) * Cnt2);
      if (*HWName == '\0')                                                                                     /* Not filled in ? */
        Done = TRUE;
      else
      {
        sprintf (TEntry.Text, "  %s", HWName);
        TEntry.Selected = 0;
        TEntry.Identifier = ((word)Cnt << 8) | Cnt2;
        if (!InsertComboEntry (&TWindow, &TEntry, NULL, -1))
          FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
      }
    }
  }
  PrintComboEntries (&TWindow);
  while (Redraw)
  {
    if (First)
      First = FALSE;
    else
      ReInitComboWindow (&IWindow, 0);
    LastValue = LastType = 0xFF;
    if (NumHWTypes > 0)
      for (Cnt = 0 ; Cnt < NumHWTypes ; Cnt ++)                                                                   /* Do all types */
      {
        TEntry.Selected = 0;
        TEntry.Identifier = 0;
        Bt1 = (byte)((*(DTable + Cnt) & 0x0000FF00) >> 8);                                                     /* (Hardware Type) */
        Bt2 = (byte)(*(DTable + Cnt) & 0x000000FF);                                                              /* (Hardware ID) */
        Bt3 = (byte)((*(DTable + Cnt) & 0x00FF0000) >> 16);                                                            /* (Value) */
        if (LastValue == 0xFF || Bt3 != LastValue)
        {
          LastValue = Bt3;
          LastType = 0xFF;                                                                   /* Force hardware type to be printed */
          switch (Bt3)
          {
            case 0 : strcpy (TEntry.Text, "{BRuns & could use  }B"); break;
            case 1 : strcpy (TEntry.Text, "{BRuns & uses       }B"); break;
            case 2 : strcpy (TEntry.Text, "{BRuns & doesn\'t use}B"); break;
            case 3 : strcpy (TEntry.Text, "{BDoesn\'t run on    }B"); break;
            default: strcpy (TEntry.Text, "{B<UNDEFINED>       }B"); break;
          }
          TEntry.Selected = ENT_NOSELECT;
          if (!InsertComboEntry (&IWindow, &TEntry, NULL, -1))
            FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
          TEntry.Selected = 0;
        }
        if (Bt1 < *HWIndex)
        {
          if (LastType == 0xFF || Bt1 != LastType)
          {
            LastType = Bt1;
            sprintf (TEntry.Text, " {+1%s}N", HWIndex + 4 + (dword)*(HWIndex + 2) * Bt1);
            TEntry.Selected = ENT_NOSELECT;
            if (!InsertComboEntry (&IWindow, &TEntry, NULL, -1))
              FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
            TEntry.Selected = 0;
          }
          HWName = HWIndex + 4 + ((dword)*(HWIndex + 2) * *HWIndex) +
                                 ((dword)*(HWIndex + 2) * *(HWIndex + 1) * Bt1) +
                                 ((dword)*(HWIndex + 2) * Bt2);
          if (*HWName == '\0')                                                                                 /* Not filled in ? */
            strcpy (TEntry.Text, "  <Undefined>");
          else
            sprintf (TEntry.Text, "  %s", HWName);
          TEntry.Identifier = ((dword)Bt3 << 16) | ((dword)Bt1 << 8) | Bt2;
        }
        else
        {
          strcat (TEntry.Text, "<Undefined>)}N");
          TEntry.Selected = ENT_NOSELECT;
        }
        if (!InsertComboEntry (&IWindow, &TEntry, NULL, -1))
          FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
      }
    PrintComboEntries (&IWindow);
    Done = FALSE;
    while (!Done)
    {
      switch (HandleWindow (&CWindow, &ButtonP))
      {
        case WINH_CLOSE   : Done = TRUE; Redraw = FALSE; Cancelled = TRUE; break;
        case WINH_RNOHAND : switch (ButtonP->ReactNum)
                            {
                              case 0 : Done = TRUE; Redraw = FALSE; Cancelled = TRUE; break;                          /* (Cancel) */
                              case 1 : Done = TRUE; Redraw = FALSE; Cancelled = FALSE; break;                             /* (Ok) */
                              case 2 : Changed = TRUE; DoAdd = TRUE; break;                                              /* (Add) */
                              case 3 : Changed = TRUE; DoDelete = TRUE; break;                                        /* (Delete) */
                              case 5 : if (++ HNum == 4)                                                              /* (Add as) */
                                         HNum = 0;
                                       switch (HNum)
                                       {
                                         case 0 : strcpy (HValue, "Runs & could use  "); break;
                                         case 1 : strcpy (HValue, "Runs & uses       "); break;
                                         case 2 : strcpy (HValue, "Runs & doesn\'t use"); break;
                                         case 3 : strcpy (HValue, "Doesn\'t run on    "); break;
                                       }
                                       UpdateButtonText (&CWindow, &THow, BF_FITTXT);
                                       break;
                            }
                            break;
      }
      if (HandleComboWindow (&TWindow, &SelectedTLine) == WINH_LINESEL)
      { Changed = TRUE; DoAdd = TRUE; }
      if (HandleComboWindow (&IWindow, &SelectedILine) == WINH_LINESEL)
      { Changed = TRUE; DoDelete = TRUE; }
      if (DoAdd)
      {
        DoAdd = FALSE;
        GetComboEntry (&TWindow, &TEntry, SelectedTLine);                                              /* Fetch entry to be added */
        NewField = ((dword)HNum << 16) | TEntry.Identifier;                                                /* Build the new field */
        Done = FALSE;
        for (Cnt = 0 ; Cnt < NumHWTypes && !Done ; Cnt ++)                                                 /* Not there already ? */
          if ((*(DTable + Cnt) & 0x0000FFFF) == TEntry.Identifier)
            Done = TRUE;
        if (Done)
        {
          ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - This combination is already there!");
          Done = FALSE;
        }
        else
        {
          for (Cnt = 0 ; Cnt < NumHWTypes && !Done ; Cnt ++)
            if (NewField < *(DTable + Cnt))                                                            /* Must be inserted here ? */
            {
              memmove (DTable + Cnt + 1, DTable + Cnt, (NumHWTypes - Cnt) * 4);                                     /* (Shift up) */
              *(DTable + Cnt) = NewField;
              Done = TRUE;
            }
          if (!Done)                                                                                /* (Must be added at the end) */
            *(DTable + NumHWTypes) = NewField;
          if (++ NumHWTypes == 1)                                                                          /* First entry added ? */
          {
            BDelete.Flags &= ~BF_GHOSTED;                                                            /* Unghost the Delete button */
            UpdateButtonText (&CWindow, &BDelete, 0);
          }
          Done = TRUE; Redraw = TRUE;
        }
      }
      else if (DoDelete)
      {
        DoDelete = FALSE;
        GetComboEntry (&IWindow, &TEntry, SelectedILine);                                            /* Fetch entry to be deleted */
        Done = FALSE;
        for (Cnt = 0 ; Cnt < NumHWTypes && !Done ; Cnt ++)                                                    /* Find the entry ? */
          if (*(DTable + Cnt) == TEntry.Identifier)
          {
            memmove (DTable + Cnt, DTable + Cnt + 1, (NumHWTypes - Cnt - 1) * 4);                                 /* (Shift back) */
            Done = TRUE;
          }
        if (Done)                                                                                               /* (Sanity check) */
        {
          if (-- NumHWTypes == 0)                                                                         /* Last entry deleted ? */
          {
            BDelete.Flags |= BF_GHOSTED;                                                               /* Ghost the Delete button */
            UpdateButtonText (&CWindow, &BDelete, 0);
          }
          Done = TRUE; Redraw = TRUE;
        }
      }
      if (GlobalKey == K_ESC)  { Done = TRUE; Redraw = FALSE; Cancelled = TRUE; }
      MouseStatus ();
    }
  }
  DestroyComboWindow (&TWindow, 0);
  DestroyComboWindow (&IWindow, 0);
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
  if (Cancelled || !Changed)
  {
    if (InitialNumHWTypes == 0)                                                              /* No entries (only after Insert!) ? */
      DeleteTapeBlock (WhichWindow, Entry, FALSE, FALSE);                         /* Then we don't have to keep the entry anymore */
    _Cancelled = TRUE;
    return (FALSE);
  }
  HWName = WorkBuffer;                                                    /* (Table is shrunk 3:4, so there's no overlap problem) */
  for (Cnt = 0 ; Cnt < NumHWTypes ; Cnt ++)                                             /* Convert the result back to tape format */
  {
    NewField = *(DTable + Cnt);
    *(HWName ++) = (byte)((NewField & 0x0000FF00) >> 8);                                                       /* (Hardware Type) */
    *(HWName ++) = (byte)(NewField & 0x000000FF);                                                                /* (Hardware ID) */
    *(HWName ++) = (byte)((NewField & 0x00FF0000) >> 16);                                                              /* (Value) */
  }
  XMSIndex = ResizeTapeBlock (WhichWindow, Entry, (dword)NumHWTypes * 3, 0);                   /* Resize the data block if needed */
  if (NumHWTypes > 0)
    CopyToXMS (Entry->AllocLength, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, XMSIndex);     /* Insert into virtual tape */
  return (TRUE);
}

bool EditCustomInfoBlock (short WhichWindow, struct TapeEntry_s *Entry)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Entry' points to the tape directory entry to be edited.                    */
/* Post  : The entry has been adjusted by the user and saved in the tape directory.                                               */
/* Import: ResizeTapeBlock.                                                                                                       */
/**********************************************************************************************************************************/

{
  dword XMSIndex;
  dword OldSize   = 0;
  dword NewSize;
  bool  Changed;
  char  CustomType[16];

  if (Entry->BlockLength > 0)                                                                      /* Editing an existing block ? */
  {
    XMSIndex = GetTapeBlockXMSIndex (WhichWindow, Entry);
    CopyFromXMS (16, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex, CustomType);                  /* Fetch type of Custom Info */
    OldSize = Entry->BlockLength - 16;
    if (OldSize > 0 && Entry->FlagByte == INSRTCI_POKES || Entry->FlagByte == INSRTCI_SCREEN)
      CopyFromXMS (Entry->AllocLength - 16, TapeWindow[WhichWindow].Tape.XMSHandle, XMSIndex + 16, WorkBuffer);  /* Get old block */
  }
  else                                                                                                  /* Nope, inserting it now */
    switch (_InsertCustomType)
    {
      case INSRTCI_POKES    : strncpy (CustomType, "POKEs           ", 16); break;
      case INSRTCI_INSTRUCT : strncpy (CustomType, "Instructions    ", 16); break;
      case INSRTCI_SCREEN   : strncpy (CustomType, "Spectrum Screen ", 16); break;
      case INSRTCI_ZXEDIT   : strncpy (CustomType, "ZX-Edit Document", 16); break;
      case INSRTCI_PICTURE  : strncpy (CustomType, "Picture         ", 16); break;
      default               : return (FALSE);
    }
  switch (Entry->FlagByte)
  {
    case INSRTCI_POKES    : Changed = EditGamePokes (OldSize, &NewSize); break;
    case INSRTCI_INSTRUCT : Changed = EditInstructionsBlock (WhichWindow, Entry, OldSize, &NewSize); break;
    case INSRTCI_SCREEN   : Changed = EditSpectrumScreen (WhichWindow, Entry, OldSize, &NewSize); break;
    case INSRTCI_ZXEDIT   : Changed = EditInstructionsBlock (WhichWindow, Entry, OldSize, &NewSize); break;
    case INSRTCI_PICTURE  : Changed = EditPictureBlock (WhichWindow, Entry, OldSize, &NewSize); break;
  }
  if (!Changed)
  {
    if (Entry->BlockLength == 0)                                                     /* No previous block, was inserting it now ? */
      DeleteTapeBlock (WhichWindow, Entry, FALSE, FALSE);                         /* Then we don't have to keep the entry anymore */
    _Cancelled = TRUE;
    return (FALSE);
  }
  if (Entry->FlagByte == INSRTCI_POKES || Entry->FlagByte == INSRTCI_SCREEN)
  {
    XMSIndex = ResizeTapeBlock (WhichWindow, Entry, NewSize + 16, 0);                          /* Resize the data block if needed */
    CopyToXMS (16, TapeWindow[WhichWindow].Tape.XMSHandle, CustomType, XMSIndex);                       /* (Over)write the header */
    if (NewSize > 0)                                                                                  /* Insert into virtual tape */
      CopyToXMS (Entry->AllocLength - 16, TapeWindow[WhichWindow].Tape.XMSHandle, WorkBuffer, XMSIndex + 16);
  }
  return (TRUE);
}

bool InsertTapeEntry (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : The user has been presented a requester to determine what type of block should be inserted. If the user did not        */
/*         cancel this requester, a block of the requested type has been inserted and TRUE is returned.                           */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
#define MAXINSTYPE 22
  struct InsTypes_s
  {
    byte BlockTypeID;
    bool (*EditHandler)(short WhichWindow, struct TapeEntry_s *Entry);
    bool AllowEmptyTape;
  } Types[MAXINSTYPE]        = {{ 0x32, EditArchiveInfoBlock,  TRUE  },                                   /* (Archive Info Block) */
                                { 0x26, EditCallSequenceBlock, FALSE },                                  /* (Call Sequence Block) */
                                { 0xF1, _InsertCIPOKEs,        TRUE  },                            /* (Custom Info - POKEs Block) */
                                { 0xF2, _InsertCIInstructions, TRUE  },                     /* (Custom Info - Instructions Block) */
                                { 0xF3, _InsertCIScreen,       TRUE  },                           /* (Custom Info - Screen Block) */
                                { 0xF4, _InsertCIZXEdit,       TRUE  },                       /* (Custom Info - ZX-Edit Document) */
                                { 0xF5, _InsertCIPicture,      TRUE  },                                /* (Custom Info - Picture) */
                                { 0x21, EditGroupStrtBlock,    TRUE  },                                    /* (Group Start Block) */
                                { 0x33, EditHardwareTypeBlock, TRUE  },                                  /* (Hardware Type Block) */
                                { 0x14, _InsertHeaderBlock,    TRUE  },                                         /* (Header Block) */
                                { 0xFF, _InsertBASICBlock,     TRUE  },                                          /* (BASIC Block) */
                                { 0x23, EditJumpToBlock,       FALSE },                                  /* (Jump To Entry Block) */
                                { 0x24, EditLoopStrtBlock,     FALSE },                                     /* (Loop Start Block) */
                                { 0x31, EditMessageBlock,      TRUE  },                                        /* (Message Block) */
                                { 0x20, EditPauseBlock,        TRUE  },                                          /* (Pause Block) */
                                { 0x12, EditPureToneBlock,     TRUE  },                                      /* (Pure Tone Block) */
                                { 0x11, _InsertRawDataFile,    TRUE  },                                        /* (Raw Datablock) */
                                { 0x27, EditRetSequenceBlock,  FALSE },                           /* (Return from Sequence Block) */
                                { 0x13, EditRowOfPulsesBlock,  TRUE  },                                  /* (Row Of Pulses Block) */
                                { 0x28, EditSelectBlock,       FALSE },                                         /* (Select Block) */
                                { 0x2A, EditRetSequenceBlock,  TRUE  },                             /* (Stop Tape if in 48K Mode) */
                                { 0x30, EditTextDescrBlock,    TRUE  }};                              /* (Text Description Block) */

  struct WindowInfo CWindow  = {218, 90, 204, 305, 204, 305, 204, 305, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "Insert Tape Entry"};
  struct ButtonInfo PAppend  = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 118, 30, 1, DBLACK, DWHITE, NULL, 0, NULL, NULL};
  struct ButtonInfo PCancel  = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 30, 280, 8, DBLACK, DWHITE, NULL, 1, "@Cancel", NULL};
  struct ButtonInfo POk      = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 122, 280, 8, DBLACK, DWHITE, NULL, 0, "@Ok", NULL};

  struct ComboInfo    TWindow;
  struct ComboEntry   TEntry;
  struct ButtonInfo  *ButtonP;
  struct TapeEntry_s *DirP;
  struct TapeEntry_s *Directory;
  struct TapeEntry_s *NewEntry;
  bool                Append      = FALSE;
  char               *TextLine;
  comboent            SelectedLine;
  byte      register  Cnt;
  byte                BlockTypeID;
  byte                MatchingBlockTypeID;
  bool                Done        = FALSE;
  bool                Cancelled   = FALSE;

  PAppend.Contents = &Append;
  _InsertCustomType = INSRTCI_NONE;
  AddButton (&CWindow, &POk, TRUE);
  AddButton (&CWindow, &PAppend, FALSE);
  AddButton (&CWindow, &PCancel, FALSE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  Wprintf (&CWindow, 1, TEXT_CENTER, "Insert entry in %s tape", WhichWindow == 0 ? "source" : "destination");
  Wprintf (&CWindow, 3, 12, "{+1Append}N");
  MakeComboWindow (&TWindow, &CWindow, COMT_SELECT, COMF_NOBUFFER | COMF_VSLIDER | COMF_SYNCSLIDER | COMF_KEYSELECT,
                   CWindow.StartX + 5, CWindow.StartY + 50, 0, 0, 32, 30, 21, NULL, FALSE, 0, NULL, DBLACK, DWHITE, 0, 0);
  TEntry.Selected = 0;
  for (Cnt = 0 ; Cnt < MAXINSTYPE ; Cnt ++)                                                    /* Insert the given block type IDs */
    if (!TapeWindow[WhichWindow].EmptyTape || Types[Cnt].AllowEmptyTape)
    {
      if (Types[Cnt].BlockTypeID == 0x11)                                                           /* Special: Normal Speed Data */
        strcpy (TEntry.Text, "Raw Datablock");                                                             /* Name it differently */
      else if (Types[Cnt].BlockTypeID == 0x14)                                                           /* Special: Header Block */
        strcpy (TEntry.Text, "Header Block");
      else if (Types[Cnt].BlockTypeID == 0xFF)                                                            /* Special: BASIC Block */
        strcpy (TEntry.Text, "BASIC Block");
      else if (Types[Cnt].BlockTypeID == 0xF1)                                              /* Special: Custom Info - POKEs Block */
        strcpy (TEntry.Text, "Custom Info - POKEs Block");
      else if (Types[Cnt].BlockTypeID == 0xF2)                                       /* Special: Custom Info - Instructions Block */
        strcpy (TEntry.Text, "Custom Info - Instructions");
      else if (Types[Cnt].BlockTypeID == 0xF3)                                             /* Special: Custom Info - Screen Block */
        strcpy (TEntry.Text, "Custom Info - Spectrum Screen");
      else if (Types[Cnt].BlockTypeID == 0xF4)                                         /* Special: Custom Info - ZX-Edit Document */
        strcpy (TEntry.Text, "Custom Info - ZX-Edit Document");
      else if (Types[Cnt].BlockTypeID == 0xF5)                                                  /* Special: Custom Info - Picture */
        strcpy (TEntry.Text, "Custom Info - Picture");
      else
      {
        GetBlockTypeInfo (Types[Cnt].BlockTypeID, NULL, NULL, &TextLine, NULL, NULL);
        strcpy (TEntry.Text, TextLine);
      }
      TEntry.Identifier = Types[Cnt].BlockTypeID;
      if (!InsertComboEntry (&TWindow, &TEntry, NULL, -1))
        FatalError (1, _ErrOutOfMemory, sizeof (struct ComboEntry));
    }
  PrintComboEntries (&TWindow);
  SelectEntry (&TWindow, 0);
  while (!Done)
  {
    switch (HandleWindow (&CWindow, &ButtonP))
    {
      case WINH_CLOSE   : Done = TRUE; Cancelled = TRUE; break;
      case WINH_RNOHAND : Done = TRUE; Cancelled = (bool)ButtonP->ReactNum; break;
    }
    switch (HandleComboWindow (&TWindow, &SelectedLine))
    {
      case WINH_LINESEL : Done = TRUE; break;
    }
    if (GlobalKey == K_ESC)  { Done = TRUE; Cancelled = TRUE; }
    MouseStatus ();
  }
  if (!Cancelled)
  {
    GetComboEntry (&TWindow, &TEntry, TWindow.LastSelected);
    BlockTypeID = (byte)TEntry.Identifier;                                                   /* Determine the BlockType to insert */
  }
  DestroyComboWindow (&TWindow, 0);
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
  if (Cancelled)
    return (FALSE);
  if (BlockTypeID == 0x11)                                                                                /* Inserting raw data ? */
    if (!FileRequester (180, 50, 48, 29, 20, _Path, _File, _Mask, "Select Raw File", TRUE, FALSE))
      return (FALSE);
  AutoInsertNewTape (WhichWindow);
  if ((NewEntry = (struct TapeEntry_s *)malloc (sizeof (struct TapeEntry_s))) == NULL)
    FatalError (1, _ErrOutOfMemory, sizeof (struct TapeEntry_s));
  memset (NewEntry, 0, sizeof (struct TapeEntry_s));
  NewEntry->BlockTypeID = BlockTypeID;                                                                      /* The requested type */
  NewEntry->ComboSelected = 0x00;
  NewEntry->Flags.IsNormalSpeed = TRUE;
  GetBlockTypeInfo (BlockTypeID, &(NewEntry->IsTapeData), &(NewEntry->IsIdentifiedData), NULL, NULL, NULL);
  NewEntry->Next = NULL;
  if (TapeWindow[WhichWindow].Directory == NULL || TapeWindow[WhichWindow].SelectedLine < 0)                      /* Empty tape ? */
  {
    TapeWindow[WhichWindow].Directory = NewEntry;
    _EntryNumber = 0;
  }
  else
  {
    GetComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &TEntry, TapeWindow[WhichWindow].SelectedLine);
    DirP = (struct TapeEntry_s *)TEntry.Identifier;                                                  /* Determine insertion point */
    _EntryNumber = GetBlockNumber (TEntry.Text);
    if (Append)                                                                                    /* Append rather than insert ? */
    {
      if (DirP->ComboSelected & EXPANDBIT)                                                              /* Collapsed Group/Loop ? */
      {
        MatchingBlockTypeID = (DirP->BlockTypeID == 0x21 ? 0x22 : 0x25);                                       /* Group or Loop ? */
        if (DirP->Next != NULL)
          DirP = DirP->Next;
        while (DirP->Next != NULL && DirP->BlockTypeID != MatchingBlockTypeID)                               /* Find matching End */
        {
          DirP = DirP->Next;
          _EntryNumber ++;
        }
      }
      NewEntry->Next = DirP->Next;
      DirP->Next = NewEntry;                                                                                 /* The actual append */
      _EntryNumber ++;
    }
    else if (TapeWindow[WhichWindow].Directory == DirP)                                        /* Insert before the first entry ? */
    {
      NewEntry->Next = TapeWindow[WhichWindow].Directory;
      TapeWindow[WhichWindow].Directory = NewEntry;
    }
    else
    {
      Directory = TapeWindow[WhichWindow].Directory;
      while (Directory->Next != NULL && Directory->Next != DirP)
        Directory = Directory->Next;
      NewEntry->Next = DirP;
      Directory->Next = NewEntry;                                                                         /* The actual insertion */
    }
  }
  TapeWindow[WhichWindow].NumEntries ++;
  TapeWindow[WhichWindow].EmptyTape = FALSE;
  for (Cnt = 0 ; Cnt < MAXINSTYPE ; Cnt ++)                                                              /* Find the edit handler */
    if (Types[Cnt].BlockTypeID == BlockTypeID)
      if (Types[Cnt].EditHandler == NULL)                                                                 /* No handler defined ? */
      {
        _EntryNumber = -1;
        return (TRUE);
      }
      else
      {
        Types[Cnt].EditHandler (WhichWindow, NewEntry);                                                  /* Go edit the new entry */
        _EntryNumber = -1;
        if (_Cancelled)
        {
          _Cancelled = FALSE;
          return (FALSE);
        }
        return (TRUE);
      }
  _EntryNumber = -1;
  return (TRUE);
}
