/**********************************************************************************************************************************/
/* Project      : TAPER                                                                                                           */
/* Module       : TPSCREEN.C                                                                                                      */
/* Description  : Screen dump                                                                                                     */
/* Version type : Program module                                                                                                  */
/* Last changed : 04-01-1999  8:30                                                                                                */
/* Update count : 14                                                                                                              */
/* OS type      : PC (DOS)                                                                                                        */
/* Compiler     : Watcom C32 Optimizing Compiler Version 10.x                                                                     */
/* Linker       : Watcom C Linker                                                                                                 */
/*                                                                                                                                */
/*                Copyleft (C) 1997-2001 ThunderWare Research Center, written by Martijn van der Heide.                           */
/*                                                                                                                                */
/*                This program is free software; you can redistribute it and/or                                                   */
/*                modify it under the terms of the GNU General Public License                                                     */
/*                as published by the Free Software Foundation; either version 2                                                  */
/*                of the License, or (at your option) any later version.                                                          */
/*                                                                                                                                */
/*                This program is distributed in the hope that it will be useful,                                                 */
/*                but WITHOUT ANY WARRANTY; without even the implied warranty of                                                  */
/*                MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                                                   */
/*                GNU General Public License for more details.                                                                    */
/*                                                                                                                                */
/*                You should have received a copy of the GNU General Public License                                               */
/*                along with this program; if not, write to the Free Software                                                     */
/*                Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                                     */
/*                                                                                                                                */
/**********************************************************************************************************************************/

#define __TPSCREEN_MAIN__

#include "wbench32/wbench32.h"
#include <io.h>
#include <fcntl.h>
#include "tpfile.h"
#include "tpconfig.h"
#include "tpmain.h"
#include "tptape.h"
#include "tpaudio.h"
#include "tpdisass.h"
#include "tpscreen.h"
#include "tpbasic.h"

#pragma pack          ()

#define MAXVWLINES   40                                                          /* Max no. Spectrum lines that fit on the screen */

static byte          *_ScreenBuffer;

static byte           _ConvertBuffer[4][12288];                                            /* This buffer is used for 2 purposes: */
                                                                                /* 1. As planedata buffer of [4][6336] (33 * 192) */
                                                                              /* 2. As readbuffer: one large array of 49152 bytes */

static byte           _SpectrumPalette[16] = {DBLACK, DBLUE, DRED, DMAGENTA, DGREEN, DCYAN, DYELLOW, DWHITE,
                                              LBLACK, LBLUE, LRED, LMAGENTA, LGREEN, LCYAN, LYELLOW, LWHITE};

static char           _PicturePath[MAXDIR + 1];

static word           _BlockOffset;
static struct WindowInfo _TBWindow  = {185, 150, 264, 232, 264, 232, 264, 232, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                       WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "Tape Block Image"};

static void _TBTimedFlash     (struct WindowInfo *Window);
static bool _SetupPicturePath (short WhichWindow, bool IsSCR);

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> START OF THE INTERNAL SCREEN LIBRARY ROUTINES <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/**********************************************************************************************************************************/

static void _TBTimedFlash (struct WindowInfo *Window)

/**********************************************************************************************************************************/
/* Pre   : A spectrum screen has to be set up and drawn already.                                                                  */
/* Post  : This is the timed flash handler. Each run toggles the flashing parts.                                                  */
/* Import: PutSpectrumScreen.                                                                                                     */
/**********************************************************************************************************************************/

{
  Window = Window; /* Keep compiler happy */
  FlashState = !FlashState;
  PutSpectrumScreen (_ScreenBuffer + _BlockOffset, _TBWindow.StartX + 4, _TBWindow.StartY + 4);
}

static bool _SetupPicturePath (short WhichWindow, bool IsSCR)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `IsSCR' is TRUE for `.SCR' files and FALSE for `.GIF' files.                */
/* Post  : The user is allowed to browse for a path to store a picture, which is stored in (global) _PicturePath.                 */
/*         The return value is TRUE if all went well, or FALSE is the user Cancelled.                                             */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  short register Cnt;
  char           Path[MAXDIR + 1];
  char           File[13];
  char           Mask[13] = "*.*";
  bool           Done = FALSE;

  Cnt = strlen (TapeWindow[WhichWindow].TapeName);
  while (-- Cnt >= 0 && TapeWindow[WhichWindow].TapeName[Cnt] != '\\' && TapeWindow[WhichWindow].TapeName[Cnt] != ':')
    ;
  if (Cnt >= 0)                                                                                      /* Is there a leading path ? */
  {
    strncpy (Path, TapeWindow[WhichWindow].TapeName, Cnt);
    Path[Cnt] = '\0';
  }
  else
    Path[0] = '\0';
  strcpy (File, TapeWindow[WhichWindow].TapeName + Cnt + 1);
  Cnt = strlen (File);
  while (File[-- Cnt] != '.')                                                                             /* Find extension start */
    ;
  strcpy (File + Cnt + 1, IsSCR ? "SCR" : "GIF");                                                             /* Change extension */
  while (!Done)
    if (FileRequester (180, 50, 48, 29, 20, Path, File, Mask, "Choose Picture Name", FALSE, FALSE))
    {
      if (MaxFloppyIdChanged)
        WriteConfig ();
      MakeFullPath (_PicturePath, Path, File);
      if (!TestFileOverwrite (_PicturePath))
        return (TRUE);
    }
    else
    {
      if (MaxFloppyIdChanged)
        WriteConfig ();
      return (FALSE);
    }
  return (FALSE);
}

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> START OF THE EXPORTED SCREEN LIBRARY ROUTINES <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/**********************************************************************************************************************************/

void DecryptBlock (short WhichWindow, byte *Start, word Length)

/**********************************************************************************************************************************/
/* Pre   : `Start' points to the start of the datablock, `Length' holds the length of the block.                                  */
/* Post  : If the block was (known to be) crypted, it is now decrypted.                                                           */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct ComboEntry ComEntry;
  word     register Cnt;
  byte              UsedXOR;
  byte              UsedADD;

  if (!TapeWindow[WhichWindow].AutoDecryptBlocks)                                                            /* Don't want this ? */
    return;
  if (GlobalDecodingScheme == DS_NONE && !DecryptAnything)                                                 /* No encryption set ? */
    return;
  GetComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &ComEntry, TapeWindow[WhichWindow].SelectedLine);
  if (((struct TapeEntry_s *)ComEntry.Identifier)->BlockTypeID != 0x14 &&                     /* Only consider 'Pure Data' blocks */
      !DecryptAnything)
    return;
  if (UseDecryptUserValues)
  { UsedXOR = DecryptXOR; UsedADD = DecryptADD; }
  else if (GlobalDecodingScheme == DS_SLOCK2 || GlobalDecodingScheme == DS_SLOCK3 || GlobalDecodingScheme == DS_NONE)
  { UsedXOR = 0x98; UsedADD = 0x0B; }
  else
  { UsedXOR = 0xC1; UsedADD = 0x11; }
  for (Cnt = 0 ; Cnt < Length ; Cnt ++)
    *(Start + Cnt) = (byte)((*(Start + Cnt) ^ UsedXOR) + UsedADD);
}

void EncryptBlock (short WhichWindow, byte *Start, word Length)

/**********************************************************************************************************************************/
/* Pre   : `Start' points to the start of the datablock, `Length' holds the length of the block.                                  */
/* Post  : If the block was (known to be) decrypted, it is now encrypted again.                                                   */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct ComboEntry ComEntry;
  word     register Cnt;
  byte              UsedXOR;
  byte              UsedADD;

  if (!TapeWindow[WhichWindow].AutoDecryptBlocks)                                                            /* Don't want this ? */
    return;
  if (GlobalDecodingScheme == DS_NONE && !DecryptAnything)                                                 /* No encryption set ? */
    return;
  GetComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &ComEntry, TapeWindow[WhichWindow].SelectedLine);
  if (((struct TapeEntry_s *)ComEntry.Identifier)->BlockTypeID != 0x14 &&                     /* Only consider 'Pure Data' blocks */
      !DecryptAnything)
    return;
  if (UseDecryptUserValues)
  { UsedXOR = DecryptXOR; UsedADD = DecryptADD; }
  else if (GlobalDecodingScheme == DS_SLOCK2 || GlobalDecodingScheme == DS_SLOCK3 || GlobalDecodingScheme == DS_NONE)
  { UsedXOR = 0x98; UsedADD = 0x0B; }
  else
  { UsedXOR = 0xC1; UsedADD = 0x11; }
  for (Cnt = 0 ; Cnt < Length ; Cnt ++)
    *(Start + Cnt) = (byte)((*(Start + Cnt) - UsedADD) ^ UsedXOR);
}

void WriteSCRFile (short WhichWindow, byte *Buffer)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `Buffer' points to the start of the displayed buffer part.                  */
/* Post  : The screenbuffer has been written as a screenshot file.                                                                */
/* Import: _SetupPicturePath.                                                                                                     */
/**********************************************************************************************************************************/

{
  int Fh;

  if (!_SetupPicturePath (WhichWindow, TRUE))
    return;
  if ((Fh = open (_PicturePath, O_BINARY | O_RDWR | O_CREAT | O_TRUNC, 0777)) == -1)
  { ErrorRequester (-1, -1, "@Stop", NULL, NULL, "Error creating file %s", _PicturePath); return; }
  if (write (Fh, Buffer, 6912) == -1)                                                                           /* Write the file */
  { close (Fh); ErrorRequester (-1, -1, "@Close", NULL, NULL, "Error writing to file %s", _PicturePath); return; }
  close (Fh);
  SoundErrorBeeper (BEEP_SHORT);
}

void WriteGIFFile (short WhichWindow, short StartX, short StartY)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT, `StartX', `StartY' hold the top-left coordinate at screen where the         */
/*         spectrum image is displayed.                                                                                           */
/* Post  : A 256x192x16 GIF-file has been made.                                                                                   */
/* Import: _SetupPicturePath.                                                                                                     */
/**********************************************************************************************************************************/

{
  byte  register  Q;                                                                                                   /* Counter */
  byte            ErrCode;
  dword          *PP;                                                                                          /* Palette Pointer */

  if (!_SetupPicturePath (WhichWindow, FALSE))
    return;
  HideMouse ();
  if ((ErrCode = GIFCreate (_PicturePath, 320, 240, 16, 6)) != GIF_OK)                             /* Choose a bigger screen size */
  { ErrorRequester (-1, -1, NULL, NULL, NULL, "Cannot create GIF file (%s)", GIFstrerror (ErrCode)); return; }
  PP = Screen.CurrentPal;
  for (Q = 0 ; Q < Screen.NumColors ; Q ++, PP ++)
    GIFSetColor (Q, (byte)(*PP & 0x000000FF), (byte)((*PP & 0x0000FF00) >> 8), (byte)((*PP & 0x00FF0000) >> 16));
  if ((ErrCode = GIFCompressImage (StartX, StartY, 256, 192)) != GIF_OK)
  { ErrorRequester (-1, -1, NULL, NULL, NULL, "Cannot compress GIF image (%s)", GIFstrerror (ErrCode)); GIFClose ();
    unlink (_PicturePath); return; }
  if ((ErrCode = GIFClose ()) != GIF_OK)
  { ErrorRequester (-1, -1, NULL, NULL, NULL, "Cannot close GIF file (%s)", GIFstrerror (ErrCode)); unlink (_PicturePath); return; }
  ShowMouse ();
  SoundErrorBeeper (BEEP_SHORT);
}

void PutSpectrumScreen (byte *ScreenBuffer, short HorOff, short VerOff)

/**********************************************************************************************************************************/
/* Pre   : `ScreenBuffer' points to the 6912 bytes spectrum screen, `HorOff' holds the horizontal, `VerOff' holds the vertical    */
/*         pixel coordinate to put the spectrum screen on the VGA display; `RemapBrightBlack' is TRUE if the ULA bug should be    */
/*         used (bright and dark black both resolve to dark black).                                                               */
/* Post  : The spectrum screen has been put on the screen.                                                                        */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  byte            PixelMask;
  byte            PaletteMask;
  byte            PixelData;
  byte            ScreenBlockIndex;
  byte            ScreenBlockCount;
  word            PlaneDataIndex;
  word            ScreenDataIndex;
  word  register  PaletteIndex;
  byte           *StartAddress;                                           /* The first address in display memory to be considered */
  byte           *Address;                                              /* The current address in display memory being considered */
  short register  CntY;                                                                              /* Vertical scanline counter */
  short register  CntX;                                                                                /* Horizontal byte counter */
  word  register  CntPlane;                                                      /* Binary representation of the current bitplane */
  byte  register  ReadPlane;                                                                           /* Linear bitplane counter */
  byte            LeftMask;
  byte            ShiftRight;                                                                        /* Bit-shift within the byte */

  StartAddress = Screen.DisplayStart + 80 * VerOff + HorOff / 8;
  ShiftRight = HorOff % 8;
  LeftMask = 0xFF >> ShiftRight;

  _SpectrumPalette[8] = (TPConfig.RemapBrightBlack ? DBLACK : LBLACK);

  ScreenContainsFlash = FALSE;
  PaletteIndex = PlaneDataIndex = 0;
  for (ScreenBlockCount = 0 ; ScreenBlockCount < 3 ; ScreenBlockCount ++)              /* Spectrum screen is composed of 3 blocks */
    for (CntY = 0 ; CntY < 8 ; CntY ++)                                                        /* 8 character lines in each block */
    {
      ScreenDataIndex = ScreenBlockCount * 2048 + CntY * 32;
      for (ScreenBlockIndex = 0 ; ScreenBlockIndex < 8 ; ScreenBlockIndex ++)                  /* Each character has 8 scan lines */
      {
        for (CntX = 0 ; CntX < 32 ; CntX ++)
        {
          PaletteMask = *(ScreenBuffer + 6144 + PaletteIndex + CntX);
          if (PaletteMask & 0x80)
            ScreenContainsFlash = TRUE;
          if ((PaletteMask & 0x80) && FlashState)
            PixelMask = _SpectrumPalette[(PaletteMask & 0x78) >> 3];
          else
            PixelMask = _SpectrumPalette[(PaletteMask & 0x07) | ((PaletteMask & 0x40) >> 3)];
          PixelData = *(ScreenBuffer + ScreenDataIndex + CntX);
          for (CntPlane = 0, ReadPlane = 1; CntPlane < Screen.NumPlanes ; CntPlane ++, ReadPlane <<= 1)
            _ConvertBuffer[CntPlane][PlaneDataIndex] = (PixelMask & ReadPlane ? PixelData : 0);
          if ((PaletteMask & 0x80) && FlashState)
            PixelMask = _SpectrumPalette[(PaletteMask & 0x07) | ((PaletteMask & 0x40) >> 3)];
          else
            PixelMask = _SpectrumPalette[(PaletteMask & 0x78) >> 3];
          for (CntPlane = 0, ReadPlane = 1; CntPlane < Screen.NumPlanes ; CntPlane ++, ReadPlane <<= 1)
            _ConvertBuffer[CntPlane][PlaneDataIndex] |= (PixelMask & ReadPlane ? ~PixelData : 0);
          PlaneDataIndex ++;
        }
        ScreenDataIndex += 256;                                                       /* One spectrum scan line down is 256 bytes */
        PlaneDataIndex ++;                                                                          /* Skip past the padding byte */
      }
      PaletteIndex += 32;                                                                     /* Palette is `correctly' addressed */
    }

  if (ShiftRight != 0)
    for (ReadPlane = 0 ; ReadPlane < Screen.NumPlanes ; ReadPlane ++)
      for (CntY = 0 ; CntY < 192 ; CntY ++)
      {
        _ConvertBuffer[ReadPlane][CntY * 33 + 32] = 0;
        for (CntX = 31 ; CntX >= 0 ; CntX --)
          _ConvertBuffer[ReadPlane][CntPlane] = (_ConvertBuffer[ReadPlane][CntPlane = CntY * 33 + CntX + 1] >> ShiftRight) |
                                                 (_ConvertBuffer[ReadPlane][CntPlane - 1] << (8 - ShiftRight));
        _ConvertBuffer[ReadPlane][CntY * 33] >>= ShiftRight;
      }

  HideMouseRange (HorOff, VerOff, HorOff + 255, VerOff + 191);
  StartScreenUpdate (HorOff, VerOff, HorOff + 255, VerOff + 191);
  for (ReadPlane = 0 ; ReadPlane < Screen.NumPlanes ; ReadPlane ++)                                       /* Handle all bitplanes */
  {
    Address = StartAddress;
    SelectPlane (ReadPlane);
    for (CntY = 0 ; CntY < 192 ; CntY ++)
    {
      *(Address ++) = (*(Address) & ~LeftMask) | (_ConvertBuffer[ReadPlane][CntY * 33] & LeftMask);
      for (CntX = 1 ; CntX < 32 ; CntX ++)
        *(Address ++) = _ConvertBuffer[ReadPlane][CntY * 33 + CntX];
      if (ShiftRight == 0)
        Address += 48;
      else
      {
        *(Address ++) = (*(Address) & LeftMask) | (_ConvertBuffer[ReadPlane][CntY * 33 + CntX] & ~LeftMask);
        Address += 47;
      }
    }
  }
  StopScreenUpdate ();
  ShowMouse ();
}

void ViewSpectrumHeader (short WhichWindow)

/**********************************************************************************************************************************/
/* Pre   : `WhichWindow' is either TWLEFT or TWRIGHT.                                                                             */
/* Post  : The selected block in that window has been shown as if it were a header block.                                         */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  struct WindowInfo CWindow  = {184, 130, 216, 100, 216, 100, 216, 100, 1, WINB_OUT | WINB_SURROUND, DBLACK, DWHITE, NULL,
                                WINF_HUSE | WINF_HCENTER | WINF_ACTIVATE | WINF_XMSBUFFER, NULL, "Tape Block Information"};
  struct ButtonInfo POk      = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 81, 75, 8, DBLACK, DWHITE, NULL, 1, "@Ok", NULL};

  struct ButtonInfo   *ButtonP;
  struct TapeRecord_s *TRecord;
  char                 NameBuffer[88];
  short                HandleRet;
  char                *TTypes[4] = { "Program", "Numerical array", "Alphanumerical array", "Bytes" };
  short                TokenLen;
  short                ExpLen;

  WhichWindow = WhichWindow; /* keep compiler happy */
  AddButton (&CWindow, &POk, TRUE);
  DrawWindow (&CWindow, NULL, WING_CLOSE, TRUE);
  TRecord = (struct TapeRecord_s *)_ScreenBuffer;
  if (TRecord->HType < 4)
    Wprintf (&CWindow, 1, 1, "{+1Type          }N %s", TTypes[TRecord->HType]);
  else
    Wprintf (&CWindow, 1, 1, "{+1Type          }N (%u)", TRecord->HType);
  ExpLen = RemapSpectrumString (TRecord->HName, 10, NameBuffer, 80, FALSE, FALSE, &TokenLen, TRUE, FALSE, FALSE);
  NameBuffer[ExpLen] = '\0';                                                                       /* Expand filename from header */
  strcat (NameBuffer, "{++{--}L");
  Wprintf (&CWindow, 2, 1, "{+1Name          }N \"%s\"", NameBuffer);
  Wprintf (&CWindow, 3, 1, "{+1Length        }N %u", TRecord->HLength);
  switch (TRecord->HType)                                                                               /* Block type from header */
  {
    case 0 : if (TRecord->HStart > 32767)
               Wprintf (&CWindow, 4, 1, "{+1Auto start    }N -----  ");
             else
               Wprintf (&CWindow, 4, 1, "{+1Auto start    }N %u", TRecord->HStart);
             Wprintf (&CWindow, 5, 1, "{+1BASIC size    }N %u", TRecord->HExtra);
             break;
    case 1 : break;
    case 2 : break;
    case 3 : Wprintf (&CWindow, 4, 1, "{+1Start address }N %u", TRecord->HStart);
             break;
    default: Wprintf (&CWindow, 4, 1, "{+1Start address }N %u", TRecord->HStart);        /* Usually in M1 blocks; just 17 bytes */
             Wprintf (&CWindow, 5, 1, "{+1Extra param   }N %u", TRecord->HExtra);
  }
  while ((HandleRet = HandleWindow (&CWindow, &ButtonP)) != WINH_RNOHAND && HandleRet != WINH_CLOSE && GlobalKey != K_ESC)
    MouseStatus ();
  DestroyWindow (&CWindow);
  GlobalKey = 0x0000;
}

void ViewSpectrumScreenBlock (short WhichWindow, word BlockLength)

/**********************************************************************************************************************************/
/* Pre   : `BlockLength' holds the length of the block. The block has already been moved to _ScreenBuffer.                        */
/* Post  : The selected entry has been converted to a screen and put in a seperate window, together with a `Make SCR' and         */
/*         `Make GIF' button. If one of these buttons has been pressed, an image file has been created.                           */
/*         The user must explicitely use the close gadget.                                                                        */
/* Import: DecryptBlock, PutSpectrumScreen, WriteSCRFile, WriteGIFFile.                                                           */
/**********************************************************************************************************************************/

{
  struct ButtonInfo PMkGIF   = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 19, 210, 10, DBLACK, DWHITE, NULL, 1, "Make @GIF",
                                NULL};
  struct ButtonInfo PMkSCR   = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 106, 210, 10, DBLACK, DWHITE, NULL, 2, "Make @SCR",
                                NULL};
  struct ButtonInfo PDecOff  = {BT_CPUSH, BC_GADGET, BF_FWIDTH | BF_KEYSELECT, 189, 210, 1, DBLACK, DWHITE, NULL, 3, "\003", NULL};
  struct ButtonInfo PSetOff  = {BT_INPUT, BC_NUMBER, BF_FWIDTH | BF_KEYSELECT | BF_FITTXT, 201, 210, 5, DBLACK, DWHITE, NULL, 4,
                                NULL, NULL};
  struct ButtonInfo PIncOff  = {BT_CPUSH, BC_GADGET, BF_FWIDTH | BF_KEYSELECT, 238, 210, 1, DBLACK, DWHITE, NULL, 5, "\004", NULL};
  struct ButtonInfo *ButtonP;
  char              Offset[6] = "0";
  bool              Done;
  bool              Redraw = TRUE;
  dword             COffset;
  int               MaxOffset;

  PSetOff.Contents = Offset;
  AddButton (&_TBWindow, &PMkGIF, TRUE);                                    /* Window MUST be created first, as the _ScreenBuffer */
  AddButton (&_TBWindow, &PMkSCR, FALSE);                                            /* is global now and used to draw the window */
  AddButton (&_TBWindow, &PDecOff, FALSE);
  AddButton (&_TBWindow, &PSetOff, FALSE);
  AddButton (&_TBWindow, &PIncOff, FALSE);
  DrawWindow (&_TBWindow, NULL, WING_CLOSE, TRUE);
  SurroundScreen (&_TBWindow, 2, 2);
  DecryptBlock (WhichWindow, _ScreenBuffer, BlockLength);
  if (BlockLength <= 6912)                                                                             /* Less than a screenful ? */
  {                                                                                 /* Then we need to fill up the remaining part */
    if (BlockLength < 6144)
    {
      memset (_ScreenBuffer + BlockLength, 0, 6144 - BlockLength);                                            /* DisplayFile part */
      MaxOffset = 0;
    }
    else
      MaxOffset = BlockLength - 6144;
    if (MaxOffset < 0)
      MaxOffset = 0;
    _BlockOffset = 6912 - BlockLength;
    if (_BlockOffset > 768)
      _BlockOffset = 768;
    memset (_ScreenBuffer + 6144 + MaxOffset, 0x38, _BlockOffset);                                             /* Attributes file */
    MaxOffset = 0;
  }
  else
    MaxOffset = BlockLength - 6912;
  while (Redraw)
  {
    _BlockOffset = atoi (Offset);                                                                          /* Read offset setting */
    PutSpectrumScreen (_ScreenBuffer + _BlockOffset, _TBWindow.StartX + 4, _TBWindow.StartY + 4);
    ReleaseTimedHandler (_TBTimedFlash, &_TBWindow);
    if (ScreenContainsFlash && TPConfig.EmulateFlash)
      RegisterTimedHandler (TPConfig.SlowFlash ? 960 : 320, _TBTimedFlash, &_TBWindow);
    Done = FALSE;
    while (!Done)
    {
      switch (HandleWindow (&_TBWindow, &ButtonP))
      {
        case WINH_RNOHAND : switch (ButtonP->ReactNum)
                            {
                              case 1 : WriteGIFFile (WhichWindow, _TBWindow.StartX + 4, _TBWindow.StartY + 4);      /* GIF button */
                                       break;
                              case 2 : WriteSCRFile (WhichWindow, _ScreenBuffer + _BlockOffset);                    /* SCR button */
                                       break;
                            }
                            break;
        case WINH_PNOHAND : switch (ButtonP->ReactNum)
                            {
                              case 3 : if (_BlockOffset > 0)                                                     /* DecOff button */
                                       {
                                         sprintf (Offset, "%d", _BlockOffset - 1);
                                         UpdateButtonText (&_TBWindow, &PSetOff, BF_FITTXT);
                                         Done = TRUE;                                                                  /* Redraw! */
                                       }
                                       break;
                              case 5 : if (_BlockOffset < MaxOffset)                                             /* IncOff button */
                                       {
                                         sprintf (Offset, "%d", _BlockOffset + 1);
                                         UpdateButtonText (&_TBWindow, &PSetOff, BF_FITTXT);
                                         Done = TRUE;                                                                  /* Redraw! */
                                       }
                                       break;
                            }
                            break;
        case WINH_EDITEND : if (ButtonP->Flags & BF_FITINPUT)                                               /* Has been changed ? */
                            {
                              COffset = atol (Offset);
                              if (COffset > MaxOffset)
                                _BlockOffset = MaxOffset;
                              else
                                _BlockOffset = COffset;
                              sprintf (Offset, "%d", _BlockOffset);
                              UpdateButtonText (&_TBWindow, &PSetOff, BF_FITTXT);
                              Done = TRUE;                                                                             /* Redraw! */
                            }
                            break;
        case WINH_CLOSE   : Done = TRUE;
                            Redraw = FALSE;
      }
      if (GlobalKey == K_ESC)                                                                                  /* <ESC> stops too */
      {
        Done = TRUE;
        Redraw = FALSE;
      }
      MouseStatus ();
    }
  }
  DestroyWindow (&_TBWindow);
  GlobalKey = 0x0000;
}

bool ViewSpectrumBASICListing (short WhichWindow, word BlockLength, bool Snap128K)

/**********************************************************************************************************************************/
/* Pre   : `BlockLength' holds the length of the tape block, `Snap128K' is TRUE for 128K tapes.                                   */
/*         The block has already been copied to _ScreenBuffer.                                                                    */
/* Post  : The BASIC Listing has been displayed in a view requester and possibly edited by the user.                              */
/*         The return value is TRUE if the listing has been edited.                                                               */
/* Import: DecryptBlock.                                                                                                          */
/**********************************************************************************************************************************/

{
  struct PrintAttributes Attributes;
  struct ButtonInfo     *ButtonP;
  struct WindowInfo      ViewWindow;
  struct ComboInfo       TextWindow;
  bool                   DropColors    = FALSE;
  bool                   RealNumbers   = TRUE;
  struct ButtonInfo      BDropColors   = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 7, 10, 0, DBLACK, DWHITE,
                                          NULL, 1, NULL, NULL};
  struct ButtonInfo      BRealNumbers  = {BT_BOOL, BC_GADGET, BF_KEYSELECT | BF_BCLEAR, 103, 10, 0, DBLACK, DWHITE,
                                          NULL, 3, NULL, NULL};
  struct ButtonInfo      BEdit         = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 187, 10, 6, DBLACK, DWHITE,
                                          NULL, 4, "@Edit", NULL};
  struct ButtonInfo      BSave         = {BT_PUSH, BC_TEXT, BF_FWIDTH | BF_KEYSELECT, 234, 10, 6, DBLACK, DWHITE,
                                          NULL, 2, "@Save", NULL};
  FILE                  *Fp;
  struct ComboEntry      NEntry;
  short                  InsertedChars;
  short                  MaxEntrySize;
  comboent               OldDFirstLine  = -1;
  byte                  *Listing        = _ScreenBuffer;
  byte                  *NextLine;
  byte                  *Maximum;
  byte                  *ListIndex;
  word                   Number;
  word                   LineNumber;
  dword                  OldLineNumber;
  bool                   More          = TRUE;
  bool                   Resizing      = FALSE;
  bool                   StillResizing = FALSE;
  bool                   Rereading     = TRUE;
  bool                   DoWrite;
  comboent               Dummy;
  char                   Line[2049];
  char                   TempFileName[MAXDIR];
  char                   SavePath[MAXDIR];
  char                   SaveFilename[13];
  char                   SaveExt[5];
  static char            SaveMask[13]  = "*.BAS";
  byte                   CFlags        = COMF_NOBUFFER | COMF_VSLIDER | COMF_SYNCSLIDER | COMF_KEYSELECT;
  short                  CurrentWidth;
  short                  CurrentHeight;
  bool                   DataChanged = FALSE;

  BDropColors.Contents = &DropColors;
  BRealNumbers.Contents = &RealNumbers;
  SelectFont (NULL, FONT_NORMAL);                                            /* Listings are of course in the Spectrum's own font */
  memset (&ViewWindow, 0, sizeof (struct WindowInfo));                                           /* Initialize the interface part */
  CurrentWidth           = 32;
  CurrentHeight          = 24;
  MaxEntrySize           = 64;
  ViewWindow.StartX      = 185;
  ViewWindow.StartY      = 150;
  ViewWindow.Width       = ViewWindow.MinWidth = SpectrumFont->Width * CurrentWidth + 27;
  ViewWindow.MaxWidth    = SpectrumFont->Width * 64 + 27;                                           /* Max width is 64 characters */
  ViewWindow.Height      = ViewWindow.MinHeight = (SpectrumFont->Height + 2) * CurrentHeight + 42;
  ViewWindow.MaxHeight   = (SpectrumFont->Height + 2) * MAXVWLINES + 42;
  ViewWindow.BorderFlags = WINB_OUT | WINB_SURROUND;
  ViewWindow.BorderSize  = 1;
  ViewWindow.APen        = SystemColors.ReqAPen;
  ViewWindow.BPen        = SystemColors.ReqOPen;
  ViewWindow.Header      = "BASIC Listing";
  ViewWindow.Flags       = WINF_ACTIVATE | WINF_HUSE | WINF_HCENTER | WINF_XMSBUFFER;

  AddButton (&ViewWindow, &BDropColors, FALSE);
  AddButton (&ViewWindow, &BRealNumbers, FALSE);
  AddButton (&ViewWindow, &BEdit, FALSE);
  AddButton (&ViewWindow, &BSave, FALSE);
  DrawWindow (&ViewWindow, NULL, WING_CLOSE | WING_RESIZEMAX, TRUE);
  Wprintf (&ViewWindow, 1, 4, "{+1Drop Colors}N");
  Wprintf (&ViewWindow, 1, 20, "{+1Real Nums}N");
  DecryptBlock (WhichWindow, _ScreenBuffer, BlockLength);

  while (More)                                                                                    /* Handle until close requested */
  {
    if (Resizing)
    {
      CurrentWidth = (short)((ViewWindow.Width - 27) / SpectrumFont->Width);
      CurrentHeight = (short)((ViewWindow.Height - 42) / (SpectrumFont->Height + 2));
      MaxEntrySize = CurrentWidth * 2;
      ViewWindow.Width  = SpectrumFont->Width * CurrentWidth + 27;
      ViewWindow.Height = (SpectrumFont->Height + 2) * CurrentHeight + 42;
      AddButton (&ViewWindow, &BDropColors, FALSE);
      AddButton (&ViewWindow, &BRealNumbers, FALSE);
      AddButton (&ViewWindow, &BEdit, FALSE);
      AddButton (&ViewWindow, &BSave, FALSE);
      DrawWindow (&ViewWindow, NULL, WING_CLOSE | WING_RESIZEMAX, TRUE);
      Wprintf (&ViewWindow, 1, 4, "{+1Drop Colors}N");
      Wprintf (&ViewWindow, 1, 20, "{+1Real Nums}N");
      Resizing = FALSE;
      Rereading = TRUE;
      StillResizing = TRUE;
    }
    if (Rereading)
    {
      MakeComboWindow (&TextWindow, &ViewWindow, COMT_VIEW, CFlags, ViewWindow.A_StartX + 12, ViewWindow.StartY + 32, 0, 1024,
                       MaxEntrySize, CurrentWidth, CurrentHeight, SpectrumFont, FALSE, 0, NULL, SystemColors.ReqAPen,
                       SystemColors.ReqBPen, 0, 0);
      TextWindow.ComboWindow.TABLength = 16;                                                   /* Spectrum has a TAB length of 16 */
      TextWindow.BlinkRate = TPConfig.EmulateFlash ? (TPConfig.SlowFlash ? 960 : 320) : 0;            /* Set Spectrum blink speed */
      Rereading = FALSE;
      RemapSpectrumUDGs ((bool)!Snap128K);
      NextLine = Listing;
      ListIndex = Listing;
      Maximum = Listing + BlockLength;
      GetComboAttributes (&TextWindow, &Attributes);
      while (ListIndex < Maximum && *ListIndex < 64 && More)                                          /* Up to the variables area */
      {
        LineNumber = (*ListIndex * 256) + *(ListIndex + 1);                                         /* Line number is big endian! */
        sprintf (Line, "%4d", LineNumber);                                                              /* Insert the line number */
        ListIndex += 2;
        Number = *ListIndex + 256 * *(ListIndex + 1);                                                              /* Line length */
        ListIndex += 2;
        NextLine = ListIndex + Number;                                                             /* Adjust pointer to next line */
        if (NextLine > Maximum)
          Number = Maximum - ListIndex;
        if (Number < 32768 && ListIndex < Maximum)
          RemapSpectrumString (ListIndex, Number, Line + 4, 2038, DropColors, TRUE, &InsertedChars, TRUE, RealNumbers, FALSE);
        More = FitComboString (&TextWindow, &Attributes, Line, CurrentWidth, 0, LineNumber, TRUE);             /* Insert the line */
        ListIndex = NextLine;                                                                            /* Prepare for next line */
      }
      if (!More)                                                                                /* Listing truncated prematurally */
      {
        if (TextWindow.NumEntries == -1)                                                                     /* No entry at all ? */
          FatalError (1, _ErrOutOfMemory, 0);                                                                        /* Then quit */
        GetComboEntry (&TextWindow, &NEntry, TextWindow.NumEntries);
        strcpy (NEntry.Text, "{+1>>> TRUNCATED <<<}N");
        PutComboEntry (&TextWindow, &NEntry, TextWindow.NumEntries, FALSE);
        More = TRUE;
      }
      if (StillResizing)
      {
        for (Dummy = 0 ; StillResizing && Dummy <= TextWindow.NumEntries ; Dummy ++)
        {
          GetComboEntry (&TextWindow, &NEntry, Dummy);
          if (NEntry.Identifier == OldLineNumber)
          {
            TextWindow.DFirstLine = Dummy;
            TextWindow.VSlider->SliderStart = TextWindow.VSlider->StepRate * Dummy;
            StillResizing = FALSE;
          }
        }
        StillResizing = FALSE;
      }
      else if (OldDFirstLine >= 0)
      {
        TextWindow.DFirstLine = OldDFirstLine;
        TextWindow.VSlider->SliderStart = TextWindow.VSlider->StepRate * OldDFirstLine;
        OldDFirstLine = -1;
      }
      PrintComboEntries (&TextWindow);                                                                      /* Print the contents */
    }
    HandleComboWindow (&TextWindow, &Dummy);
    if (GlobalKey == K_ESC)
      More = FALSE;
    else
    {
      switch (HandleWindow (&ViewWindow, &ButtonP))
      {
        case WINH_CLOSE     : More = FALSE; break;
        case WINH_BUTTBOOL  : OldDFirstLine = TextWindow.DFirstLine;                                             /* (Drop Colors) */
                              DestroyComboWindow (&TextWindow, 0);
                              Rereading = TRUE;
                              break;
        case WINH_WINRESIZE : Resizing = TRUE;
                              GetComboEntry (&TextWindow, &NEntry, TextWindow.DFirstLine);
                              OldLineNumber = NEntry.Identifier;                     /* Remember which line number was at the top */
                              DestroyComboWindow (&TextWindow, 0);                             /* Full destroy, as we must reread */
                              DestroyWindow (&ViewWindow);
                              break;
        case WINH_RNOHAND   : if (ButtonP->ReactNum == 3)                                                               /* (Save) */
                              {
                                _splitpath (TapeWindow[WhichWindow].TapeName, SavePath, Line, SaveFilename, SaveExt);
                                strcat (SavePath, Line);
                                strcat (SaveFilename, ".BAS");
                                DoWrite = FALSE;
                                if (FileRequester (180, 50, 48, 29, 20, SavePath, SaveFilename, SaveMask, "Choose BASIC Save Name",
                                    FALSE, FALSE))
                                {
                                  MakeFullPath (Line, SavePath, SaveFilename);
                                  if (!TestFileOverwrite (Line))
                                    if ((Fp = fopen (Line, "w")) == NULL)
                                      ErrorRequester (-1, -1, "@Close", NULL, NULL, "Error creating file %s:\n    %s",
                                                      Line, StrError (errno));
                                    else
                                      DoWrite = TRUE;
                                }
                              }
                              else                                                                                      /* (Edit) */
                              {
                                DoWrite = FALSE;
                                if (TPConfig.EditPath[0] == '\0')                                          /* No editor defined ? */
                                  ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - No editor has been defined");
                                else
                                {
                                  MakeFullPath (TempFileName, TmpDirectory, mktemp (NULL));
                                  strcpy (TempFileName + strlen (TempFileName) - 3, "BAS");
                                  if ((Fp = fopen (TempFileName, "wb")) == NULL)                     /* Create the temp text file */
                                    ErrorRequester (-1, -1, NULL, NULL, NULL, "ERROR - Cannot create %s\n      - %s",
                                                    TempFileName, StrError (errno));
                                  else
                                    DoWrite = TRUE;
                                }
                              }
                              if (DoWrite) 
                              {
                                NextLine = Listing;
                                ListIndex = Listing;
                                Maximum = Listing + BlockLength;
                                while (ListIndex < Maximum && *ListIndex < 64 && More)
                                {
                                  Number = (*ListIndex * 256) + *(ListIndex + 1);
                                  sprintf (Line, "%4d", Number);
                                  ListIndex += 2;
                                  Number = *ListIndex + 256 * *(ListIndex + 1);
                                  ListIndex += 2;
                                  NextLine = ListIndex + Number;
                                  RemapSpectrumString (ListIndex, Number, Line + 4, 2038, DropColors, TRUE, &InsertedChars,
                                                       FALSE, FALSE, FALSE);
                                  fputs (Line, Fp);
                                  ListIndex = NextLine;
                                }
                                fclose (Fp);
                                if (ButtonP->ReactNum == 3)
                                  SoundErrorBeeper (BEEP_SHORT);
                                else
                                {
                                  GetComboEntry (&(TapeWindow[WhichWindow].IndexWindow), &NEntry,
                                                 TapeWindow[WhichWindow].SelectedLine);
                                  if (EditBASICBlock (WhichWindow, (struct TapeEntry_s *)NEntry.Identifier, TempFileName,
                                                      &BlockLength, ListIndex - Listing, FALSE))
                                  {
                                    OldDFirstLine = TextWindow.DFirstLine;
                                    DestroyComboWindow (&TextWindow, 0);
                                    Rereading = TRUE;
                                    memcpy (Listing, _ObjectData, BlockLength);
                                    SignalTapeChanged (WhichWindow, TRUE);
                                    DataChanged = TRUE;
                                  }
                                  unlink (TempFileName);
                                }
                              }
                              break;
      }
      if (More && GlobalKey == K_ESC)
        More = FALSE;
    }
    MouseStatus ();
  }
  DestroyComboWindow (&TextWindow, 0);
  DestroyWindow (&ViewWindow);
  GlobalKey = 0x0000;
  RemapSpectrumUDGs (TRUE);                                                                                     /* Reset the UDGs */
  return (DataChanged);
}

void AllocateScreenBuffer (void)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : The _ScreenBuffer and _ObjectData buffers are allocated.                                                               */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if ((_ScreenBuffer = (byte *)malloc ((word)49152)) == NULL)
    FatalError (1, _ErrOutOfMemory, (word)49152);
  WorkBuffer = _ScreenBuffer;
  if ((_ObjectData = (byte *)malloc ((dword)65536)) == NULL)                        /* Create Disassembler/HexDump working buffer */
    FatalError (1, _ErrOutOfMemory, (word)65536);
}
